; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; ---------------------------------------------------------------------------
;         .::** EXEC HELPER -- Jump to SubRoutine replacements **::.
; ---------------------------------------------------------------------------

; **
; EH_1668::  Extend the sign of the LOW byte of [R0] (bit #7) into
;            the HIGH byte of [R0], setting Z if the LOW byte of
;            [R0] is = 0.
;

EH_1668::                               ; [R0] <- ( [R0] << 8 ) .SAR. 8

        di                              ; don't allow CopyToVRAM to interrupt

        pop             hl              ; hl <- R0    (OBJ_01)
        pop             de              ; de <- $FF00 (OBJ_02)

        ld              a, [hl+]        ; a <- LOW byte of [R0]
        add             a, a            ; is bit #7 set ?
        jr              nc, .NoSign     ; no, so don't "extend" the sign

        ld              [hl], d         ; set HIGH byte of [R0] = $FF

        set             3, b            ; set Sign (bit #3)
.NoZero
        res             2, b            ; reset Zero (bit #2)

        reti                            ; return, enabling interrupts

.NoSign

        res             3, b            ; reset Sign (bit #3)

        ld              [hl], e         ; set HIGH byte of [R0] = 0

        jr              nz, .NoZero

        set             2, b            ; set Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

; **
; EH_167D::
;

EH_167D::

        ; NOTE:  Real object code definitions are located in CPU_NEW.ASM.

        ; OBJ_00:       EH_167D
        ; OBJ_01:       R0
        ; OBJ_02:       SYSTEM_RAM + $015E * 2

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl -> R0

        ld              a, [hl]         ; a <- LOW byte of [R0]
        or              a               ; is it = 0 ?

        pop             hl              ; hl -> SYSTEM_RAM + $015E * 2

        jr              z, .Exit        ; yes, so exit (must be in 1..16)

        ld              [R5], sp        ; save SP in R5 (zero'd below)

        ld              sp, hl
        pop             de              ; de <- SYSTEM_RAM + $015E * 2

        ei                              ; interrupts can't be OFF for so long

        ld              h, a            ; h = LOW byte of [R0]

        cp              17              ; is [R0] >= 17 ?
        jr              c, .LessThan    ; no, so Bitmask we be OK

        ld              a, 16           ; make sure BitMask will be = $FFFF
.LessThan
        dec             a               ; put in 0..15
        add             a, a
        ld              l, a            ; l = ( LOW byte of [R0] - 1 ) * 2

.GetRandBit

        ld              a, d            ; a = HIGH byte of [$035E]
        swap            a               ; put bit #11 in bit #15
        xor             d               ; and xor w/ bit #15 of [$035E]

        swap            a
        rrca                            ; put bit #15 in bit #2
        xor             e               ; and xor it w/ bit #2 of [$035E]

        rrca
        rrca                            ; put bit #2 in bit #0
        xor             e               ; and xor it w/ bit #0 of [$035E]

        rrca                            ; put bit #0 in CY

        rl              e               ; rotate in "random" bit to bit #0
        rl              d

        dec             h
        jr              nz, .GetRandBit

        ld              h, BitCntToMask >> 8  ; hl -> BitCntToMask
                                              ;       [ LOW of [R0] * 4 ]

        ld              a, [hl+]        ; a = LOW byte of BitMask
        and             e               ;   = LOW byte of BitMask & [$035E]
        ld              [R0], a         ; store out to LOW byte of [R0]

        ld              a, [hl]         ; a = HIGH byte of BitMask
        and             d               ;   = HIGH byte of BitMask & [$035E]
        ld              [R0 + 1], a     ; store out to HIGH byte of [R0]

        di                              ; disable interrupts to restore stack

        push            de              ; update SYSTEM_RAM @ $015E * 2

        ld              sp, R5          ; sp -> R5
        pop             hl

        ld              de, 0
        push            de              ; [R5] <- 0

        ld              sp, hl          ; SP <- [SP_1] (restore stack)
.Exit
        RETURN          0               ; return, enabling interrupts

; **
; EH_1766:  Take the MO record address in [R1] (positioned at any
;           field), and place its corresponding MO number in [R0].
;

EH_1766::                               ; [R0] <- ( [R1] - $031D ) >> 3

        di                              ; don't allow CopyToVRAM to interrupt

        pop             hl              ; hl <- R1    (OBJ_01)
        pop             de              ; de <- $071D (OBJ_02)

        ld              a, [hl-]        ; a <- LOW byte of $031D + MO # * 8

        sub             e               ; a = ..nnn???
        rrca                            ;   = ?..nnn??
        rrca                            ;   = ??..nnn?
        rrca                            ;   = ???..nnn
        and             d               ;   = .....nnn (in case not @ MoAttr)

        ld              [hl], 0         ; store out to HIGH byte of [R0] (0)
        dec             l
        ld              [hl], a         ; store out to LOW byte of [R0]

        RETURN          0               ; return, enabling interrupts

        ; ^ 20 + 4 = 24 cycles, 14 bytes

; **
; EH_176C:  Take the MO number in [R0] and place its corresponding
;           MO record address in [R1] positioned at its 1st field.
;

EH_176C::                               ; [R1] <- $031D + [R0] << 3

        di                              ; don't allow CopyToVRAM to interrupt

        pop             hl              ; hl <- R0 (OBJ_01)

        ld              a, [hl]         ; a <- MO # (0-7)

        add             a, a            ; a = MO # * 2
        add             a, a            ;   = MO # * 4
        add             a, a            ;   = MO # * 8

        add             $1D             ; a = LOW byte of $031D + MO # * 8

        pop             hl              ; hl <- R1 (OBJ_02)

        ld              [hl+], a        ; store out to LOW byte of [R0]
        ld              [hl], l         ; store out to HIGH byte of [R0] (3)

        RETURN          0               ; return, enabling interrupts

        ; ^ 18 + 4 = 22 cycles, 12 bytes

; **
; EH_1629:     Compute separate X & Y velocities based on a direction
;              and single velocity value.
;
; Parameters:  [R0] = Velocity
;              [R1] = Direction (0-15)
;
; Returns:     [R0] = XVel in HIGH, YVel in LOW
;              [R1] = YVel
;

EH_1629::

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- R0    (OBJ_01)
        pop             de              ; de <- $0080 (OBJ_02)

        ld              [SP_1], sp      ; save RetAddr for Exit

        ld              sp, SP_1        ; use STACK_1 for EXEC_162A

        ei                              ; interrupts can't be off for so long

        ; get absolute value of Velocity, reverse Direction if negative

        ld              c, [hl]         ; c <- LOW byte of Velocity
        inc             l
        ld              a, [hl+]
        ld              b, a            ; b <- HIGH byte of Velocity

        and             e               ; is Velocity negative ?
        jr              z, .NotNeg      ; no, so leave it alone

        xor             a
        sub             c
        ld              c, a

        ld              a, d            ; = ld a, 0 (preserve CY)
        sbc             b
        ld              b, a            ; bc = -1 * Velocity

        ld              a, 8            ; reverse Direction
.NotNeg
        add             [hl]            ; a <- Direction

        push            bc              ; save Velocity
        push            af              ; save Direction

        add             4               ; rotate Direction by 90 degrees

        call            .GetVel         ; a <- XVel

        ld              [R0 + 1], a     ; HIGH byte of [R0] <- XVel

        pop             af              ; restore Direction
        pop             bc              ; restore Velocity

        add             8               ; rotate Direction by 180 degrees

        call            .GetVel         ; a <- YVel

        ld              hl, R0

        ld              [hl+], a        ; LOW byte of [R0]  <- YVel
        inc             l
        ld              [hl+], a
        ld              [hl], b         ; [R1] <- YVel

        ld              bc, rROMB0 + cROMB0 & $FF

        di                              ; disable interrupts to load SP

        pop             hl
        ld              sp, hl          ; return to "caller"

        reti

.GetVel

        ; look-up Sin

        and             $0F             ; keep it in 0..15
        add             a, a            ; multiply by 2 for word-table

        add             SinTab & $FF    ; *

        ld              l, a
        ld              h, SinTab >> 8

        ld              a, [hl+]
        ld              d, [hl]
        ld              e, a            ; de <- Sin (fp 12.5)

        ; multiply Sin by the Velocity

        ld              l, h            ; set Accumulator = 0
.Loop
        srl             b
        rr              c               ; CY = 1 if 1's, 2's 4's, etc.

        jr              nc, .Skip

        add             hl, de          ; add Cos / Sin
.Skip
        sla             e
        rl              d               ; de = Cos / Sin * 2, 4, 8, etc.

        ld              a, c
        or              b               ; tested all bits ?
        jr              nz, .Loop       ; no, so more "multiplying" to do

        ; convert Result to integer, round off

        add             hl, hl          ; hl = XXIIIIII IIFFFFF.
        add             hl, hl          ;    = XIIIIIII IFFFFF..
        add             hl, hl          ;    = IIIIIIII FFFFF...

        ld              c, $80
        add             hl, bc          ; round off Result

        ld              a, h            ; a = XVel or YVel

        ret

; **
; EH_1757:     Place XVel of XYVel in [R0], YVel of XYVel in [R1],
;              extending the sign bit into the HIGH byte of either
;              register.
;
; Parameters:  [R0] = XVel << 8 + YVel
;
; Modifies:    [R0] = XVel
;              [R1] = YVel
;

EH_1757::

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- R0 + 1 (OBJ_01)
        pop             de              ; de <- $0100  (OBJ_02)

        ld              a, [hl-]        ; a <- XVel
        ld              e, [hl]         ; e <- YVel

        ld              [hl+], a        ; set LOW byte of [R0] = XVel

        rlca
        and             d
        cpl
        inc             a               ; a = 0 if XVel.7 reset, -1 if set

        ld              [hl+], a        ; set HIGH byte of [R0] = 0 / -1

        ld              a, e            ; a = YVel

        ld              [hl+], a        ; set LOW byte of [R1] = YVel

        rlca
        and             d
        cpl
        inc             a               ; a = 0 if YVel.7 reset, -1 if set

        ld              [hl], a         ; set HIGH byte of [R1] = 0 / -1

        RETURN          0               ; return, enabling interrupts

        ; 28 + 4 = 32 cycles, 18 bytes

; **
; EH_1A2F:
;

EH_1A2F::

        di                              ; don't allow CopyToVRAM to interrupt

        pop             hl              ; hl <- OBJ_01 (R0)

        ld              [SP_1], sp      ; save SP in [SP_1]

        ld              sp, hl          ; sp -> R0

        pop             de              ; de <- R0 (DX)
        pop             bc              ; bc <- R1 (DY)

        ei                              ; interrupts can't be off so long (?)

        ld              a, b
        or              c               ; is DY = 0 ?
        jr              z, .QIorII      ; yes, so in Quadrant I or II

        bit             7, b            ; is DY < 0 (negative) ?
        jr              z, .QIIIorIV    ; no, so in Quadrant III or IV

        ; DY is signed, so in Quadrant I or Quadrant II

        xor             a
.QIorII
        ld              l, a
        ld              h, a            ; set DirBase = 0 (QI)

        sub             c
        ld              c, a

        ld              a, h            ; = ld a, 0 (preserve CF)
        sbc             b
        ld              b, a            ; bc = -1 * DY

        bit             7, d            ; is DX.15 = 1 (signed) ?
        jr              z, .GetDir      ; no, so in Quadrant I

        ; DX and DY are signed, so in Quadrant II

        xor             a
        sub             e
        ld              e, c
        ld              c, a

        ld              a, h            ; = ld a, 0 (preserve CY)
        sbc             d
        ld              d, b            ; de = DY
        ld              b, a            ; bc = -1 * DX

        ld              l, 4            ; set DirBase = 4 (QII)

        jr              .GetDir

.QIIIorIV

        ; DY is positive, so in Quadrant III or Quadrant IV

        bit             7, d            ; is DX.15 = 1 (signed) ?
        jr              nz, .QIII       ; yes, so in Quadrant III

        ; DY and DX are positive, so in Quadrant IV

        ld              l, c
        ld              h, b
        ld              c, e
        ld              b, d            ; bc = DX
        ld              e, l
        ld              d, h            ; de = DY

        ld              hl, 12          ; set DirBase = 12 (QIV)

        jr              .GetDir

.QIII

        ; DY is positive and DX negative, so in Quadrant III

        ld              hl, 8           ; set DirBase = 8 (QIII)

        xor             a
        sub             e
        ld              e, a

        ld              a, h            ; = ld a, 0 (preserve CF)
        sbc             d
        ld              d, a            ; de = -1 * DX

.GetDir

        push            hl              ; save DirBase

        ld              a, e            ; a = LOW byte of DX
        sub             c               ;   = LOW byte of DX - DY
        ld              a, d            ; a = HIGH byte of DX
        sbc             b               ;   = HIGH byte of DX - DY

        jr              c, .LessThan    ; if CY = 1, DX < DY

        ; if DY * 4 < DX, set DirOfs = 0

        ld              l, c
        ld              h, b            ; hl = DY

        add             hl, hl
        add             hl, hl          ; hl = DY * 4

        ld              a, l            ; a = LOW byte of DY * 4
        sub             e               ;   = LOW byte of DY * 4 - DX
        ld              a, h            ; a = HIGH byte of DY * 4
        sbc             d               ;   = HIGH byte of DY * 4 - DX

        ld              a, -1           ; set DirOfs = 0 (CY will be added)

        jr              c, .SetDir      ; if CY = 1, DY * 4 < DX

        ; if ( DX - DY ) * 4 >= DX, set DirOfs = 1

        ld              a, e            ; a = LOW byte of DX
        sub             c               ;   = LOW byte of DX - DY
        ld              l, a

        ld              a, d            ; a = HIGH byte of DX
        sbc             b               ;   = HIGH byte of DX - DY
        ld              h, a

        add             hl, hl
        add             hl, hl          ; hl = ( DX - DY ) * 4

        ld              a, l            ; a = LOW of ( DX - DY ) * 4
        sub             e               ;   = LOW of ( DX - DY ) * 4 - DX
        ld              a, h            ; a = HIGH of ( DX - DY ) * 4
        sbc             d               ;   = HIGH of ( DX - DY ) * 4 - DX

        ld              a, 1            ; set DirOfs = 1 (or 2)

        jr              .SetDir

.LessThan

        ; if ( DY - DX ) * 4 < DY, set DirOfs = 2

        ld              a, c            ; a = LOW of DY
        sub             e               ;   = LOW of DY - DX
        ld              l, a

        ld              a, b            ; a = HIGH of DY
        sbc             d               ;   = HIGH of DY - DX
        ld              h, a

        add             hl, hl
        add             hl, hl          ; hl = ( DY - DX ) * 4

        ld              a, l            ; a = LOW of ( DY - DX ) * 4
        sub             c               ;   = LOW of ( DY - DX ) * 4 - DY
        ld              a, h            ; a = HIGH of ( DY - DX ) * 4
        sbc             b               ;   = HIGH of ( DY - DX ) * 4 - DY

        ld              a, 1            ; set DirOfs = 2 (CY will be added)

        jr              c, .SetDir      ; if CY = 1, ( DY - DX ) * 4 < DY

        ; if DX * 4 >= DY, set DirOfs = 3

        ld              l, e
        ld              h, d            ; hl = DX

        add             hl, hl
        add             hl, hl          ; hl = DX * 4

        ld              a, l            ; a = LOW of DX * 4
        sub             c               ;   = LOW of DX * 4 - DY
        ld              a, h            ; a = HIGH of DX * 4
        sbc             b               ;   = HIGH of DX * 4 - DY

        ld              a, 3            ; set DirOfs = 3 (or 4)
.SetDir
        pop             de              ; restore DirBase (0, 4, 8, 12)

        adc             e               ; a = DirBase + DirOfs (PreDir)

        ; return values in R1 & R5 (assume R0 need not be preserved)

        ld              hl, R5

        ld              [hl+], a        ; set LOW byte of [R5] = PreDir
        ld              [hl], d         ; set HIGH byte of [R5] = 0

        and             $0F             ; a = ( DirBase + DirOfs ) & $0F
        ld              e, a            ; save PostDir in e

        push            de              ; [R1] <- PostDir

        di                              ; disable interrupts to load SP

        ld              sp, SP_1
        pop             hl
        ld              sp, hl          ; sp <- [SP_1]

        pop             bc              ; bc <- OBJ_02 (rROMB0 + cROMB0 & $FF)

        reti
