; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; **
; HOME_130B:
;

HOME_130B::

        ld              hl, SYSTEM_RAM + $11B * 2

        ld              l, [hl]
        sla             l               ; hl -> MoRecAddr @ MoAttr

        inc             l               ; skip LOW byte of MoAttr
        ld              d, [hl]         ; d <- HIGH byte of MoAttr

        inc             l               ; pos MoRec @ COLLTAB
        ld              a, [hl+]        ; a <- LOW byte of COLLTAB

        or              [hl]            ; is COLLTAB = 0 ?
        jr              z, .NextMO      ; yes, so go to the next MO

        inc             l               ; pos @ LOW byte of MoXPos

        ld              sp, R6          ; registers as a STACK for HOME_130B

        push            hl              ; save MoRec @ MoXPos

        ; store out OldXPos & OldYPos to [$0319] & [$031A] (respectively)

        ld              bc, SYSTEM_RAM + $119 * 2

        ld              a, [hl+]
        ld              [bc], a
        inc             c

        ld              a, [hl+]
        ld              [bc], a         ; [$0319] <- MoXPos
        inc             c

        ld              a, [hl+]
        ld              [bc], a
        inc             c

        ld              a, [hl+]
        ld              [bc], a         ; [$031A] <- MoYPos

        ; check if MO is linked to previous MO before proceeding

        ld              a, [SCRATCHPAD_RAM + $01C * 2]  ; a <- MoRotBit
        ld              c, a
        ld              a, [SCRATCHPAD_RAM + $01A * 2]  ; a <- MoLink

        and             c               ; linked with previous MO ?
        jr              z, .AddXYVel    ; no, so add X & Y velocities

        ; -------------------------------------------------------------------
        ;                Current MO is linked with previous MO
        ; -------------------------------------------------------------------

        ld              c, [hl]         ; c <- MoYVel (vertical displacement)
        inc             l
        ld              b, [hl]         ; b <- MoXVel (horiz. displacement)

        ld              a, l
        sub             ( 4 - 2 ) * 2 + 1
        ld              l, a            ; hl -> MoRec @ MoXPos

        sub             8 * 2
        ld              e, a
        ld              d, h            ; * de -> LOW byte of PrevXPos

        ld              a, [de]         ; a <- LOW byte of PrevXPos (frac)
        inc             e
        ld              [hl+], a        ; store out LOW byte of NewXPos

        ld              a, [de]         ; a <- HIGH byte of PrevXPos (int)
        inc             e

        add             b               ; NewXPos = PrevXPos + MoXVel

        cp              [hl]            ; same as OldXPos ?
        ld              [hl+], a        ; store out NewXPos (in case new)

        ld              b, a            ; b = HIGH byte of NewXPos

        ld              a, [de]         ; a <- LOW byte of PrevYPos (frac)

        jr              nz, .Diff       ; no, so ...

        ; NewXPos did not change, now check if NewYPos changes

        inc             e
        ld              [hl+], a        ; store out LOW byte of PrevYPos

        ld              a, [de]         ; a <- HIGH byte of PrevYPos (int)

        add             c               ; NewYPos = PrevYPos + MoYVel

        cp              [hl]            ; same as OldYPos ?
        jp              z, .Exit        ; yes, so position did not change

        ld              [hl-], a        ; store out NewYPos (in case new)

        dec             l               ; pos @ HIGH byte of NewXPos
        dec             l               ; pos @ LOW byte of NewXPos

        ld              c, [hl]         ; c <- LOW byte of PrevXPos

        jr              .GetOffHndlr

.Diff

        ; NewXPos did change, so position will change regardless of NewYPos

        inc             e
        ld              [hl+], a        ; store out LOW byte of PrevYPos

        ld              a, [de]         ; a <- HIGH byte of PrevYPos (int)

        add             c               ; NewYPos = PrevYPos + MoYVel

        ld              [hl-], a        ; store out NewYPos

        dec             l               ; pos @ HIGH byte of NewXPos
        dec             l               ; pos @ LOW byte of NewXPos

        ld              c, [hl]         ; c <- LOW byte of PrevXPos

        jr              .GetOffHndlr

.NextMO

        di                              ; disable interrupts to restore stack

        ld              sp, $442E       ; return to EXEC @ $1217

        reti                            ; return, enabling interrupts

.AddXYVel

        ; -------------------------------------------------------------------
        ;               Add X & Y velocities to X & Y positions
        ; -------------------------------------------------------------------

        bit             7, d            ; is MoAttr.15 = 1 ("still" set) ?
        jp              nz, .Exit       ; yes, so don't add velocities

        ld              a, [hl+]
        ld              b, [hl]         ; b <- MoXVel (HIGH byte of MoXYVel)
        ld              d, a            ; d <- MoYVel (LOW byte of MoXYVel)

        or              b               ; are both MoXVel and MoYVel = 0 ?
        jp              z, .Exit        ; yes, so Mo's pos. will not change

        ld              a, [SCRATCHPAD_RAM + $015 * 2]  ; a <- DcmlPt

        cpl
        add             6 + 1           ; $00 = 6, $02 = 4, $04 = 2
        jr              c, .Even        ; if CY = 1, shift even # of times

        add             16 - 7 + 1      ; $09 = 7, $0C = 4, $0E = 2, $10 = 0
        or              $01             ;     = 7,     = 5,     = 3,     = 1
.Even
        ld              c, a            ; save SarCount c

        xor             a               ; a = LOW byte of MoXVel (0)
        ld              e, a            ; e = LOW byte of MoYVel (0)
.SarLoop
        sra             d
        rra                             ; MoYVel = MoYVel >> 1
        sra             b
        rr              e               ; MoXVel = MoXVel >> 1

        dec             c               ; dec SarCount
        jr              nz, .SarLoop

        dec             l               ; pos MoRec @ LOW byte of MoXYVel
        dec             l               ; pos MoRec @ HIGH byte of MoYPos
        dec             l               ; pos MoRec @ LOW byte of MoYPos

        add             [hl]            ; a = LOW of MoYPos + MoYVel
        ld              [hl+], a        ; store out LOW of MoYPos + MoYVel

        ld              a, [hl]         ; a <- HIGH of MoYPos
        adc             a, d            ; a = HIGH of MoYPos + MoYVel
        ld              [hl-], a        ; store out HIGH of MoYPos + MoYVel

        dec             l               ; pos MoRec @ HIGH byte of MoXPos
        dec             l               ; pos MoRec @ LOW byte of MoXPos

        ld              a, [hl]         ; a <- LOW  of MoXPos
        add             e               ; a = LOW of MoXPos + MoXVel
        ld              [hl+], a        ; store out LOW of MoXPos + MoXVel

        ld              c, a            ; c = LOW byte of NewXPos

        ld              a, [hl]         ; a <- HIGH of MoXPos
        adc             a, b            ; a = HIGH of MoXPos + MoXVel
        ld              [hl-], a        ; store out HIGH of MoXPos + MoXVel

        ld              b, a            ; b = HIGH byte of NewXPos

.GetOffHndlr

        ; -------------------------------------------------------------------
        ;               Get OffHndlrType (0-3) and OffHndlrAddr
        ; -------------------------------------------------------------------

        dec             l               ; pos @ HIGH byte of COLLTAB

        ld              a, [hl-]
        ld              e, [hl]         ; e <- LOW byte of COLLTAB
        ld              l, a            ; l <- HIGH byte of COLLTAB

        ld              h, AddrHiToROMB0 >> 8

        ld              a, [hl]         ; a <- #/ROM bank where COLLTAB is
        ldh             [cROMB0], a     ; note new ROM bank for ISRs
        ld              [rROMB0], a     ; switch to new ROM bank

        inc             h               ; hl -> AddrHiToReal

        ld              h, [hl]         ; a <- HIGH of COLLTAB >> 1 (real)
        ld              l, e

        add             hl, hl          ; hl -> COLLTAB (real)

        ld              e, [hl]         ; e <- LOW byte of OffHndlrAddr
        inc             l
        ld              a, [hl+]        ; a <- OffHndlrType

        or              a               ; is OffHndlrType = 0 ?
        jp              z, .Exit        ; yes, so not handled here

        ld              d, [hl]         ; d <- HIGH byte of OffHndlrAddr

        dec             a               ; is OffHndlrType = 1 ?
        jp              nz, .ChkHndlr2  ; no, so check handler type 2, etc.

        ; -------------------------------------------------------------------
        ;        OffHnldrType #1 (handles "invalid" coordinate wraps)
        ; -------------------------------------------------------------------

        pop             hl              ; hl -> MoRec @ NewXPos

        ld              a, [SYSTEM_RAM + $119 * 2 + 1]  ; a <- OldXPos HIGH
        xor             b
        add             a, a            ; did NewXPos.15 change ?
        jr              nc, .NoWrapX    ; no, so no X-coord. wrap possible

        ld              a, b            ; a = HIGH byte of NewXPos
        add             a, a
        xor             b
        add             a, a            ; is NewXPos.15 = NewXPos.14 ?
        jr              c, .NoWrapX     ; no, so no X-coord. wrap possible

        ld              a, [SYSTEM_RAM + $119 * 2]
        ld              [hl+], a
        ld              a, [SYSTEM_RAM + $119 * 2 + 1]
        ld              [hl+], a        ; overwrite NewXPos w/ OldXPos

        ld              a, e
        or              d               ; is OffHndlrAddr = 0 ?
        jr              z, .WrapX       ; yes, so check for Y-coord. wrap

        inc             l               ; skip LOW byte of NewYPos
        ld              a, [hl]         ; a <- HIGH byte of NewYPos

        ld              [ehNewYPos], a  ; save HIGH of NewYPos for EXEC_1358

        ; load up registers w/ values expected to be passed to OffHndlr

        di                              ; don't allow VBlankFunc to interrupt

;        ld              sp, R5 + 2      ; update from R5..R0 using the stack

        ld              hl, $1358       ; set RetAddr = $1358 (see EXEC_1358)

        push            hl              ; [R5] <- RetAddr
        push            de              ; [R4] <- OffHndlrAddr
        push            bc              ; [R3] <- NewXPos

        ld              a, b            ; a = HIGH byte of NewXPos
        rlca                            ; put bit #7 in bit #0
        and             $01             ; a = 0 if RIGHT, 1 if LEFT
        ld              c, a            ; save it in c for write to R2

        ld              b, 0            ; set HIGH byte of R2 = 0

.DispatchX

        ld              hl, SYSTEM_RAM + $11C * 2

        ld              [hl], e
        inc             l
        ld              [hl], d         ; [$031C] <- HndlrAddr

        ld              l, ( SYSTEM_RAM + $11B * 2 ) & $FF

.DispatchY

        push            bc              ; [R2] <- 0 if RIGHT, 1 if LEFT

        ld              l, [hl]
        ld              h, $03          ; hl <- [$031B] (MoRecAddr)

        push            hl              ; [R1] <- MoRec @ MoAttr

        ld              a, [SCRATCHPAD_RAM + $01B * 2]
        ld              c, a            ; c <- [$011B] (MoNum)

        push            bc              ; [R0] <- MoNum

        ; registers are all set, so it's time to call the "OffHndlr" routine

        ld              bc, rROMB0 + cROMB0 & $FF

        ld              l, d            ; l = HIGH byte of OffHndlr
        ld              h, c            ; hl -> AddrHiToROMB0

        ld              a, [hl]         ; a <- #/ROM bank where OffHndlr is
        ld              [c], a          ; note new ROM bank for ISRs
        ld              [bc], a         ; switch to new ROM bank

        inc             h               ; hl -> AddrHiToReal
        ld              h, [hl]         ; h <- HIGH of OffHndlr >> 1 (real)

        ld              a, l            ; a = HIGH byte of OffHndlr (CP1610)
        sub             h
        ldh             [R7_Conv], a

        ld              l, e
        add             hl, hl

        ld              sp, hl          ; sp -> OffHndlr

        reti                            ; return, enabling interrupts

.NoWrapX

        ; no illegal wrap on X, so let's check for an illegal wrap on Y

        inc             l               ; skip LOW byte of MoXPos
        inc             l               ; skip HIGH byte of MoXPos
.WrapX
        inc             l               ; skip LOW byte of MoYPos

        ld              a, [hl-]        ; a <- HIGH byte of NewYPos
        add             a, a            ; is NewYPos in 128..255 ?
        jp              nc, .Exit       ; no, so no Y-coord. wrap is possible

        rra                             ; restore HIGH byte of NewYPos
        ld              b, a            ; and save it in b

        ld              c, [hl]         ; c <- LOW byte of NewYPos

        ld              a, [SYSTEM_RAM + $11A * 2]
        ld              [hl+], a
        ld              a, [SYSTEM_RAM + $11A * 2 + 1]
        ld              [hl], a         ; overwrite NewYPos w/ OldYPos

        ld              a, e
        or              d               ; is OffHndlrAddr = 0 ?
        jp              z, .Exit        ; yes, so nothing more to do

        ; load up registers w/ values expected to be passed to OffHndlr

        di                              ; don't allow VBlankFunc to interrupt

;        ld              sp, R5 + 2      ; update from R5..R0 using the stack

        ld              hl, $1212       ; set RetAddr = $1212

        push            hl              ; [R5] <- RetAddr
        push            de              ; [R4] <- OffHndlrAddr
        push            bc              ; [R3] <- NewYPos

        ld              a, b            ; a = HIGH byte of NewYPos
        rlca
        rlca                            ; put NewYPos.14 in CY

        ld              b, 0            ; set HIGH byte of R2 = 0

        ld              a, b
        adc             a, $02          ; a = 2 if BOTTOM, 3 if TOP
        ld              c, a            ; save it in c for write to R2

        ld              hl, SYSTEM_RAM + $11B * 2

        jr              .DispatchY      ; continue above to save code

.ChkHndlr2

        dec             a               ; is OffHndlrType = 2 ?
        jr              nz, .OffHndlr3  ; no, so must be handler type 3

        ; -------------------------------------------------------------------
        ;           OffHnldrType #2 (leaves physical screen region)
        ; -------------------------------------------------------------------

        pop             hl              ; hl -> MoRec @ NewXPos

        ld              a, b            ; a = HIGH byte of NewXPos
        cp              168             ; is NewXPos in 0..167 ?
        jr              c, .OnScrnX     ; yes, so NewXPos is on-screen

        ; went off screen at the LEFT or RIGHT, so restore OldXPos

        ld              a, [SYSTEM_RAM + $119 * 2]
        ld              [hl+], a
        ld              a, [SYSTEM_RAM + $119 * 2 + 1]
        ld              [hl+], a        ; overwrite NewXPos w/ OldXPos

        ld              a, e
        or              d               ; is OffHndlrAddr = 0 ?
        jr              z, .OffScrnX    ; yes, so don't dispatch to OffHndlr

        inc             l               ; skip LOW byte of NewYPos
        ld              a, [hl]         ; a <- HIGH of NewYPos

        ld              [ehNewYPos], a  ; save HIGH of NewYPos for EXEC_13A4

        ; load up registers w/ values expected to be passed to OffHndlr

        di                              ; don't allow VBlankFunc to interrupt

;        ld              sp, R5 + 2      ; update from R5..R0 using the stack

        ld              hl, $13A4       ; set RetAddr = $13A4 (see EXEC_13A4)

        push            hl              ; [R5] <- RetAddr
        push            de              ; [R4] <- OffHndlrAddr
        push            bc              ; [R3] <- NewXPos

        ; if NewXPos in 168..178 = RIGHT, if NewXPos in 179..255 = LEFT

        ld              a, b            ; a = HIGH byte of NewXPos

        ld              bc, 0           ; set HIGH byte of R2 = 0

        add             256 - 179       ; CY = 1 if NewXPos in 179..255
        rl              c               ; c = 0 if RIGHT, 1 if LEFT

        jr              .DispatchX      ; dispatch below to save code

.OnScrnX
        inc             l               ; skip LOW byte of NewXPos
        inc             l               ; skip HIGH byte of NewXPos
.OffScrnX
        inc             l               ; skip LOW byte of NewYPos

        ld              a, [hl-]        ; a <- HIGH byte of NewYPos
        cp              104             ; is NewXPos in 0..103 ?
        jp              c, .Exit        ; yes, so object is on-screen

        ; went off screen at the TOP or BOTTOM, so restore OldYPos

        ld              b, a            ; b = HIGH byte of NewYPos (integer)
        ld              c, [hl]         ; c <- LOW byte of NewYPos (fraction)

        ld              a, [SYSTEM_RAM + $11A * 2]
        ld              [hl+], a
        ld              a, [SYSTEM_RAM + $11A * 2 + 1]
        ld              [hl], a         ; overwrite NewYPos w/ OldYPos

        ld              a, e
        or              d               ; is OffHndlrAddr = 0 ?
        jp              z, .Exit        ; yes, so don't dispatch to OffHndlr

        ; load up registers w/ values expected to be passed to OffHndlr

        di                              ; don't allow VBlankFunc to interrupt

;        ld              sp, R5 + 2      ; update from R5..R0 using the stack

        ld              hl, $1212       ; set RetAddr = $1212

        push            hl              ; [R5] <- RetAddr
        push            de              ; [R4] <- OffHndlrAddr
        push            bc              ; [R3] <- NewYPos

        ; if NewYPos in 104..114 = BOTTOM, if NewYPos in 115..255 = TOP

        ld              a, b            ; a = HIGH byte of NewYPos

        ld              bc, 1           ; set HIGH byte of R2 = 0

        add             256 - 115       ; CY = 1 if if NewXPos in 115..255
        rl              c               ; c = 2 if BOTTOM, 3 if TOP

        ld              hl, SYSTEM_RAM + $11B * 2

        jp              .DispatchY      ; dispatch below to save code

.OffHndlr3

        ; -------------------------------------------------------------------
        ;           OffHnldrType #3 (leaves user-defined rectangle)
        ; -------------------------------------------------------------------

        ; make sure old coordinates lie inside in the view

        ld              hl, SYSTEM_RAM + $119 * 2 + 1

        ld              a, [SCRATCHPAD_RAM + $016 * 2]  ; a <- View_X2

        cp              [hl]            ; is View_X2 < OldXPos ?
        jr              c, .Exit        ; yes, so not previously in view

        ld              a, [SCRATCHPAD_RAM + $017 * 2]  ; a <- View_X1

        cp              [hl]            ; is View_X1 > OldXPos ?
        jr              c, .ChkYIns     ; no, so make sure OldYPos inside
        jr              nz, .Exit       ; yes, so not previously in view
.ChkYIns
        inc             l               ; pos @ LOW byte of OldYPos
        inc             l               ; pos @ HIGH byte of OldYPos

        ld              a, [SCRATCHPAD_RAM + $018 * 2]  ; a <- View_Y2

        cp              [hl]            ; is View_Y2 < OldYPos ?
        jr              c, .Exit        ; yes, so not previously in view

        ld              a, [SCRATCHPAD_RAM + $019 * 2]  ; a <- View_Y1

        cp              [hl]            ; is View_Y1 > OldYPos ?
        jr              c, .ChkXOut     ; no, so check if NewXPos outside
        jr              nz, .Exit       ; yes, so not previously in view

        ; now check to see if new coordinates lie outside the view

.ChkXOut
        pop             hl              ; hl -> MoRec @ NewXPos

        ld              a, [SCRATCHPAD_RAM + $016 * 2]  ; a <- View_X2

        cp              b               ; is View_X2 < NewXPos ?
        jr              c, .OffHoriz    ; yes, so went off view to the RIGHT

        ld              a, [SCRATCHPAD_RAM + $017 * 2]  ; a <- View_X1

        cp              b               ; is View_X1 > NewXPos ?
        jr              c, .ChkYOut     ; no, so check if NewYPos outside
        jr              nz, .OffHoriz   ; yes, so went off view to the LEFT
.ChkYOut
        inc             l               ; skip LOW byte of MoXPos
        inc             l               ; skip HIGH byte of MoXPos
        inc             l               ; skip LOW byte of MoYPos
.XOut
        ld              b, [hl]         ; b <- HIGH byte of MoYPos

        ld              a, [SCRATCHPAD_RAM + $018 * 2]  ; a <- View_Y2

        cp              b               ; is View_Y2 < NewYPos ?
        jr              c, .OffVert     ; yes, so went off view at BOTTOM

        ld              a, [SCRATCHPAD_RAM + $019 * 2]  ; a <- View_Y1

        cp              b               ; is View_Y1 > NewYPos ?
        jr              c, .Exit        ; no, so didn't go off view at TOP
        jr              z, .Exit        ; no, so didn't go off view at TOP

.OffVert

        ; went out of view at the TOP or BOTTOM, so restore OldYPos

        ld              a, [SYSTEM_RAM + $11A * 2 + 1]  ; a <- OldYPos HIGH
        ld              [hl-], a        ; store out HIGH byte of OldYPos

        ld              c, [hl]         ; c <- LOW byte of NewYPos

        ld              a, [SYSTEM_RAM + $11A * 2]      ; a <- OldYPos LOW
        ld              [hl], a         ; store out LOW byte of OldYPos

        ; load up registers w/ values expected to be passed to OffHndlr

        di                              ; don't allow VBlankFunc to interrupt

;        ld              sp, R5 + 2      ; update from R5..R0 using the stack

        ld              hl, $1212       ; set RetAddr = $1212

        push            hl              ; [R5] <- RetAddr
        push            de              ; [R4] <- OffHndlrAddr
        push            bc              ; [R3] <- NewYPos

        ld              b, 0            ; set HIGH byte of R2 = 0

        ld              a, $03
        sbc             a, b            ; a = 2 if BOTTOM, 3 if TOP
        ld              c, a            ; save it in c for write to R2

        ld              a, e
        or              d               ; is OffHndlrAddr = 0 ?
        jr              z, .Exit        ; yes, so all that setup for nothing

        ld              hl, SYSTEM_RAM + $11B * 2

        jp              .DispatchY      ; dispatch below to save code

.OffHoriz

        ; went out of view at the LEFT or RIGHT, so restore OldXPos

        ld              a, [SYSTEM_RAM + $119 * 2]
        ld              [hl+], a
        ld              a, [SYSTEM_RAM + $119 * 2 + 1]
        ld              [hl+], a        ; overwrite NewXPos w/ OldXPos

        inc             l               ; skip LOW byte of MoYPos
        ld              a, [hl]         ; a <- HIGH byte of MoYPos

        ld              [ehNewYPos], a  ; save HIGH of NewYPos for EXEC_138B

        ; load up registers w/ values expected to be passed to OffHndlr

        di                              ; don't allow VBlankFunc to interrupt

;        ld              sp, R5 + 2      ; update from R5..R0 using the stack
;        ld              hl, $138B       ; set RetAddr = $138B (see EXEC_138B)
;        push            hl              ; [R5] <- RetAddr

        ld              sp, $138B       ; set RetAddr = $138B (see EXEC_138B)
        ld              [R5], sp        ; [R5] <- RetAddr

        ld              sp, R4 + 2      ; update from R4..R0 using the stack

        push            de              ; [R4] <- OffHndlrAddr
        push            bc              ; [R3] <- NewXPos

        ld              b, 0            ; set HIGH byte of R2 = 0

        ld              a, $01
        sbc             a, b            ; a = 0 if RIGHT, 1 if LEFT
        ld              c, a            ; save it in c for write to R2

        ld              a, e
        or              d               ; is OffHndlrAddr = 0 ?
        jr              z, .XOut        ; yes, so check TOP and BOTTOM too!

        jp              .DispatchX      ; no, so dispatch above to save code

.Exit
        ld              bc, rROMB0 + cROMB0 & $FF

        ld              a, BANK(EXEC)   ; a = #/bank where EXEC is
        ld              [c], a          ; note new ROM bank for ISRs
        ld              [bc], a         ; switch to new ROM bank

        di                              ; * disable interrupts to restore stack

        ld              sp, $4424       ; return to EXEC @ $1212

        reti                            ; return, enabling interrupts

; **
; HOME_17D6:  Decrement active counters in SCRATCHPAD_RAM at [$0125-6..]
;             When they reach 0, call associated routine in PLIST.
;

HOME_17D6::

        ld              sp, DS_STACK + 6

        ld              a, [UDB_Flag]
        dec             a               ; is UDB_Flag = 1 ?
        call            z, UDB_Update   ; yes, so need to call UDB_Update

        ld              a, [ehProcCount]
        or              a               ; is ProcCount = 0 ?
        jr              z, DS_Exit      ; yes, so nothing to do

        ld              hl, SCRATCHPAD_RAM + $025 * 2

        ld              b, a            ; b = ProcCount

DS_Loop:

        ld              a, [hl+]
        inc             l
        ld              d, [hl]         ; d <- HIGH byte of Ctr
        ld              e, a            ; e <- LOW byte of Ctr

        or              d               ; is Ctr = 0 ?
        jr              z, DS_PN2       ; yes, so process is frozen

        bit             7, d            ; is Ctr.15 = 1 (negative) ?
        jr              nz, DS_PN2      ; yes, so process is frozen

        dec             de              ; dec Ctr

        ld              a, d
        ld              [hl-], a        ; store out HIGH byte of Ctr
        dec             l
        ld              [hl], e         ; store out LOW byte of Ctr

        or              e               ; is Ctr now = 0 ?
        jr              nz, DS_PN1      ; no, so move on to next Ctr

        ; Ctr hit 0

        ld              c, l            ; c = LOW byte of CtrAddr

        push            bc              ; save ProcCount / CtrAddr

        ld              b, h            ; bc -> Ctr

        ld              de, ehProcList - ( SCRATCHPAD_RAM + $025 * 2 )

        add             hl, de          ; hl -> ehProcList @ ProcAddr

        ld              a, [hl+]
        ld              e, a
        ld              a, [hl+]
        ld              d, a            ; de <- ProcAddr

        ; refresh Ctr

        ld              a, [hl+]        ; a <- LOW byte of CtrRefresh
        ld              [bc], a         ; store out LOW byte of CtrRefresh
        inc             c
        inc             c               ; pos @ LOW byte of Ctr

        ld              a, [hl+]        ; a <- HIGH byte of CtrRefresh
        ld              [bc], a         ; store out HIGH byte of CtrRefresh

        ; dispatch to Proc

        ld              bc, rROMB0 + cROMB0 & $FF

        ld              l, d
        ld              h, c            ; hl -> AddrHiToROMB0

        ld              a, [hl]         ; a <- #/ROM bank where Proc is
        ld              [c], a          ; note ROM bank # for ISRs
        ld              [bc], a         ; switch to new ROM bank

        inc             h               ; hl -> AddrHiToReal
        ld              h, [hl]         ; h <- HIGH of ProcAddr >> 1 (Real)

        ld              a, l            ; l = HIGH byte of ProcAddr
        sub             h
        ldh             [R7_Conv], a

        ld              l, e            ; l = LOW byte of ProcAddr
        add             hl, hl          ; hl = ProcAddr (Real)

        di                              ; disable interrupts to load SP

        ld              sp, $17FA
        ld              [R5], sp        ; [R5] <- $17FA (RetAddr)

        ld              sp, hl          ; sp = ProcAddr (Real)

        reti                            ; return, enabling interrupts

HOME_17FA::                             ; return from Proc dispatch

        ld              sp, DS_STACK + 4

        pop             bc              ; rest. ProcCount / CtrAddr

        ld              l, c
        ld              h, SCRATCHPAD_RAM >> 8

DS_PN1: inc             l
        inc             l
DS_PN2: inc             l
        inc             l               ; pos @ next Ctr

        dec             b               ; dec ProcCount
        jr              nz, DS_Loop

        ld              bc, rROMB0 + cROMB0 & $FF

DS_Exit:

        di                              ; disable interrupts to load SP

        ld              sp, $503A       ; return to EXEC @ $181D

        reti                            ; return, enabling interrupts

GH_Dispatch_1:

        ; re-initialize the screen saver so it won't time out

        ld              l, $1E * 2      ; hl -> HIGH byte of ScrnSaver

        ld              [hl], $19       ; set HIGH byte of ScrnSaver = $19

GH_Dispatch_2:

        ; a  = ControlCode
        ; de = RetAddr
        ; b  = HandTabOfs
        ; c  = 0 if LEFT, 1 if RIGHT

        ; load up registers w/ values expected to be passed to HandHndlr

        ld              hl, R5          ; hl -> R5

        ld              [hl], e
        inc             l
        ld              [hl], d         ; [R5] <- PARAM_1 (RetAddr)

        ld              l, R0 & $FF     ; hl -> R0

        ld              d, l            ; zero d in 1 cycle (used below)

        ld              [hl+], a

        rlca
        and             l
        cpl
        inc             a               ; a = 0 or -1

        ld              [hl+], a        ; [R0] <- PARAM_0 (ControlCode)

        ld              a, c
        rrca
        ld              [hl+], a
        ld              [hl], d         ; [R1] <- 0 if LEFT, 1 if RIGHT

        ; get HANDTAB from SYSTEM_RAM and add the appropriate offset

        ld              hl, SYSTEM_RAM + $015D * 2  ; hl -> HANDTAB

        ld              a, [hl+]        ; a <- LOW byte of HANDTAB
        add             b
        ld              e, a            ; e = LOW byte of HANDTAB + offset

        ld              a, [hl]         ; a <- HIGH byte of HANDTAB
        adc             a, d
        ld              l, a            ; l = LOW byte of HANDTAB + offset

        ; switch to ROM bank where HANDTAB + offset is, and get HandHndlr

        ld              bc, rROMB0 + cROMB0 & $FF

        ld              h, c            ; hl -> AddrHiToROMB0 >> 8

        ld              a, [hl]         ; a <- ROM bank # where HANDTAB is
        ld              [c], a          ; note new ROM bank for ISRs
        ld              [bc], a         ; switch to new ROM bank

        inc             h
        ld              h, [hl]         ; h <- HIGH of HANDTAB (Real) >> 1

        ld              l, e
        add             hl, hl          ; hl -> HANDTAB (Real)

        ld              a, [hl+]
        inc             hl
        ld              l, [hl]         ; l <- HIGH byte of HandHndlrAddr
        ld              e, a            ; e <- LOW byte of HandHndlrAddr

        or              l               ; is HandHndlrAddr = 0 ?
        jr              z, .Abort       ; yes, so all that set-up for nothing

        ; now switch to ROM bank where _HandHndlr_ is, and dispatch to it

        ld              h, c            ; hl -> AddrHiToROMB0

        ld              a, [hl]         ; a <- ROM bank # where HandHndlr is
        ld              [c], a          ; note new ROM bank for ISRs
        ld              [bc], a         ; switch to new ROM bank

        inc             h
        ld              h, [hl]         ; h <- HIGH of HandHndlr (Real) >> 1

        ld              a, l            ; a = HIGH of HandHndlr (CP1610)
        sub             h
        ldh             [R7_Conv], a

        ld              l, e
        add             hl, hl          ; hl -> HandHndlr (Real)

        di                              ; disable interrupts to restore stack

        ld              sp, hl          ; sp -> HandHndlr (Real)

        reti                            ; return, enabling interrupts

.Abort

        ; HandHndlrAddr is = 0, so nothing to dispatch to

        ld              a, BANK(EXEC)   ; a <- ROM bank # where EXEC is
        ld              [c], a          ; note new ROM bank for ISRs
        ld              [bc], a         ; switch to new ROM bank

        ret
