; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; ---------------------------------------------------------------------------
;                     .::** IMAGIC HELPER ROUTINES **::.
; ---------------------------------------------------------------------------

; **
; IH_50BB:        Dispatch to routine specified by RomRec #0-#1 of
;                 MO that corresponds to the bit which is set in
;                 ProcBitMap.
;
; Notes:          Also known as "ONESHOT" in DRACULA.SRC.
;
; Parameters:     * OS_PB = ProcBitMap      * OS_RR = RomRec
;                 * OS_RA = RetAddr         * OS_C1 = Convert #1
;                 * OS_C2 = Convert #2      * OS_CA = ContAddr
;
;                 * Must be set in advance.
;
; Modifies:       af, de, hl
;

rIH_50BB::

        ld              sp, OS_STACK + 4  ; use OS_STACK for IH_50BB

rOS_PB: ld              hl, SCRATCHPAD_RAM + $017 * 2

        ld              a, [hl]         ; a <- [$0117] (ProcBitMap)
        or              a               ; is ProcBitMap = 0 ?
        jr              z, rOS_Exit     ; yes, so no reason to check ea. bit

        ld              [hl], 0         ; [$0117] (ProcBitMap) <- 0

        ld              e, -1           ; examine all 8 bits
        scf                             ; set ProcBitMap.7 after 1st rotate

rOS_Loop:

        inc             e               ; inc ProcMoNum
        rra                             ; is ProcBitMap.X = 1 ?
        jr              nc, rOS_Loop    ; no, so keep looping until it is

        bit             3, e            ; is ProcMoNum = 8 ?
        jr              nz, rOS_Exit    ; yes, so checked all 8 MOs

        ; ProcBitMap.X = 1, so get MO's RomRecAddr from RomRecList

        ld              d, a            ; d = ProcBitMap >> X

        push            de              ; [OS_STACK + 2] <- e (ProcMoNum)
                                        ; [OS_STACK + 3] <- d (ProcBitMap)

        ld              d, 0            ; zero d for saving below

rOS_RR: ld              a, $3D          ; a = ( $03D )
        add             e               ;   = ( $03D + MO # )
        add             a, a            ;   = ( $03D + MO # ) * 2

        ld              sp, R5 + 2      ; sp -> R6 (1st push writes to R5)

rOS_RA: ld              hl, $50D7
        push            hl              ; [R5] <- RetAddr (CP1610)

        ld              l, a
        ld              h, ( SYSTEM_RAM + $0100 * 2 ) >> 8 ; hl -> RomRecList
                                                           ;       [ MO # ]

        ld              a, [hl+]        ; a <- LOW of RomRecAddr (CP1610)
        add             2
        ld              h, [hl]         ; h <- HIGH of RomRecAddr (CP1610)

        jr              nc, .NoCarry

        inc             h
.NoCarry
        ld              l, a

        ; got RomRecAddr, now get RomRec #0-#1 (DispAddr)

        push            hl              ; [R4] <- RomRecAddr + 2 (CP1610)
        push            de              ; [R3] <- MO #

rOS_C1: ld              de, $4000 >> 1 - $5000 - 1
        add             hl, de

        jr              nc, rIH_50D7    ; is CY = 0, assume RomRecAddr = 0

        add             hl, hl          ; hl = RomRecAddr + 1 (Real)

        ld              a, [hl-]
        dec             l
        ld              l, [hl]         ; al <- RomRec #0-#1 (DispAddr)

rOS_C2: add             ( $4000 >> 1 - $5000 ) >> 8
        ld              h, a

        add             hl, hl          ; hl = DispAddr (Real)

        di                              ; disable interrupts to load SP

        ld              sp, hl

        reti

rOS_Exit:

        di                              ; disable interrupts to load SP

rOS_CA: ld              sp, $41BC       ; return to DRACULA @ $50DE

        reti

rIH_50BB_End::

rIH_50D7::

        ld              sp, OS_STACK + 2

        pop             de              ; e <- [EH_STACK + 2] (ProcCount)
                                        ; d <- [EH_STACK + 3] (ProcBitMap)

        ld              a, d            ; a = ProcBitMap >> X

        jr              rOS_Loop

rIH_50D7_End::

; **
; IH_50E0:        Checks for collisions.
;
; Notes:          Also known as "COLLCHK" in DRACULA.SRC.
;
; Parameters:     * CC_XP = XPos            * CC_CD = CollInfo Dist.
;                 * CC_C1 = Convert #1        CC_IM = Info & Mask
;                 * CC_CI = CollInfo        * CC_CA = ContAddr
;                 * CC_RA = RetAddr         * CC_C2 = Convert #2
;
;                 * Must be set in advance.
;
; Modifies:       af, bc, de, hl
;

rIH_50E0::

        ld              sp, CC_STACK + 4 * 2

        ld              c, 0            ; set MoNum = 0

rCC_OutLoop:

        ld              a, c            ; a = MO #
        add             a, a            ;   = MO # * 2

rCC_XP: add             ( SYSTEM_RAM + $0125 * 2 + 1 ) & $FF

        ld              l, a
        ld              h, ( SYSTEM_RAM + $0100 * 2 ) >> 8 ; hl -> XP[ MO # ]

        bit             0, [hl]         ; is XP.8 (Coll) = 1 ?
        jr              z, rCC_MoNext   ; no, so check the next MO

        add             24 * 2 - 1
        ld              l, a            ; hl -> RomRecList[ MO # ]

        ld              e, [hl]
        inc             l
        ld              d, [hl]         ; de <- RomRecAddr

        ; MO's Collision bit is set, and it's RomRecAddr <> 0

rCC_CD: add             24 * 2          ; pos @ CollInfoList
        ld              l, a

        ld              a, [hl+]
        ld              b, [hl]         ; ba <- CollInfo

rCC_C1: ld              hl, $4000 >> 1 - $5000 + 5
        add             hl, de

        jr              nc, rCC_MoNext  ; if CY = 0, assume RomRecAddr = 0

        add             hl, hl          ; hl = RomRecAddr + 5 (Real)

        ld              e, [hl]         ; e = LOW of CollMask
        and             e
        ld              d, a            ; d = LOW of CollMask & CollInfo

        inc             l               ; pos @ HIGH of CollMask

        ld              a, [hl+]        ; a = HIGH of CollMask
        ld              [CC_IM + 1], a  ; store it out to "Info & Mask"
        and             b               ; a = HIGH of CollMask & CollInfo
        ld              [CC_IM + 2], a  ; store it out to "Info & Mask"

        or              d               ; is CollInfo now = 0 ?
        jr              z, rCC_MoNext   ; yes, so no collisions to process

        ; CollInfo is not 0, so find which bits are set & handle them

        ld              b, 10           ; CollBitCtr = 10

rCC_InLoop:

        srl             d               ; is CollInfo.X = 1 ?
        jr              c, rCC_Disp     ; yes, so dispatch to handler

        srl             e               ; is CollMask.X = 1 ?
        jr              nc, rCC_Cont    ; no, so no CollEntry for X

        inc             l
        inc             hl
        inc             l
        inc             hl              ; pos @ next CollEntry
rCC_Cont:
        ld              a, b
        cp              3               ; is CollBitCtr = 3 ?
        jr              nz, rCC_Skip    ; no, so leave CollInfo / CollMask

rCC_IM: ld              de, 0           ; de = CollInfo / CollMask (HIGH)
rCC_Skip:
        dec             b               ; is CollBitCtr now = 0 ?
        jr              nz, rCC_InLoop  ; no, so keep checking bits

rCC_MoNext:

        inc             c               ; inc MO #

        bit             3, c            ; is MO # = 8 ?
        jr              z, rCC_OutLoop  ; no, so check the next MO

rCC_CI: ld              hl, SYSTEM_RAM + $0155 * 2

        xor             a
.ZeroLoop
        ld              [hl+], a
        ld              [hl+], a        ; [CollInfo] <- 0

        dec             c
        jr              nz, .ZeroLoop

        ld              bc, rROMB0 + cROMB0 & $FF

        di                              ; disable interrupts to load SP

rCC_CA: ld              sp, $427E       ; return to DRACULA @ $513F

        reti

rCC_Disp:

        ; CollInfo.X is set, and MO #X was flagged in CollMask

        srl             e               ; CollMask = CollMask >> 1

        push            de              ; save CollInfo / CollMask
        push            bc              ; save CollBitCtr / MoNum

        ld              a, [hl+]
        ld              e, a
        inc             hl

        ld              a, [hl+]
        ld              d, a            ; de <- CollHndlrAddr
        inc             hl

        push            hl              ; save CollEntryAddr (CP1610)

        ld              a, 10
        sub             b               ; a = 10 - CollBitCtr

        ld              sp, R2 + 2      ; sp -> R3 (next push writes to R2)

        ld              b, 0
        push            bc              ; [R2] <- MoNum
        ld              c, a
        push            bc              ; [R2] <- CollBitNum

        di                              ; disable interrupts to load SP

        ld              bc, rROMB0 + cROMB0 & $FF

rCC_RA: ld              sp, $5128

        ld              [R5], sp        ; [R5] <- RetAddr

rCC_C2: ld              hl, $4000 >> 1 - $5000
        add             hl, de

        add             hl, hl
        ld              sp, hl          ; sp = CollHndlrAddr (Real)

        reti

rIH_50E0_End::

rIH_5128::

        ld              sp, CC_STACK + 2

        pop             hl              ; rest. CollEntryAddr (CP1610)
        pop             bc              ; rest. CollBitCtr / MoNum
        pop             de              ; rest. CollInfo / CollMask

        jr              rCC_Cont        ; *

rIH_5128_End::

; **
; IH_5357::
;
; Notes:          Also known as "UPDSTIC" in DRACULA.SRC.
;
; Parameters:     * US_XP = XPos            * US_SB = SticBase
;                 * US_CA = ContAddr        * US_SC = SticComp
;
;                 * Must be set in advance.
;
; Modifies:       af, bc, de, hl
;

rIH_5357::

        ld              sp, SP_1        ; use STACK_1 for IH_5357

rUS_XP: ld              bc, SYSTEM_RAM + $0125 * 2 - 1

        ld              a, c            ; a = LOW byte STIC address * 2 (S)
rUS_SB: sub             $25 * 2 - 1
        ld              l, a            ; l = LOW byte STIC address * 2 (H)

        ld              h, STIC >> 8    ; hl -> STIC[ a * 2 ]

        call            US_UpdSTIC      ; update STIC register

        ld              a, c
rUS_SC: cp              ( $25 + 24 ) * 2 - 1  ; updated 24 registers ?
        jr              nz, rUS_SB            ; no, so keep looping

rUS_Exit:

        ld              bc, rROMB0 + cROMB0 & $FF

        di                              ; disable interrupts to load SP

rUS_CA: ld              sp, $43C4       ; return to DRACULA @ $51E2

        reti

rUS_Same::

        ld              a, l
        cp              24 * 2 - 1      ; updated 24 registers ?
        jr              z, rUS_Exit     ; yes, so get out of loop

        inc             l               ; pos @ LOW of next STIC reg. (H)

rUS_UpdSTIC::

        inc             c               ; pos @ LOW of next STIC reg. (S)

        ld              a, [bc]         ; a = LOW byte of "new" reg. value
        inc             c
        cp              [hl]            ; is it = LOW of old reg. value ?
        ld              [hl+], a        ; store out LOW of "new" reg. value
        ld              a, [bc]         ; a = HIGH byte of "new" reg. value
        jr              nz, .Diff       ; yes, so process new reg. value

        cp              [hl]            ; is it = HIGH of old reg. value ?
        jr              z, rUS_Same     ; yes, so process new reg. value
.Diff
        ld              [hl-], a        ; store out HIGH of "new" reg. value

        ld              d, a            ; d = HIGH byte of STIC reg. value
        ld              e, [hl]         ; e = LOW byte of STIC reg. value

        inc             h               ; = ld h, stRegToHandler >> 8

        ld              a, [hl+]
        ld              h, [hl]
        ld              l, a

        jp              [hl]            ; jump to handler for this register

rIH_5357_End::

; **
; IH_51E2:        Copy the STIC's collision registers (located @ $0018-
;                 $001F) to the "software copy", or CollInfo.  Unlike
;                 the original IMAGIC routine, the STIC collision register
;                 is not OR'd w/ the old CollInfo, because we are
;                 guaranteed in this case that collisions are handled
;                 each interation of the loop (see COLLCHK, above).
;
; Notes:          Also known as "GETCOLL" in DRACULA.SRC.
;
; Parameters:     * GC_CI = CollInfo        * GC_CA = ContAddr
;
;                 * Must be set in advance.
;
; Modifies:       af, bc, de, hl
;
; Costs:          157 cycles, 35 bytes (vs. 216 cycles, 42 bytes).
;

rIH_51E2::

rGC_CI: ld              hl, SYSTEM_RAM + $155 * 2  ; hl -> CollInfo
        ld              de, 8                      ; copy 8 registers

        di                              ; protect Attr. reg. of MO #7

        ld              sp, STIC + $018 * 2        ; sp -> STIC's CollRegs
.Loop
        pop             bc              ; bc <- STIC's CollReg

        ld              a, c
        ld              [hl+], a        ; store out LOW byte of CollReg
        ld              a, b
        ld              [hl+], a        ; store out HIGH byte of CollReg

        dec             e
        jr              nz, .Loop

        ; interrupts off for 106 cycles, so allow pending interrups to occur

        ei                              ; small window for pending interrupts

        ld              bc, rROMB0 + cROMB0 & $FF

        di                              ; protect Attr. reg. of MO #7

        push            de
        push            de
        push            de
        push            de
        push            de
        push            de
        push            de
        push            de              ; zero all 8 CollRegs (in 32 cycles)

        ; ^ 150 cycles (total)

rGC_CA: ld              sp, $43EC       ; return to DRACULA @ $51F6

        reti

rIH_51E2_End::

;
; Notes:          Also known as "CYCANIM" in DRACULA.SRC.
;
; Parameters:     * CA_RR = RomRecList      * CA_C1 = Convert #1
;                 * CA_RS = RomRecToSQPC    * CA_SR = SQPCToRomRec
;                 * CA_EA = EndAddr         * CA_CA = ContAddr
;
;                 * Must be set in advance.
;
; Modifies:       af, bc, de, hl
;
; Costs:          88 bytes.
;

rIH_514A::

        ld              sp, SP_1        ; use STACK_1 for IH_514A

rCA_RR: ld              hl, SYSTEM_RAM + $13D * 2  ; hl -> RomRecList
rCA_C1: ld              b, ( $4000 >> 1 - $5000 ) >> 8

rCA_Loop:

        ; get RomRecAddr, add RomRecOfs, and get NumPics

        ld              e, l
        ld              d, h            ; de -> RomRecAddr

        ld              a, [hl+]
        ld              h, [hl]         ; h <- HIGH of RomRecAddr (CP1610)
        ld              l, a            ; l <- LOW of RomRecAddr (CP1610)

        or              h               ; is RomRecAddr = 0 ?

        ld              c, 4            ; * outside loop ?
        add             hl, bc
        add             hl, hl          ; hl -> RomRec @ NumPics (Real)

        ld              c, [hl]         ; c <- NumPics

rCA_RS: ld              hl, ( SYSTEM_RAM + $145 * 2 ) - ( SYSTEM_RAM + $13D * 2 ) + 1
        add             hl, de          ; hl -> SQPC

        jr              z, rCA_SR       ; yes, so don't sequence animation

        ; ^ 23 cycles

        ; IMAGIC "SQPC"                                    nnnn  = PicNum
        ;                                                  Rrrrr = DlyRefresh
        ; 15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00  ddddd = Delay
        ;  m  ?  d  d  d  d  d  R  r  r  r  r  n  n  n  n  m     = Minus

        ld              a, [hl-]
        ld              e, [hl]         ; e <- rrrrnnnn (LOW byte of SQPC)
        ld              d, a            ; d <- m?dddddR (HIGH byte of SQPC)

        and             $3E             ; isolate Delay in bits 1-5
        jr              z, rCA_Zero     ; if ZY = 1, Delay was already = 0

        sub             $02             ; dec Delay
        jr              nz, rCA_NotZero ; if ZF = 0, Delay was in 2..31

        ; advance PicNum, make sure it's in 0..NumPics - 1, and update it

        ld              a, e            ; a = rrrrnnnn (LOW byte of SQPC)
        and             $0F             ;   = ....nnnn

        inc             a               ; inc PicNum

        bit             7, d            ; is Minus = 1 ?
        jr              z, .Cont1       ; no, so leave the PicNum inc'd

        sub             2               ; dec PicNum
        jr              nc, .Cont1      ; if CY = 0, PicNum was > 0

        ld              a, c            ; a = NumPics
        dec             a               ; conv NumPics to MaxPic
.Cont1
        cp              c               ; is PicNum < NumPics ?
        jr              c, .Cont2       ; yes, so don't wrap to zero

        xor             a               ; set PicNum = 0
.Cont2
        xor             e
        and             $0F             ; preserve PicNum
        xor             e               ; set PicNum

        ld              [hl], a         ; * store out LOW byte of SQPC

rCA_Zero:                               ; Delay was already = 0

        ; set Delay = DelayRefresh, and update Delay

        ld              a, d            ; a = m?dddddR (HIGH byte of SQPC)
        rrca                            ;   = Rm?ddddd (CY = R)

        ld              a, e            ; a = rrrrnnnn (LOW byte of SQPC)
        rra                             ;   = Rrrrrnnn
        rrca                            ;   = nRrrrrnn
        rrca                            ;   = nnRrrrrn

        ; ^ 6 cycles

rCA_NotZero:                            ; Delay was in 2..31

        xor             d
        and             $3E             ; preserve DelayRefresh
        xor             d               ; set Delay = DelayRefresh

        inc             l               ; * re-pos @ HIGH byte of SQPC

        ld              [hl], a         ; store out HIGH byte of SQPC

rCA_SR: ld              de, ( SYSTEM_RAM + $13D * 2 ) - ( SYSTEM_RAM + $145 * 2 ) + 1
        add             hl, de          ; hl -> RomRecList

        ld              a, l
rCA_EA: cp              ( SYSTEM_RAM + $145 * 2 ) & $FF
        jr              nz, rCA_Loop

        ld              bc, rROMB0 + cROMB0 & $FF

        di

rCA_CA: ld              sp, $42AA       ; return to DRACULA @ $5155

        reti

rIH_514A_End::

