; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; ---------------------------------------------------------------------------
;               .::** Initialization routines and tables **::.
; ---------------------------------------------------------------------------

GB_Reset:

        ld              [GB_Type], a    ; save the GB type for later

        cp              $11                 ; is it a GameBoy Color ?
        call            z, CPU_SetDblSpeed  ; yes, so set the CPU in 2x mode
        di                                  ; don't let anything interrupt

	; wait until we are in VBlank so we can reliably disable the display

.VBL    ldh		a, [rLY]	; a <- LCDC Y-Coordinate (0..153)
	cp		144		; is it >= 144 (in V-Blank) ?
	jr		c, .VBL         ; no, so keep looping until it is

        xor		a
	ldh		[rLCDC], a	; disable display (VRAM accessable)

        ld              a, STATF_LYC    ; specify interrupt when LYCEQULY
        ldh             [rSTAT], a      ; [$FF41] (LCD Status) <- a

        ld              a, IEF_VBLANK | IEF_LCDC | IEF_TIMER
        ldh             [rIE], a        ; set VBL, LCDC, & TIMER interrupts

        xor             a
        ldh             [rTIMA], a      ; set Timer Counter = 0
        ld              a, $80
        ldh             [rTMA], a       ; set Timer Modulo = 128

        xor             a
        ldh             [rTAC], a       ; make sure Timer stopped (for now)

        ; -------------------------------------------------------------------
        ; Copy GROM ' '..'~' to Upper Tile Bank (for writing debugging info)
        ; -------------------------------------------------------------------

        ld              a, 1
        ldh             [rVBK], a       ; select bank 1 / upper-tile bank

        ld              hl, _VRAM
        ld              de, GROM_8
        ld              bc, 95 * 8 + $0100
.Copy_GROM
        ld              a, [de]         ; a <- byte from GROM_8
        inc             de
        ld              [hl+], a
        ld              [hl+], a        ; store it out as color 3

        dec             c
        jr              nz, .Copy_GROM
        dec             b
        jr              nz, .Copy_GROM

        xor             a
        ldh             [rVBK], a       ; select bank 0 / lower-tile bank

        ld              hl, _VRAM
        ld              bc, 8 * 8 * 16
.InitVRAM
        ld              [hl+], a
        dec             c
        jr              nz, .InitVRAM
        dec             b
        jr              nz, .InitVRAM

        ; -------------------------------------------------------------------
        ; Initialize JoyPad-related variables
        ; -------------------------------------------------------------------

        xor             a
        ld              [JoyDown], a      ; assume no buttons are down
        dec             a
        ld              [JoyUp], a        ; assume all buttons are up

        ; -------------------------------------------------------------------
        ; Clear HIGH bytes of SCRATCHPAD_RAM (so they return 0 when read)
        ; -------------------------------------------------------------------

        xor             a
        ld              hl, SCRATCHPAD_RAM
        ld              bc, 256 * 2

        call            MemFill         ; clear SCRATCHPAD_RAM

;        ld              hl, SCRATCHPAD_RAM + 1
;        xor             a
;        ld              c, a            ; do 256 locations
;.InitScratch
;        ld              [hl+], a
;        inc             l               ; skip LOW byte of SCRATCHPAD_RAM
;        dec             c
;        jr              nz, .InitScratch

        ; -------------------------------------------------------------------
        ; Copy random bitmaps from GRAM_16 to GRAM_8 & zero the HIGH bytes
        ; -------------------------------------------------------------------

        ld              a, 1
        ldh             [rSVBK], a      ; switch to WRAM bank #1 (GRAM_16)

        ld              hl, GRAM_16
        ld              de, GRAM_8
        xor             a
        ld              bc, 64 * 8      ; do 512 locations
.InitGRAM
        ld              a, [hl+]        ; a <- LOW byte from GRAM_16
        ld              [de], a         ; store it out to GRAM_8
        inc             de

        ld              [hl], 0         ; clear HIGH byte of GRAM_16
        inc             hl

        dec             c
        jr              nz, .InitGRAM
        dec             b
        jr              nz, .InitGRAM

        ; -------------------------------------------------------------------
        ; Copy various routines from ROM to Low RAM and High RAM
        ; -------------------------------------------------------------------

        ld              hl, VBlankFixUp_ROM
        ld              de, VBlankFixUp
        ld              bc, VBlankFixUp_ROM_End - VBlankFixUp_ROM
        call            MemCopy

        ld              hl, HBlankFunc_ROM
        ld              de, HBlankFunc
        ld              bc, HBlankFunc_ROM_End - HBlankFunc_ROM
        call            MemCopy

        ld              hl, CopyToOAM_ROM
        ld              de, CopyToOAM
        ld              bc, CopyToOAM_ROM_End - CopyToOAM_ROM
        call            MemCopy         ; copy CopyToOAM to High RAM

        ; -------------------------------------------------------------------
        ; Initialize AddrHiToReal and AddrHiToROMB0 using MappingData
        ; -------------------------------------------------------------------

        ld              hl, MappingData
        ld              de, AddrHiToReal
.InitMapping
        ld              a, [hl+]         ; a <- RealAddr >> 1 for "page"
        ld              [de], a          ; store it out to AddrHiToReal
        dec             d                ; pos @ AddrHitoROMB0
        ld              a, [hl+]         ; a <- ROM bank # for this "page"
        ld              [de], a          ; store it out to AddrHiToROMB0
        inc             d                ; re-pos @ AddrHitoReal
        inc             e
        jr              nz, .InitMapping

        ; -------------------------------------------------------------------
        ; Init variables in High RAM
        ; -------------------------------------------------------------------

        ld              hl, ColorToRGB_ROM
        ld              de, ColorToRGB
        ld              bc, 16 * 2      ; RGB data of 16 colors (2 bytes ea.)
        call            MemCopy         ; copy ColorToRGB_ROM to High RAM

        xor             a
        ldh             [STIC_Mode], a        ; init to ColorStack mode
        ldh             [BC_Active], a        ; disable vertical blockout
        ldh             [STIC_Display_Hi], a  ; disable display

        ;

        ld              [STIC_Analyzed], a    ; necessary ??
        ld              [STIC_Mode_New], a
;        ld              [PageReady], a

        ; -------------------------------------------------------------------
        ; Init everything having to do w/ proper display of the border
        ; -------------------------------------------------------------------

        xor             a                     ; color #0
        ld              bc, 8 * 2             ; 8 lines @ 2 bytes ea.
        ld              hl, _VRAM + 193 * 16  ; hl -> tile #193's bitmap data

        call            MemFill

        dec             c               ; = ld c, $FF

        ld              hl, _VRAM + 195 * 16  ; hl -> tile #195's bitmap data

        REPT            8

        ld              [hl], c         ; store out LSB of color #0 ($FF)
        inc             l
        ld              [hl+], a        ; store out MSB of color #0 ($00)

        ENDR

        ld              hl, _VRAM + 197 * 16  ; hl -> tile #260's bitmap data

        REPT            8

        ld              [hl+], a        ; store out LSB of color #0 ($00)
        ld              [hl], c         ; store out MSB of color #0 ($FF)
        inc             l

        ENDR

        ld              a, c
        ld              c, 8 * 2              ; 8 lines @ 2 bytes ea.
        ld              hl, _VRAM + 199 * 16  ; hl -> tile #262's bitmap data

        call            MemFill

        ; write a series of $F9s for tile #256's data (set it to color 0)

;        xor             a                     ; solid block of color 0
;        ld              bc, 8 * 2             ; 8 lines at 2 bytes each
;        ld              hl, _VRAM + 256 * 16  ; hl -> tile #256's bitmap data
;        call            MemFill

        ; set the palette index for all tiles to #0 (where border color is)

        ld              a, 1            ; a = 1
        ldh             [rVBK], a       ; select bank 1 / tile attributes

        xor             a               ; set palette index to 0
        ld              bc, 32 * 32 * 2 ; do 2,096 locations (set both maps)
        ld              hl, _SCRN0      ; set tiles on both _SCRN0 and _SCRN1
        call            MemFill

        ;

        ld              hl, _SCRN1 + ( 32 * 5 ) + 15

        ld              a, $08
        ld              [hl+], a
        ld              [hl+], a
        ld              [hl+], a
        ld              [hl+], a

        ld              hl, CurFrames

        xor             a
        ld              [hl+], a        ; set CurFrames = 0
        ld              [hl], a         ; set TotFrames = 0

        ; load both tile maps w/ #66 (border color locations)

        xor             a
        ld              [rVBK], a       ; select bank 0 / tile indices

        ld              bc, 32 * 32 * 1 ; do 2,096 locations (set both maps)
        ld              hl, _SCRN0      ; set tiles on both _SCRN0 and _SCRN1
        call            MemFill

        ld              a, 66           ; fill w/ tile #66
        ld              bc, 32 * 32 * 1 ; do 2,096 locations (set both maps)
        call            MemFill

        ;

        ld              hl, _SCRN1 + ( 32 * 5 ) + 15

        xor             a
        ld              [hl+], a
        ld              [hl+], a
        ld              [hl+], a
        ld              [hl+], a

        ld              hl, WINPAL_0 + 3 * 2

        ld              a, $9F
        ld              [hl+], a
        ld              [hl], $0F

        ld              hl, WINPAL_0 + 7 * 2

        ld              a, $4A
        ld              [hl+], a
        ld              [hl], $29

        ; set the border color to 0 initially so we don't have a white bkg

        ld              hl, rBCPS       ; hl -> Bkg Color Palette Spec Reg

        ld              a, $80
        ld              [hl+], a        ; auto-advance, palette #0, color #0
        xor             a
        ld              [hl], a
        ld              [hl], a         ; set RGB value = 0 (black)

        ; -------------------------------------------------------------------
        ; Init everything having to do w/ vertical card blockout
        ; -------------------------------------------------------------------

        ; write out two solid blocks of color #2 for the "column" sprites

        xor             a                      ; a = 0
        ld              bc, $FF << 8 + 8 * 2   ; do 2 tiles at 8 rows ea.
        ld              hl, _VRAM + 64 * 16    ; hl -> tile #64's bitmap data
.InitColumn_Tile
        ld              [hl+], a        ; 1st line <- all zeros
        ld              [hl], b         ; 2nd line <- all ones (color 2)
        inc             l
        dec             c
        jr              nz, .InitColumn_Tile

        ld              bc, 16
        call            MemFill         ; tile #66 <- border tile (color 0)

        ; initialize the OAM for the six "column" sprites

        ld              a, 5
        ldh             [rSVBK], a      ; switch to WRAM bank #5

        ld              hl, LocalOAM    ; position at sprite #0's OAM
        ld              bc, 64 << 8 + 6 ; tile #64, do 6 sprites
        ld              e, BkgOrgX + 8
        xor             a
.InitColumn_OAM
        ld              [hl+], a        ; Y-Coordinate = 0
        ld              [hl], e         ; X-Coordinate = 8
        inc             l
        ld              [hl], b         ; Tile Index = 64
        inc             l
        ld              [hl+], a        ; Attribute = 0
        dec             c
        jr              nz, .InitColumn_OAM

        ; let's also initialize the OAM for sprites 6 & 7 (not used yet)

        ld              [hl+], a        ; Y-Coordinate = 0
        inc             l               ; skip X-Coordinate
        inc             l               ; skip TileIndex
        inc             l               ; skip Attribute

        ld              [hl+], a        ; Y-Coordinate = 0

        ; -------------------------------------------------------------------
        ; Init everything having to do w/ propery display of the window
        ; -------------------------------------------------------------------

        ; position window in the upper left hand corner of the screen

        xor             a
        ld              [rWY], a        ; Window Y Position = 0
        ld              a, 7
        ld              [rWX], a        ; Window X Position = 7

        xor             a
        ldh             [BrdrColRGB], a
        ldh             [BrdrColRGB + 1], a   ; set BorderColor = 0 (Black)
        ld              [WINPAL_0], a
        ld              [WINPAL_0 + 1], a

        ; -------------------------------------------------------------------
        ; Setup everything having to do with Moving Ojects
        ; -------------------------------------------------------------------

        ld              a, 5
        ldh             [rSVBK], a      ; switch to WRAM bank #5

        ld              hl, LocalOAM + 8 * 4  ; pos @ sprite #8's OAM
        ld              bc, 8 * 4             ; do 32 sprites (4 for ea. MO)
        xor             a
.MO_Loop
        ld              [hl+], a        ; store out Y Coord.
        ld              [hl+], a        ; store out X Coord.
        ld              [hl], b         ; store out Tile Index
        inc             l
        ld              [hl+], a        ; store out Attributes
        inc             b
        inc             b               ; inc Tile Index by 2
        dec             c
        jr              nz, .MO_Loop

        call            InitMO          ; init MO record structure

        ; -------------------------------------------------------------------
        ; Set up MO_CsCols and MO_BgCols arrays from MO_ColorAssignments
        ; -------------------------------------------------------------------

        ld              hl, MO_FgCols
        ld              a, 1
        ld              c, 8
.MO_InitFgColLoop
        ld              [hl+], a
        dec             c
        jr              nz, .MO_InitFgColLoop

        ld              hl, MO_BgCols
        ld              bc, MO_CsCols
        ld              de, MO_ColorAssignments

.MO_InitColLoop

        ld              a, [de]         ; a <- color assignment for this MO
        inc             de

        cpl
        cp              $04             ; is it a ColorStack index in 0..3 ?
        jr              nc, .MO_BgCol_1 ; no, so just store out a -1..-16

        add             ColorStack & $FF  ; a = $CD + CS index # (STIC reg.)
.MO_BgCol_1
        ld              [bc], a         ; store out CsCol byte
        inc             c

        cpl
        cp              $10             ; is it a fixed BG color in 0..15 ?
        jr              c, .MO_BgCol_2  ; yes, so store it out as is

        xor             a               ; use black until CS index # defined
.MO_BgCol_2
        add             OBJPAL >> 7
        ld              [hl+], a        ; store out BgCol byte

        bit             3, c
        jr              z, .MO_InitColLoop

        ; -------------------------------------------------------------------
        ; Fill SYSTEM_RAM with $FFs so 1st writes will be recoginized as new
        ; -------------------------------------------------------------------

;        ld              a, $FF
;        ld              bc, 384 * 2
;        ld              hl, SYSTEM_RAM
;        call            MemFill         ; * let's get rid of this !!!

        ; -------------------------------------------------------------------
        ; Init HBlankFunc related variables
        ; -------------------------------------------------------------------

        ld              a, BkgOrgY - 1  ; interrupt 1 row before window = OFF
        ld              [LY_WinOff], a
        ldh             [rLYC], a       ; LY Compare <- LY_WinOff

        ld              hl, hbfRoutine
        ld              a, HBlankFunc & $FF
        ld              [hl+], a
        ld              [hl], HBlankFunc >> 8  ; hbfRoutine <- HF_WinOff

        ; -------------------------------------------------------------------
        ; Init Scroll X and Scroll Y registers for the background
        ; -------------------------------------------------------------------

        ld              a, -1 * BkgOrgY
        ldh             [rSCY], a
        ld              [rSCY_New], a

        ld              a, -1 * BkgOrgX
        ldh             [rSCX], a
        ld              [rSCX_New], a

        ld              a, 1
        ldh             [rSVBK], a      ; switch back to WRAM bank #1
        ld              [UDB_Flag], a   ; set UDB_Flag = 1

        ld              a, $FF
        ld              bc, 384 * 2
        ld              hl, SYSTEM_RAM
        call            MemFill         ; * let's get rid of this !!!

        ld              hl, stRegToHandler_ROM
        ld              de, stRegToHandler
        ld              bc, 256
        call            MemCopy

        ld              hl, BitCntToMask_ROM
        ld              de, BitCntToMask
        ld              bc, 16 * 2
        call            MemCopy

;        ld              hl, EXEC_167D_ROM
;        ld              de, EXEC_167D
;        ld              bc, EXEC_167D_ROM_END - EXEC_167D_ROM
;        call            MemCopy

        ; IMAGIC HELPER

        ld              hl, rIH_50BB
        ld              de, IH_50BB
        ld              bc, rIH_50D7_End - rIH_50BB
        call            MemCopy         ; copy both rIH_50BB & rIH_50D7

        ld              hl, rIH_50E0
        ld              de, IH_50E0
        ld              bc, rIH_50E0_End - rIH_50E0
        call            MemCopy         ; COLLCHK

        ld              hl, rIH_5128
        ld              de, IH_5128
        ld              bc, rIH_5128_End - rIH_5128
        call            MemCopy

        ld              hl, rIH_5357
        ld              de, IH_5357
        ld              bc, rIH_5357_End - rIH_5357
        call            MemCopy         ; UPDSTIC

        ld              hl, rIH_51E2
        ld              de, IH_51E2
        ld              bc, rIH_51E2_End - rIH_51E2
        call            MemCopy         ; GETCOLL

        ld              hl, rIH_514A
        ld              de, IH_514A
        ld              bc, rIH_514A_End - rIH_514A
        call            MemCopy         ; OBJCYC

        ld              hl, rIH_515D
        ld              de, IH_515D
        ld              bc, rIH_515D_End - rIH_515D
        call            MemCopy         ; BACKCYC

        ld              hl, rIH_52B3
        ld              de, IH_52B3
        ld              bc, rIH_52B3_End - rIH_52B3
        call            MemCopy         ; ADDVEL

        ld              hl, GcToIc_ROM
        ld              de, GcToIc
        ld              bc, gcNumConstants * 2
        call            MemCopy

        xor             a
        ld              hl, HandDelay
        ld              bc, icNumConstants * 2
        call            MemFill         ; clear HandDelay

;        dec             a
;        ld              hl, WINPAL_0
;        ld              bc, 2 * 4 * 2
;        call            MemFill         ; clear WINPAL_0 *

        ld              [ehKeyClicks], a  ; * see IMAGIC games

        ret                             ; return, but don't enable interrupts

; **
; CopyToOAM_ROM:  Transfer the LocalOAM at $D200 (in WRAM bank #3) to the
;                 OAM at $FE00, at the same time updating object palettes
;                 #2-#7 from the MO_Colors table.
;
; Parameters:     a  = HIGH byte of LocalOAM
;                 c  = LOW byte of rOCPS
;                 b  = $08
;                 hl = rOCPD
;
; Notes:          Copied to High RAM during startup, and called
;                 by VBlankFunc.
;
; Costs:          168 cycles, 30 bytes.
;

CopyToOAM_ROM:

        ldh             [rDMA], a       ; [$FF46] <- DMA Xfer & Start Addr

        ; NOTE: the DMA transfer should take about 160 cycles, during this
        ; time, only memory in the range $FF80-$FFFE (HiRAM) is accessible.

.UpdObjPal_Loop

        ld              [c], a          ; pos @ color #1 of current palette

        ; write color entry #1

        pop             de              ; de <- RGB color value of entry #1

        ld              [hl], e         ; store LOW byte of RGB color value
        ld              [hl], d         ; store HIGH byte of RGB color value

        ; write color entry #2

        pop             de              ; de <- RGB color value of entry #2

        ld              [hl], e         ; store LOW byte of RGB color value
        ld              [hl], d         ; store HIGH byte of RGB color value

        ; write color entry #3

        pop             de              ; de <- RGB color value of entry #3

        ld              [hl], e         ; store LOW byte of RGB color value
        ld              [hl], d         ; store HIGH byte of RGB color value

        add             b               ; advance to next palette
        jr              nc, .UpdObjPal_Loop

        ; ^ 27 cycles per iteration, 161 cycles to update last 6 palettes

        jp              CopyToOAM_Ret   ; return to finish job in VBlankFunc

CopyToOAM_ROM_End:

; **
; InitMO:      Set up the 8 moving object records by setting both attribute
;              bytes to $21 (bit #5 must always be set), "clearing" the data
;              address, and writing the correct VRAM addresses of each MO.
;
; Parameters:  None.
;
; Notes:       In order for STIC_CollInit to function properly, the
;              collision bit must be set, the X coordinate valid for
;              collisions (1-167), and the attributes YM, XM, YR, XS,
;              YS, and YF must be set to something other than 1, 0, 0,
;              0, 1, 0 (respectively).  InitMO must also be run before
;              STIC_CollInit.
;

InitMO:

         ld             hl, MO
         ld             bc, _VRAM + $21
         ld             de, 128 << 8 + GROM_8 >> 8
         xor            a
.Loop
         ld             [hl], c         ; init MO attribute byte #0
         inc            l

         ld             [hl], c         ; set X-Coord. to 33 initially
         inc            l
         inc            l               ; skip Y-Coord.

         ld             [hl], c         ; init MO attribute byte #1
         inc            l

         ld             [hl], GROM_8 & $FF
         inc            l
         ld             [hl], e         ; set DataAddr = GROM #0
         inc            l

         ld             [hl+], a        ; store out LOW byte of VRAM address
         ld             [hl], b         ; store out HIGH byte of VRAM address
         inc            l

         add            d               ; advance to VRAM address of next MO
         jr             nc, .Loop

         inc            b               ; increment HIGH byte of VRAM address

         bit            2, b            ; reached a VRAM address of $8400 ?
         jr             z, .Loop        ; no, so keep looping

         ; zeros need to be stored to the following locations for
         ; STIC_Analyze to work (they're used as "end-of-structure" markers)

         ld             [hl+], a        ; zero MO[ MO # * 8 ]

         inc            l               ; \
         inc            l               ; / unused locations

         ld             [hl], a         ; zero MO[ MO # * 8 + 3 ]

         ret

; **
; VBlankFixUp_ROM:
;
; Costs:            30 bytes.
;

VBlankFixUp_ROM:

        ldh             a, [cROMB0]     ; a <- cROMB0 (current ROM bank #)
        ld              [rROMB0], a     ; restore rROMB0

        ld              sp, vbfRegs

        pop             hl
        pop             de
        pop             bc
        pop             af              ; restore registers from vbfRegs

rVF_M2: ld              sp, 0           ; sp <- SP at interrupt - 2

        reti                            ; return, enabling interrupts

VBlankFixUp_ROM_End:

; **
; HBlankFunc_ROM:  Turn the window ON and OFF at the appropriate times.
;
; Notes:           MODE 2: ~38-41   cycles (OAM)
;                  MODE 3: ~84-87   cycles (OAM/VRAM)
;                  MODE 0: ~100-103 cycles (HBLANK)
;
;                  No more than ~110-116 cycles of "interrupt lag" would be
;                  acceptable in order for HBlankFunc to execute properly.
;

HBlankFunc_ROM::

rHF_WinOff::

        push            af              ; hbfRoutine <- af

        ; if the STIC's display is NOT enabled, let's leave the window ON

        ldh             a, [STIC_Display_Hi]
        or              a               ; is the STIC's display ON ?
        jr              z, rHF_DO       ; no, so leave the window ON

        push            de              ; hbfDE      <- de
        push            hl              ; hbfHL      <- hl

        ; ^ 18 cycles

        ; wait until in H-Blank so window doesn't go OFF until the NEXT line

.HBL    ldh             a, [rSTAT]      ; a <- LCDC Status
        and             $03             ; are we in the H-Blank period ?
        jr              nz, .HBL        ; no, so let's wait until we are

        ; ^ 7-12 cycles in HBLANK

        ; restore palettes #0 & #1 which will be used by the background

        ld              hl, rBCPS       ; hl -> Bkg Color Palette Spec Reg
        ld              sp, WINPAL_1    ; sp -> WINPAL_1 @ palette #0

        ld              a, $80
        ld              [hl+], a        ; auto-advance, palette #0, color #0

        REPT            2 * 4           ; do 2 palettes @ 4 colors ea.

        pop             de              ; de <- RGB color value

        ld              [hl], e
        ld              [hl], d         ; store out RGB color value

        ENDR

        ; ^ 66 cycles

;        ldh             a, [rSTAT]         ; a <- LCDC Status
;        and             $03                ; are we still in H-Blank period ?
;        jp              nz, InvalidCode_1  ; no, so report error !!!

        ; turn window OFF & set LY Compare to where we'll turn it back ON

        ld              a, LCDC_WINOFF
        ldh             [rLCDC], a      ; turn the window OFF

        ; ^ 5 cycles

        ld              a, LY_WinOn     ; a = row at which window goes ON
        ldh             [rLYC], a       ; store it out to LY Compare register

        ; put HF_WinOn in hbfRoutine so it gets called next LCDC interrupt

        ld              sp, hbfHL

        pop             hl              ; hl <- hbfHL
        pop             de              ; de <- hbfDE
        pop             af              ; af <- hbfRoutine

        call            HF_M2           ; hbfRoutine <- HF_WinOn

rHF_WinOn::

        push            af              ; hbfRoutine <- af
        push            de              ; hbfDE      <- de
        push            hl              ; hbfHL      <- hl

        ; wait until in H-Blank so window doesn't go ON until the NEXT line

.HBL    ldh             a, [rSTAT]      ; a <- LCDC Status
        and             $03             ; are we in the H-Blank period ?
        jr              nz, .HBL        ; no, so let's wait until we are

;        xor             a
;        ldh             [rSCX], a       ; set Scroll X = 0 (avoid distortion)

        ; set palettes #0 & #1 which will now be used by the window

        ld              hl, rBCPS       ; hl -> Bkg Color Palette Spec Reg
        ld              sp, WINPAL_0    ; sp -> WINPAL_0 @ palette #0

        ld              a, $80
        ld              [hl+], a        ; auto-advance, palette #0, color #0

;        REPT            2 * 4           ; do 2 palettes @ 4 colors ea.

        pop             de              ; de <- RGB color value

        ld              [hl], e
        ld              [hl], d         ; store out RGB color value

;        ENDR

        ; turn window ON & set LY Compare to where we'll turn it back OFF

        ld              a, LCDC_WINON
        ldh             [rLCDC], a      ; turn the window ON

        ld              a, [LY_WinOff]  ; a = row at which window goes OFF
        ldh             [rLYC], a       ; store it out to LY Compare register

        ; put HF_WinOff in hbfRoutine so it gets called next LCDC interrupt

        ld              sp, hbfHL       ; *

        pop             hl              ; hl <- hbfHL
        pop             de              ; de <- hbfDE
rHF_DO: pop             af              ; af <- hbfRoutine

        ld              [hbfRoutine], sp  ; hbfRoutine <- hbfWinOff

rHF_M2: ld              sp, 0           ; sp = SP at interrupt - 2

        reti                            ; return, enabling interrupts

HBlankFunc_ROM_End::

MappingData:

  INCLUDE "MAPPING.ASM"




