; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

NOCASH              EQU     1

GAME_NAME           EQUS    "IGB\\ASTRSMSH"

GAME_SPLITS         EQU     0

mtMattel            EQU     0
mtIntv              EQU     1
mtImagic            EQU     2
mtPacMan            EQU     3
mtCentipede         EQU     4
mtOther             EQU     5

MAPPING             EQU     mtMattel

ctDefault           EQU     0
ctOther             EQU     1

CONTROLS            EQU     ctDefault

ptDefault           EQU     0
ptBeauty            EQU     1
ptOther             EQU     2

PATCHES             EQU     ptOther

; ---------------------------------------------------------------------------
;                   .::** WRAM Bank 2 declarations **::.
; ---------------------------------------------------------------------------

MakeUp_GROM     EQU   $D000        ; $D000-$DFFF, fixed here

; ---------------------------------------------------------------------------
;                   .::** WRAM Bank 3 declarations **::.
; ---------------------------------------------------------------------------

PALTAB_0        EQU   $D0C0        ; $D0C0-$D0FF  RGB,     must be at $XXC0

SingleInUseBy   EQU   $D100        ; $D100-$D13E  stored in even addresses
SingleIndices   EQU   $D100        ; $D101-$D13F  stored in odd addresses

PALTAB_DMY0     EQU   $D181        ; $D181        \
PALTAB_DMY1     EQU   $D183        ; $D183         \
PALTAB_DMY2     EQU   $D185        ; $D185         / see Bkg_Init
PALTAB_DMY3     EQU   $D187        ; $D187        /

PALTAB_1        EQU   $D1C0        ; $D1C0-$D1FF  InUseBy, must be at $XXC0

PairInUseBy     EQU   $D200        ; $D200-$D2FF  aligned on multiples of 256
PairIndices     EQU   $D300        ; $D300-$D3FF  aligned on multiples of 256

BDD_TileNdx     EQU   $D400        ; $D400-$D57F  aligned on 256 multiples
BDD_TileAttr    EQU   $D600        ; $D600-$D77F  aligned on 256 multiples

BktGetCols_Stop EQU   $D7FA        ; $D7FA, must be initialized to $80

CARD            EQU   $D800        ; $D800-$DF7F, fixed here

BktGetCols_Flag EQU   $DF82        ; $DF82, bits 0-2 must always be reset

; ---------------------------------------------------------------------------
;                   .::** WRAM Bank 4 declarations **::.
; ---------------------------------------------------------------------------

OBJGRAM         EQU   $D000        ; $D000-$D7F7, aligned on multiples of 256

; ---------------------------------------------------------------------------
;                   .::** WRAM Bank 5 declarations **::.
; ---------------------------------------------------------------------------

LocalOAM        EQU   $D200        ; $D200-$D29F, fixed here

ML_BOTTOM       EQU   $D2FE
ML_TOP          EQU   $D2FF
MOVELIST        EQU   $D300        ; $D300-$D3FF, aligned on multiples of 256

; ---------------------------------------------------------------------------
;                   .::** WRAM Bank 6 declarations **::.
; ---------------------------------------------------------------------------

MakeUp_GRAM     EQU   $D000        ; $D000-$D3FF, fixed here
MakeUp_DMY      EQU   $D400        ; $D400-$D4FF, aligned on multiples of 256

CL_TOP          EQU   $D5FE
CL_BOTTOM       EQU   $D5FF
COPYLIST        EQU   $D600        ; $D600-$D7FF, aligned on multiples of 512

TILE            EQU   $D800        ; $D800-$DFFF, fixed here

; ---------------------------------------------------------------------------
;                         .::** EQUATES **::.
; ---------------------------------------------------------------------------

INCLUDE	            "HARDWARE.INC"
INCLUDE             "MACROS.ASM"

; macro parameters

mpROMB0             EQU     $001   ; if rROMB0 is to be restored, see RETURN
mpCmp               EQU     $002   ; if result is thrown away, see SUB_RD_F
mpSub               EQU     $004   ; if result is stored, see SUB_RD_F
mpDI                EQU     $008   ; if macro should disable interrupts
mpR0inH             EQU     $010   ; if H contains R0 >> 8
mpJAinAE            EQU     $020   ; if Jump Adress is in AE (see JUMP)
mpJAinLA            EQU     $040   ; if Jump Adress is in LA (see JUMP)
mpSDBD              EQU     $080   ; if SDBD mode set
mpSDBD_R13          EQU     $180   ; if HIGH byte = LOW byte (see PEEK)

; compile-time options

R13_SDBD            EQU     0      ; 0 = Code that handles SDBD for
                                   ;     IMPLIED instructions using
                                   ;     R1-R3 is NOT included.
                                   ; 1 = Code is included (slower).

SYNC_DISPLAY        EQU     0      ; 0 = Process is allowed to execute
                                   ;     even if the screen has not yet
                                   ;     been updated by VBlankFunc (fast).
                                   ; 1 = Process does not execute until
                                   ;     VBlankFunc updates the screen
                                   ;     (slow, but smooth scrolling)

NEED_PRIORITIES     EQU     1      ; 0 = Background palette is packed more
                                   ;     tightly, without concern for object
                                   ;     priorities -- better for colorful
                                   ;     screens.
                                   ; 1 = Background palette is managed so
                                   ;     that prioritized objects display
                                   ;     correctly.

CLICK_REDUCTION     EQU     1      ; 0 = rAUDxENV is set to 0 when volume = 0
                                   ;     (may produce more clicks)
                                   ; 1 = Channel's output enable bit in
                                   ;     rAUDTERM is reset when volume = 0
                                   ;     (may produce less clicks)

COLOR_TUNE          EQU     0      ; 0 = Color tuning disabled.
                                   ; 1 = Color tuning enabled via the
                                   ;     D-Button, Button A, and Button B
                                   ;     (color # & RGB values at bottom)

GRAM_MODIFIED       EQU     1      ; 0 = GRAM_Flags are always examined
                                   ;     (faster only if many GRAM writes)
                                   ; 1 = GRAM_Flags are examined only when
                                   ;     _new_ values are written to GRAM
                                   ;     (faster only if few GRAM writes)

COUNT_FRAMES        EQU     0      ; 0 = Frame counting disabled.
                                   ; 1 = Frame counting enabled (shows
                                   ;     number of frames "rendered" per
                                   ;     second in terms of percentage)

; structure & record sizes

CARD_RECSIZE        EQU     8
CARD_SIZE           EQU     240 * CARD_RECSIZE

PALTAB_COLSIZE      EQU     2
PALTAB_PALSIZE      EQU     4 * PALTAB_COLSIZE
PALTAB_SIZE         EQU     8 * PALTAB_PALSIZE

MO_RECSIZE          EQU     8
MO_SIZE             EQU     8 * MO_RECSIZE + 4

COLLMO_RECSIZE      EQU     21
COLLMO_SIZE         EQU     8 * COLLMO_RECSIZE + 1

; other

BkgOrgY             EQU     3 * 8       ; NOTE: this value cannot be 0 or 1
BkgOrgX             EQU     0
LY_WinOn            EQU     BkgOrgY + 96 - 1

LCDC_WINON          EQU     %11110100   ; LCDCF_ON        1
                                        ; LCDCF_WIN9C00   1
                                        ; LCDCF_WINON     1
                                        ; LCDCF_BG8000    1
                                        ; LCDCF_BG9800    0
                                        ; LCDCF_OBJ16     1
                                        ; LCDCF_OBJOFF    0
                                        ; LCDCF_BGOFF     0 *

LCDC_WINOFF         EQU     %11000111   ; LCDCF_ON        1
                                        ; LCDCF_WIN9C00   1
                                        ; LCDCF_WINOFF    0
                                        ; LCDCF_BG8800    0
                                        ; LCDCF_BG9800    0
                                        ; LCDCF_OBJ16     1
                                        ; LCDCF_OBJON     1
                                        ; LCDCF_BGON      1

; gameboy controller constants

gcPadRight          EQU     0
gcPadUpRight        EQU     1
gcPadUp             EQU     2
gcPadUpLeft         EQU     3
gcPadLeft           EQU     4
gcPadDownLeft       EQU     5
gcPadDown           EQU     6
gcPadDownRight      EQU     7

gcButtonA           EQU     8
gcButtonB           EQU     9
gcSelect            EQU     10
gcStart             EQU     11

gcNumConstants      EQU     12

; intellivision controller constants

icD0                EQU     0
icD1                EQU     1
icD2                EQU     2
icD3                EQU     3
icD4                EQU     4
icD5                EQU     5
icD6                EQU     6
icD7                EQU     7
icD8                EQU     8
icD9                EQU     9
icD10               EQU     10
icD11               EQU     11
icD12               EQU     12
icD13               EQU     13
icD14               EQU     14
icD15               EQU     15

icK0                EQU     16
icK1                EQU     17
icK2                EQU     18
icK3                EQU     19
icK4                EQU     20
icK5                EQU     21
icK6                EQU     22
icK7                EQU     23
icK8                EQU     24
icK9                EQU     25
icKC                EQU     26
icKE                EQU     27

icA0                EQU     28
icA1                EQU     29
icA2                EQU     30

icNumConstants      EQU     31

; ---------------------------------------------------------------------------
;                    .::** Start of HOME section **::.
; ---------------------------------------------------------------------------

SECTION "SineTable",HOME[$0000]

Zeros:                          ; $0000-$000F, fixed here (see BktWrite_FGBGGROM0)

        DB      0, 0, 0, 0, 0, 0, 0, 0
        DB      0, 0, 0, 0, 0, 0, 0, 0 ; *

DbToGc:                         ; $0010-$001F, no special alignment

        DB      0               ; %0000 (not used)
        DB      gcPadRight      ; %0001
        DB      gcPadLeft       ; %0010
        DB      0               ; %0011 (not used)
        DB      gcPadUp         ; %0100
        DB      gcPadUpRight    ; %0101
        DB      gcPadUpLeft     ; %0110
        DB      0               ; %0111 (not used)
        DB      gcPadDown       ; %1000
        DB      gcPadDownRight  ; %1001
        DB      gcPadDownLeft   ; %1010
        DB      0               ; %1011 (not used)
        DB      0               ; %1100 (not used)
        DB      0               ; %1101 (not used)
        DB      0               ; %1110 (not used)
        DB      0               ; %1111 (not used)

SinTab:                         ; $0020-$003F, fixed here (see EH_1629)

        DW      0, 24, 45, 59, 64, 59, 45, 24
        DW      0, -24, -45, -59, -64, -59, -45, -24

SECTION	"VBlank_IRQ_Jump",HOME[$0040]

        jp              VBlankFunc

SECTION	"LCDC_IRQ_Jump",HOME[$0048]

        ld              [HF_M2 + 1], sp ; save SP at interrupt - 2

        ld              sp, hbfRoutine
        ret                             ; PC <- [hbfRoutine]

SECTION	"Timer_Overflow_IRQ_Jump",HOME[$0050]

        jp              TimerFunc

SECTION	"Serial_IRQ_Jump",HOME[$0058]

	reti

SECTION	"Joypad_IRQ_Jump",HOME[$0060]

	reti

;        INCLUDE         "RTPAT.ASM"

SECTION "AddrHiToPoke_STIC",HOME[$00D3 - $0004]

ifMVO_STIC_1::                          ; MoVe Out to STIC

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; e <- LOW byte of OBJ_01 (RS)
                                        ; d <- HIGH byte of OBJ_01 (BB)

        ld              a, d            ; a = LOW byte of STIC address
        ld              d, c            ; de -> RS

        ; ^ 6 cycles, 4 bytes

AP_STIC::                               ; STIC @ $0000-$003F

;        di                              ; don't allow VBlankFunc to interrupt

        add             a, a

;AP_STIC_1::                             ; entry point used by IMGC_4AFC

        ld              l, a            ; l = LOW byte STIC address * 2

        ld              h, STIC >> 8    ; hl -> STIC[ a * 2 ]

        ld              a, [de]         ; a = LOW byte of "new" reg. value
        inc             e
        cp              [hl]            ; is it = LOW of old reg. value ?
        ld              [hl+], a        ; store out LOW of "new" reg. value
        ld              a, [de]         ; a = HIGH byte of "new" reg. value
        jr              nz, .Diff       ; yes, so process new reg. value

        ; ^ 16/15 cycles, 11 bytes

        cp              [hl]            ; is it = HIGH of old reg. value ?
        jr              z, .Same        ; yes, so no reason to process value

        ; ^ 5/4 cycles, 3 bytes

.Diff
        ld              [hl-], a        ; store out HIGH of new reg. value

        ld              d, a            ; d = HIGH byte of STIC reg. value
        ld              e, [hl]         ; e = LOW byte of STIC reg. value

        inc             h               ; = ld h, stRegToHandler >> 8

        ld              a, [hl+]
        ld              h, [hl]
        ld              l, a

        jp              [hl]            ; jump to handler for this register

        ; ^ 13 cycles, 9 bytes

.Same
        ld              a, l            ; a = #/STIC register to poke to

        cp              $20 << 1 + 1    ; is it DisplayEnable ($20) ?
        jr              z, .Diff        ; yes, so prepare to enable display

        ; ^ 6/5 cycles, 5 bytes

        cp              $21 << 1 + 1    ; was it GraphicsMode ($21) ?
        jr              z, .Diff        ; yes, so set FG/BG mode

        ; ^ 5/4 cycles, 4 bytes

        reti                            ; return, enabling interrupts

        ; ^ 4 cycles, 1 byte

SECTION	"GameBoy_Header_Start",HOME[$0100]

        jp              Start
        nop

DB      $CE, $ED, $66, $66, $CC, $0D, $00, $0B
DB      $03, $73, $00, $83, $00, $0C, $00, $0D
DB      $00, $08, $11, $1F, $88, $89, $00, $0E
DB      $DC, $CC, $6E, $E6, $DD, $DD, $D9, $99
DB      $BB, $BB, $67, $63, $6E, $0E, $EC, $CC
DB      $DD, $DC, $99, $9F, $BB, $B9, $33, $3E

DB      "INTV. FOR GBC", $0, $0         ; name (upper case), remaining = $0
DB      $80                             ; $80 = Color GB, else $00 or other
DB      $00, $00			; Licensee code (new), HI & LO nybbles
DB      $00				; GB/SGB Indicator (00=GB, 03=SGB)
DB      $19                             ; Cartridge type (19=ROM+MCB5)
DB      $03				; ROM Size (256 kB, 16 banks)
DB      $00				; RAM Size (None)
DB      $01                             ; Destination code (0=JPN, 1=NON-JPN)
DB      $00				; Licensee code (old)
DB      $01				; Mask ROM Version number = 1
DB      $FF				; Complement check (Important)
DB      $FF, $FF			; Cheksum, needs to be calculated!

SECTION	"Game_Code_Start",HOME[$0150]

; ---------------------------------------------------------------------------
; Data in ROM that needs to be aligned on special multiples should go here
; ---------------------------------------------------------------------------

MO_ColorAssignments:        ; @ $0150-$0157, no special alignment required

IF      GAME_SPLITS > 0
  INCBIN   "MOBKGCOL.BIN"
ENDC
  ; if value is in 0..15,  background color is fixed for duration of game
  ; if value is in -1..-4, background is set to color stack element 1..4

;  DB       13               ; background color for MO #0
;  DB       13               ; background color for MO #1
;  DB       13               ; background color for MO #2
;  DB       13               ; background color for MO #3
;  DB       13               ; background color for MO #4
;  DB       13               ; background color for MO #5
;  DB       13               ; background color for MO #6
;  DB       13               ; background color for MO #7

; Intellivision RGB color values, copied into high RAM at start-up

ColorToRGB_ROM:             ; @ $0158-$0177, no special alignment required

  RGBSet   0,  0,  0        ; 0  = Black
  RGBSet   2,  12, 62       ; 1  = Blue
  RGBSet   48, 0,  0        ; 2  = Red
  RGBSet   40, 62, 28       ; 3  = Tan
  RGBSet   2,  32, 0        ; 4  = Dark Green
  RGBSet   0,  44, 0        ; 5  = Green
  RGBSet   63, 56, 6        ; 6  = Yellow
  RGBSet   63, 63, 63       ; 7  = White
  RGBSet   34, 34, 34       ; 8  = Gray
  RGBSet   18, 38, 50       ; 9  = Cyan
  RGBSet   63, 35, 3        ; 10 = Orange
  RGBSet   13, 24, 10       ; 11 = Brown
  RGBSet   63, 21, 21       ; 12 = Pink
  RGBSet   34, 32, 62       ; 13 = Light Blue
  RGBSet   8,  62, 2        ; 14 = Yellow-Green
  RGBSet   34, 4,  20       ; 15 = Purple

SECTION "AddrHiToPoke_ScratchPad",HOME[$01D3]

SECTION_01D3_Start::

AP_ScratchPad::                         ; SCRATCHPAD_RAM @ $0100-$01FF

        add             a, a            ; is address >= $180 ?
        ld              l, a            ; l = LOW byte of address << 1

        jr              c, .High        ; yes, so in SCRATCHPAD_RAM "high"

        ; ^ 5/4 cycles, 4 bytes

        ld              h, SCRATCHPAD_RAM >> 8  ; hl -> SCRATCHPAD_RAM
                                                ;       [ a * 2 ]

        ld              a, [de]         ; a <- LOW byte of [RS]
        ld              [hl], a         ; store it out to SCRATCHPAD_RAM

        reti                            ; return, enabling interrupts

        ; ^ 10 cycles, 5 bytes

.High

        ld              h, ( SCRATCHPAD_RAM + $080 * 2 ) >> 8

;        cp              ( $F0 << 1 ) & $FF  ; is the address >= $1F0 ?

        add             b               ; is it >= $1F0 ? (values $1E9-$1EF
                                        ; may pass thru, depending on flags)
        jr              c, AP_AY8914    ; yes, so _probably_ poked to AY8914

        ; ^ 6/5 cycles, 5 bytes

        ld              a, [de]         ; a <- LOW byte of [RS]
        ld              [hl], a         ; store it out to SCRATCHPAD_RAM

        reti                            ; return, enabling interrupts

        ; ^ 8 cycles, 3 bytes

ifMVO_AY8914_1::                        ; MoVe Out to AY8914

        ld              h, ( SCRATCHPAD_RAM + $0F0 * 2 ) >> 8

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; e <- LOW byte of OBJ_01 (RS)
                                        ; d <- HIGH byte of OBJ_01 (BB)

        ld              l, d            ; hl -> SCRATCHPAD_RAM "high"
        ld              d, c            ; de -> RS

        ; ^ 8 cycles, 6 bytes

AP_AY8914:                              ; AY8914 @ $01F0-$01FF

        ld              a, [de]         ; a <- LOW byte of [RS]

        cp              [hl]            ; "new" value same as old value ?
        jr              nz, PA_Diff     ; no, so process new value

        ; ^ 7/6 cycles, 5 bytes

        ; poking the same value, so only process it if the location is $01FA

        ld              a, l            ; a = ( $F0 + Reg ) << 1

        cp              ( $FA << 1 ) & $FF  ; is it the "envelope type" reg ?
        jr              nz, ET_Exit         ; no, so nothing to process

        ; ^ 6/5 cycles, 5 bytes

        ld              a, [de]         ; restore 8-bit value that was poked

        ; ^ 2 cycle, 1 byte

AY8914_EnvType:                         ; AY8914 Register #10 @ $01FA

        ld              hl, ayEnvType

        add             a, a
        ld              [hl-], a        ; update ayEnvType
        ld              [hl], h         ; update ayGetEnv (set bit #4)

        ; ^ 8 cycles, 6 bytes

ET_Exit:

        reti                            ; return, enabling interrupts

        ; ^ 4 cycles, 1 byte

PA_Diff:

;        di                              ; don't allow VBlankFunc to interrupt

        ld              [hl], a         ; write new value to SCRATCHPAD_RAM

        inc             h               ; = ld h, ayRegToHandler >> 8

        ld              e, [hl]
        inc             l
        ld              h, [hl]
        ld              l, e            ; hl <- address of handling routine

        jp              [hl]            ; jump to AY8914 handling routine

        ; ^ 11 cycles, 8 bytes

ErrorMsgs:

        DB              "  Intellivision for ", 0
        DB              "            GameBoy ", 0
        DB              "               v1.1 ", 0

        DB              "By Carl Mueller, Jr.", 0
        DB              " Copyright @2000,03 ", 0

        DB              "Please Reset GameBoy", 0

        DB              "   Invalid OpCode   ", 0
;        DB              "        $           ", 0

RegStr:

        DB              "R =$", 0

icBLACK EQU              0 << 10 +  0 << 5 +  0  ; text        - black color
icCREAM EQU             25 << 10 + 27 << 5 + 30  ; background  - cream color
icRED   EQU              0 << 10 +  0 << 5 + 28  ; power down  - red color
icGREY  EQU             10 << 10 + 10 << 5 + 10  ; hex address - grey

icPAL   DW              icCREAM, icBLACK, icBLACK, icBLACK
        DW              icCREAM, icBLACK, icBLACK, icGREY
        DW              icCREAM, icBLACK, icBLACK, icRED

RT_Patch::

        ld              hl, rtPatData
.AddrLoop
        ld              a, [hl+]        ; a = sppppccc
        ld              d, a

        and             $07             ; isolate Count - 1 in bits 0-2
        ld              c, a            ; save Count - 1 in c

        xor             d               ;   = spppp...
        ret             z               ; if ZF = 1, no more patch entries

        rrca                            ;   = .spppp..
        rrca                            ;   = ..spppp.
        rrca                            ;   = ...spppp
        or              $C0             ;   = 11.spppp (HIGH of PatchAddr)

        ld              d, a            ; d = HIGH of PatchAddr

        ld              a, [hl+]
        ld              e, a            ; e = LOW of PatchAddr

        inc             c               ; c = Count
.DataLoop
        ld              a, [hl+]        ; a <- PatchByte

        ld              [de], a         ; store out PatchByte
        inc             de

        dec             c               ; dec Count
        jr              nz, .DataLoop

        jr              .AddrLoop

SECTION_01D3_End::

        PRINTT          "- SECTION_01D3: "
        PRINTV          256 - ( SECTION_01D3_End - SECTION_01D3_Start )
        PRINTT          " bytes still available.\n"

SECTION "AddrHiToPoke_SystemLow",HOME[$02D3 - 4]

SECTION_02D3_Start::

ifMVO_BKTB_1::                          ; MoVe Out to BacKTab

        ; OBJ_00:       ifMVO_BKTB
        ; OBJ_01:       BB .SHL. 8 + RS .SHL. 1

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; e <- LOW byte of OBJ_01 (RS)
                                        ; d <- HIGH byte of OBJ_01 (BB)

        ld              a, d            ; a = LOW byte of BACKTAB address
        ld              d, c            ; de -> RS

        ; ^ 6 cycles, 4 bytes

AP_BackTab:                             ; BACKTAB @ $0200-$02EF

        ld              l, a            ; l = LOW byte of BACKTAB address
        ld              h, SYSTEM_RAM >> 9

        add             hl, hl          ; hl -> SYSTEM_RAM[ a * 2 ]

        ; first check if BACKTAB word is new before calling BktWrite_XX

        ld              a, [de]         ; a = LOW byte of "new" BACKTAB word
        inc             e
        cp              [hl]            ; is it = LOW of old BACKTAB word ?
        ld              [hl+], a        ; store out LOW of "new" BACKTAB word
        ld              a, [de]         ; a = HIGH byte of "new" BACKTAB word
        jr              nz, .Diff       ; yes, so update BACKTAB w/ new word

        cp              [hl]            ; is it = HIGH of old BACKTAB word ?
        jr              z, .Exit        ; yes, so no reason to update BACKTAB
.Diff
        ld              [hl-], a        ; store out HIGH of new BACKTAB word

        ld              d, a            ; d = HIGH byte of new BACKTAB word
        ld              e, [hl]         ; e = LOW byte of new BACKTAB word

        ; ^ 25 cycles, 17 bytes

        ; SYSTEM_RAM "low" location IS new, so see if it's a BACKTAB location

        ld              a, h            ; a = 11.1...p (p = 0 or 1)
        rrca                            ;   = p11.1... (CY = p)
        ld              a, l
        rra                             ; a = LOW of SYSTEM_RAM address

        sub             $F0             ; really a BACKTAB location ?

;        ld              a, l            ; a = LOW of SYSTEM_RAM address << 1
;
;        sub             ( $F0 << 1 ) & $FF  ; really a BACKTAB location ?
        jp              c, BktWrite         ; yes, so do a write to BACKTAB
        jr              nz, .Exit           ; no, but not $02F0 (UDB) either

        inc             a               ; a = 1
        ld              [UDB_Flag], a   ; [UDB_Flag] <- 1
.Exit
        reti                            ; return, enabling interrupts

; **
; HOME_1414:  Setup the locations [$0111-$0113] with the _next_ even MO #
;             (plus resolution bits for even and odd in bits 0-1), and the
;	      "calculated" picture #s of the even (in [$0112]), and odd
;             (in [$0113]), based on the #/frames of animation and picture
;	      base fields located by each MO's "COLLTAB" field, at the same
;	      time cycling animations using the SQPC field of each MO.
;

HOME_1414::

        ld              sp, SP_1        ; use STACK_1 for EXEC_1414

        ld              a, [SCRATCHPAD_RAM + $011 * 2]

	add		2 << 2          ; advance to next MO pair
        rrca
        rrca                            ; put MO # in bits 0-2
        and             $06             ; keep MO # in 0, 2, 4, and 6
        ld              c, a            ; save a copy in c

        inc             a               ; do odd MO of pair first

        call            .CycleAnimations  ; cycle animations for odd MO

        ld              [SCRATCHPAD_RAM + $013 * 2], a  ; store odd's pic #

        ld              a, c            ; a = MO # in bits 1-3
        rra                             ; put MO # in bits 0-1 (CY was = 0)

        call            .CycleAnimations  ; cycle animations for even MO

        ld              hl, SCRATCHPAD_RAM + $012 * 2

        ld              [hl-], a        ; [$112] <- even MO's pic #
        dec             l
        ld              [hl], c         ; [$111] <- MO # + res. bits

        xor             a
        ld              [SCRATCHPAD_RAM + $014 * 2], a  ; [$114] <- 0

        ld              bc, rROMB0 + cROMB0 & $FF

        ld              a, BANK(EXEC)   ; a = #/bank where EXEC is

        ld              [c], a          ; cROMB0 <- new ROM bank # (for ISRs)
        ld              [bc], a         ; rROMB0 <- new ROM bank #

        di                              ; disable interrupts to restore stack

        ld              sp, $42A0       ; return to EXEC @ $1150

        RETURN          0               ; return, enabling interrupts

.CycleAnimations

        ld              hl, SCRATCHPAD_RAM + $004 * 2

        cp              [hl]            ; is MO in use (in 0..[$104] - 1) ?
        jr              nc, .ZeroPic    ; no, so set picture # to zero

        ; find MO's record in SYSTEM_RAM, then get YRES and COLLTAB pointer

        swap            a               ; a = MO # * 16
        add             $01D << 1       ;   = MO # * 16 + $01D << 1
        ld              l, a
        ld              h, ( SYSTEM_RAM + $0100 * 2 ) >> 8

        ld              a, [hl+]        ; a <- LOW byte of MO's attrib. word
        add             a, a
        add             a, a            ; put YRES in CY
        rl              c               ; put YRES in bit #0 of c

        inc             l               ; skip HIGH byte of MO's attrib. word

        ld              a, [hl+]
        ld              e, [hl]         ; e <- HIGH byte of COLLTAB
        ld              b, a            ; b <- LOW byte of COLLTAB

        or              e               ; is COLLTAB = 0 ?
        ret             z               ; yes, so object is not "alive"

        ; now switch to the ROM bank where COLLTAB is (saves code below)

        ld              d, AddrHiToROMB0 >> 8

        ld              a, [de]         ; a <- #/ROM bank where COLLTAB is

        ldh             [cROMB0], a     ; [cROMB0] <- ROM bank # (for ISRs)
        ld              [rROMB0], a     ; [rROMB0] <- ROM bank #

        inc             d
        ld              a, [de]         ; a <- HIGH of COLLTAB >> 1 (real)

        sla             b
        rla
        ld              d, a
        ld              e, b            ; de -> COLLTAB (real)

        ; add sequence speed to the counter and adv. the pic. # if necessary

        ld              a, l
        add             ( 5 - 1 ) * 2 - 1
        ld              l, a            ; pos MO record @ LOW byte of SQPC

 	ld		a, [hl+]        ; a <- LOW byte of SQPC

 	and		$F0             ; isolate LSBs of SQ (in bits 4-7)
 	xor		[hl]            ; xor them w/ HIGH byte of SQPC
        and		$FC             ; preserve bits 2-7
        xor	        [hl]		; put MSBs of SQ in bits 0-1

        cp              $F3             ; is it = maximum speed (63) ?
        jr              z, .MaxSpeed    ; yes, so always inc. the picture #

 	rrca
 	rrca                            ; put MSBs in bits 6-7

        add             [hl]            ; add it to the counter in bits 2-7
        ld              [hl-], a        ; and update the HIGH byte of SQPC
        jr              c, .PicAdv      ; if CY = 1, advance picture #

        ; counter did not overflow, so just get the picture # & add its base

        ld		a, [hl]		; a <- LOW byte of SQPC

        ld              hl, 5 * 2
        add             hl, de          ; hl -> COLLTAB @ picture base

	and		$0F		; isolate picture # in bits 0-3
.ChkYRES
        bit             0, c            ; is MO's YRES set (double res.) ?
        jr		z, .NoYRES      ; no, so don't double the offset

        add		a, a		; double res., so double the offset
.NoYRES
        add		[hl]            ; add the picture base
        inc		a		; put picture in 1..255 (0 = invalid)

	ret

        ; counter did overflow, so inc picture # and compare it to #/frames

.MaxSpeed
	dec		l		; pos @ LOW byte of SQPC
.PicAdv
        inc	        [hl]		; advance to next picture

        ld              a, l            ; save (pointer to) LOW byte of SQPC

        ld              hl, 4 * 2
        add             hl, de          ; hl -> COLLTAB @ #/frames

        ld              e, a
        ld              d, ( SYSTEM_RAM + $0100 * 2 ) >> 8

 	ld		a, [de]         ; a <- LOW byte of new SQPC word
 	and		$0F		; isolate picture # in bits 0-3

 	cp 		[hl]            ; is it = #/frames of animation ?

;        inc             l               ; pos @ HIGH byte of #/frames
        inc             hl              ; pos @ HIGH byte of #/frames
        inc             hl              ; pos @ LOW byte of picture base

;        jr              c, .ChkYRES     ; * no, so picture will not be zero'd
        jr              nz, .ChkYRES    ; no, so picture will not be zero'd

        ; went one picture beyond last of sequence, so set picture # = 0

        ld              a, [de]         ; a <- LOW byte of new SQPC word
        and             $F0             ; clear picture # in bits 0-3
        ld              [de], a         ; update LOW byte of new SQPC word

        ld		a, [hl]         ; a = picture base
	inc		a               ; put picture in 1..255 (0 = invalid)

	ret

.ZeroPic

        ; MO not in the "active" range (0..[$104] - 1), so set pic # = 0

        xor             a               ; set picture # = 0 (invalid)
        rl              c               ; set MO's res. bit = 0 (not used)

	ret

SECTION_02D3_End::

        PRINTT          "- SECTION_02D3: "
        PRINTV          256 - ( SECTION_02D3_End - SECTION_02D3_Start )
        PRINTT          " bytes still available.\n"

SECTION "AddrHiToPoke_SystemHigh",HOME[$03D3]

SECTION_03D3_Start::

AP_SysHi::                              ; SYSTEM_RAM "high" @ $0300-$035F

        ld              h, ( SYSTEM_RAM + $0100 * 2 ) >> 8

        add             a, a
        ld              l, a            ; hl -> SYSTEM_RAM[ ( $0100 + a ) * 2 ]

        ld              a, [de]         ; a <- LOW byte of [RS]
        ld              [hl+], a        ; store out LOW byte of [RS]

        inc             e               ; pos @ HIGH byte of [RS]

        ld              a, [de]         ; a <- HIGH byte of [RS]
        ld              [hl], a         ; store out HIGH byte of [RS]

        reti                            ; return, enabling interrupts

        ; ^ 17 cycles, 10 bytes

UDB_Update::

        ; get the UDB address from SYSTEM_RAM @ $02F0 at convert to "real"

        ld              hl, SYSTEM_RAM + $0F0 * 2

        ld              e, [hl]         ; e <- LOW byte of address
        inc             l

        call            .CP1610ToReal   ; hl -> UDB

        ; copy the 8 GRAM offsets in the UDB to UDB_CardNdx

        ld              de, UDB_CardNdx
.NdxLoop
        ld              a, [hl+]        ; a <- LOW of GRAM offset

        ld              [de], a         ; store it out to UDB_CardNdx
        inc             e

        ld              a, [hl+]        ; a <- HIGH of GRAM offset

        and             $01             ; isolate MSB of CardNdx
        add             GRAM_8 >> 8     ; add GRAM_8 base

        ld              [de], a         ; store it out to UDB_CardNdx
        inc             e

        bit             4, e            ; copied all 8 GRAM offsets ?
        jr              z, .NdxLoop     ; no, so continue to loop

        ; get the KEYCLICKS decle from UDB and save it in RAM

        ld              de, ( $0C - $1C ) * 2
        add             hl, de          ; hl -> UDB @ KEYCLICKS

        ld              a, [hl]         ; a <- KeyClicks
        ld              [ehKeyClicks], a

        ; get the OBJGRAM address from UDB and convert to "real"

        ld              de, ( $00 - $0C ) * 2
        add             hl, de          ; hl -> UDB @ OBJGRAM

        ld              e, [hl]         ; e <- LOW byte of address
        inc             l
        inc             hl

        ld              a, [hl]         ; a <- HIGH byte of address
        or              e               ; is OBJGRAM = 0 (avoid Echo RAM) ?
        jr              z, .Proc        ; yes, so don't copy any pictures

        call            .CP1610ToReal   ; hl -> OBJGRAM

        ld              a, 4            ; a = #/WRAM bank where OBJGRAM is
        ldh             [cSVBK], a      ; [cSVBK] <- WRAM bank # (for ISRs)
        ldh             [rSVBK], a      ; [rSVBK] <- WRAM bank #

        ld              de, OBJGRAM
.ObjLoop
        ld              a, [hl+]        ; a <- 8-bit value from cartridge
        inc             hl
        ld              [de], a         ; store it out to OBJGRAM
        inc             de

        bit             3, d            ; copied 256 pictures yet ?
        jr              z, .ObjLoop     ; no, so keep looping

        ld              a, 1            ; a = #/WRAM bank where R0, etc. is
        ldh             [cSVBK], a      ; [cSVBK] <- WRAM bank # (for ISRs)
        ldh             [rSVBK], a      ; [rSVBK] <- WRAM bank #

        ;

.Proc

        ld              hl, SYSTEM_RAM + $0F0 * 2

        ld              e, [hl]         ; e <- LOW byte of address
        inc             l

        call            .CP1610ToReal   ; hl -> UDB

        ld              de, ( $02 - $14 ) * 2
        add             hl, de          ; hl -> PLIST address

        ld              e, [hl]         ; a <- LOW of PLIST address
        inc             l
        inc             hl

        call            .CP1610ToReal   ; hl -> PLIST

        ld              de, ehProcList
        ld              c, 6 * 4 + 2    ; copy 6 entries + 2 for EOT
.ProcCopy
        ld              a, [hl+]
        inc             hl
        ld              [de], a
        inc             e

        dec             c
        jr              nz, .ProcCopy

        dec             c               ; c = -1

        ld              hl, ehProcList  ; hl -> ehProcList
.ProcCount
        ld              a, [hl+]        ; a <- LOW byte of ProcAddr
        ld              d, [hl]         ; d <- HIGH byte of ProcAddr
        inc             l

        inc             l
        inc             l               ; skip CtrRefresh

        inc             c               ; inc ProcCount

        or              d               ; is ProcAddr = 0 ?
        jr              nz, .ProcCount

        ld              a, c
        ld              [ehProcCount], a

        ld              c, cROMB0 & $FF

        ;

        ld              a, BANK(EXEC)   ; a = #/ROM bank where EXEC is
        ld              [c], a          ; [cROMB0] <- ROM bank # (for ISRs)
        ld              [bc], a         ; [rROMB0] <- ROM bank #

        reti                            ; return, making sure interrupts ON

.CP1610ToReal

        ld              l, [hl]         ; l <- HIGH byte of address
        ld              h, c            ; hl -> AddrHiToROMB0

        ld              a, [hl]         ; a <- #/ROM bank
        ld              [c], a          ; [cROMB0] <- ROM bank # (for ISRs)
        ld              [bc], a         ; [rROMB0] <- ROM bank #

        inc             h
        ld              h, [hl]         ; h <- HIGH of address (real) >> 1

        ld              l, e            ; l = LOW byte of address
        add             hl, hl          ; hl -> UDB

        ret

INCLUDE         "PATCH.ASM"

SECTION_03D3_End::

        PRINTT          "- SECTION_03D3: "
        PRINTV          256 - ( SECTION_03D3_End - SECTION_03D3_Start )
        PRINTT          " bytes still available.\n"

SECTION "AddrHiToPoke_Stack",HOME[$04D3]

SECTION_04D3_Start::

AP_Stack::                              ; SYSTEM_RAM "high" @ $0300-$035F

        ld              l, a
        ld              h, c            ; = ld h, AddrHiToROMB0 >> 8

        inc             l

        ld              a, [de]         ; a <- LOW byte of [RS]
        ld              [hl-], a        ; store out LOW byte of [RS]

        inc             e               ; pos @ HIGH byte of [RS]
        inc             h

        ld              a, [de]         ; a <- HIGH byte of [RS]
        ld              [hl], a         ; store out HIGH byte of [RS]

        reti                            ; return, enabling interrupts

; ---------------------------------------------------------------------------
;                  .::** CP1610 Emulation Code **::.
; ---------------------------------------------------------------------------

CPU_BEGIN:

INCLUDE         "CPU_NEW.ASM"
INCLUDE         "HOMEHLPR.ASM"
INCLUDE         "EHJSR.ASM"  ; EXEC Helper -- Jump to SubRoutine replacements
        IF      COLOR_TUNE == 1
INCLUDE         "COLTUNE.ASM"
        ENDC
INCLUDE         "COPYRITE.ASM"

CPU_END:

        PRINTT  "- CPU: "
        PRINTV  CPU_END - CPU_BEGIN + 1
        PRINTT  " bytes consumed by CP1610 implementing function code.\n"

; ---------------------------------------------------------------------------
;          .::** Library includes for HOME section go here **::.
; ---------------------------------------------------------------------------

LIBINC_BEGIN:

INCLUDE "STIC.ASM"          ; STIC_Write, GRAM_Write, Bkt_Write, etc.

SetCol0::

        ei                              ; interrupts can't be off for so long

        ld              a, 3
        ldh             [cSVBK], a
        ldh             [rSVBK], a      ; switch to WRAM #3 (for PALTAB_0)

        ld              hl, ColorStack  ; hl -> ColorStack

        ld              a, [hl]         ; a <- ColNum << 1

        add             ColorToRGB & $FF
        ld              l, a            ; hl -> ColorToRGB[ ColNum << 1 ]

        ld              d, [hl]         ; d <- LOW byte of RGB value

        ld              hl, ColorStack  ; hl -> ColorStack

        ld              a, c
        cp              cROMB0 & $FF    ; color already set ?
        jr              nz, .SetCS0     ; yes, so don't choose a new color

        xor             a               ; a = 0 (Black)

        cp              [hl]            ; is ColorStack #0 = Black ?
        jr              nz, .SetCS0     ; no, so make it Black

        ld              a, e            ; a = 2 << 1 (Red), etc.
.SetCS0
        ld              [hl], a         ; ColorStack #0 <- a

        add             ColorToRGB & $FF
        ld              c, a            ; c -> ColorToRGB

        ld              hl, PALTAB_0    ; hl -> PALTAB_0 (RGB)
        ld              e, PALTAB_PALSIZE
.PalLoop
        ld              a, d            ; a = LOW byte of (old) RGB value

        cp              [hl]            ; same color ?
        jr              nz, .Skip       ; no, so don't overwrite

        ld              a, [c]          ; a <- LOW byte of RGB value
        ld              [hl+], a        ; store out to LOW byte of Color #0

        inc             c               ; pos @ HIGH byte of RGB value

        ld              a, [c]          ; a <- HIGH byte of RGB value
        ld              [hl-], a        ; store out to HIGH byte of Color #0

        dec             c               ; re-pos @ LOW byte of RGB value
.Skip
        ld              a, l
        add             e               ; pos @ Color #0 of next Palette
        ld              l, a

        jr              nz, .PalLoop    ; no, so keep looping

        ld              a, 1            ; * necessary ?
        ldh             [cSVBK], a
        ldh             [rSVBK], a      ; switch back to WRAM #1

        ld              c, cROMB0 & $FF

        di                              ; disable interrupts to load SP

        ret

;INCLUDE "PATCH2.ASM"

; **
; ReadJoy:     Reads the status of the joypad, putting newly pressed buttons
;              in JoyPress, while updating JoyDown with those that are
;              currently being held down.
;
; Parameters:  None.
;
; Modifies:
;
; Costs:       64 bytes.
;

ReadJoy:

        ; first read in the status of the d-button by selecting P14
        ld              a, P1F_5        ; select P14 by setting it low
        ldh             [rP1], a        ; [$FF00] <- a (read joypad info)
        ldh             a, [rP1]
        ldh             a, [rP1]        ; wait a few cycles (??)
        and             $0F             ; isolate right, left, up & down bits
        swap            a
        ld              b, a

        ; now read in the status of buttons a, b, select, and start
        ld              a, P1F_4        ; select P15 by setting it low
        ldh             [rP1], a        ; [$FF00] <- a (read joypad info)
        ldh             a, [rP1]
        ldh             a, [rP1]
        ldh             a, [rP1]
        ldh             a, [rP1]
        ldh             a, [rP1]
        ldh             a, [rP1]        ; wait a few MORE cycles (??)
        and             $0F             ; isolate a, b, select and start bits

        or              b               ; combine both sets
        ld              b, a

        ld              a, [JoyUp]      ; a <- old JoyUp data
        xor             b               ; a = mask for changed buttons
        and             b               ; allow only changed buttons to pass
        ld              [JoyRelease], a

        ld              a, b
        ld              [JoyUp], a

        cpl                             ; flip all bits so that UP=0, DOWN=1
        ld              b, a

        ld              a, [JoyDown]    ; a <- old JoyDown data
        xor             b               ; a = mask for changed buttons
        and             b               ; allow only changed buttons to pass
        ld              [JoyPress], a

        ld              a, b
        ld              [JoyDown], a

        ld              a, P1F_4 + P1F_5  ; deselect P14 and P15
        ldh             [rP1], a          ; reset joypad (???)

        ret

LIBINC_END:

                PRINTT      "- LIBINC: "
                PRINTV      LIBINC_END - LIBINC_BEGIN + 1
                PRINTT      " bytes consumed by library includes.\n"

SECTION_04D3_End::

        PRINTT          "- SECTION_04D3: "
        PRINTV          13312 - ( SECTION_04D3_End - SECTION_04D3_Start )
        PRINTT          " bytes still available.\n"

SECTION "AddrHiToPoke_GRAM_Low",HOME[$38D3 - 4]

ifMVO_GALO_1::                          ; MoVe Out to GrAm LOw

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; e <- LOW byte of OBJ_01 (RS)
                                        ; d <- HIGH byte of OBJ_01 (BB)

        ld              a, d            ; a = LOW byte of GRAM address
        ld              d, c            ; de -> RS

SECTION_38D3_Start::

AP_GRAM_Low::                           ; GRAM "low" @ $3800-$38FF

        ld              l, a
        ld              h, GRAM_8 >> 8  ; hl -> GRAM_8[ a ]

        ld              a, [de]         ; a <- LOW byte of [RS]

        cp              [hl]            ; is it equal to the old value ?
        jr              z, .Exit        ; yes, so nothing to do

        ld              [hl], a         ; store out LOW of [RS] to GRAM_8

        ld              h, GRAM_16 >> 9

        add             hl, hl          ; hl -> GRAM_16[ a * 2 ]

        ld              [hl], a         ; store out LOW of [RS] to GRAM_16

        ld              a, l
        and             $F0             ; isolate LSBs of card index
        sub             GRAM_16 >> 8    ; take off GRAM_16's base address
        add             h               ; put MSBs of card index in bits 0-1

        ld              l, a
        ld              h, GRAM_Flags >> 8  ; hl -> GRAM_Flags[ CardNdx ]

        ld              [hl], h         ; set GRAM_Flag <> 0

        IF              GRAM_MODIFIED == 1

        ld              l, h
        ld              [hl], h         ; set GRAM_Modified <> 0

        ENDC
.Exit
        reti                            ; return, enabling interrupts

        ; ^ 32 cycles

INCLUDE "MEMORY.ASM"        ; MemCopy, MemFill

SECTION_38D3_End::

        PRINTT      "- SECTION_38D3: "
        PRINTV      256 - ( SECTION_38D3_End - SECTION_38D3_Start )
        PRINTT      " bytes still available.\n"

SECTION "AddrHiToPoke_GRAM_High",HOME[$39D3 - 4]

ifMVO_GAHI_1::                          ; MoVe Out to GrAm HIgh

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; e <- LOW byte of OBJ_01 (RS)
                                        ; d <- HIGH byte of OBJ_01 (BB)

        ld              a, d            ; a = LOW byte of GRAM address
        ld              d, c            ; de -> RS

SECTION_39D3_Start::

AP_GRAM_High:                           ; GRAM "high" @ $3900-$39FF

        ld              l, a
        ld              h, ( GRAM_8 + $0100 ) >> 8  ; hl -> GRAM_8[ a ]

        ld              a, [de]         ; a <- LOW byte of [RS]

        cp              [hl]            ; is it equal to the old value ?
        jr              z, .Exit        ; yes, so nothing to do

        ld              [hl], a         ; store out LOW of [RS] to GRAM_8

        ld              h, ( GRAM_16 + $0100 * 2 ) >> 9

        add             hl, hl          ; hl -> GRAM_16[ a * 2 ]

        ld              [hl], a         ; store out LOW of [RS] to GRAM_16

        ld              a, l
        and             $F0             ; isolate LSBs of card index
        sub             GRAM_16 >> 8    ; take off GRAM_16's base address
        add             h               ; put MSBs of card index in bits 0-1

        ld              l, a
        ld              h, GRAM_Flags >> 8  ; hl -> GRAM_Flags[ CardNdx ]

        ld              [hl], h         ; set GRAM_Flag <> 0

        IF              GRAM_MODIFIED == 1

        ld              l, h
        ld              [hl], h         ; set GRAM_Modified <> 0

        ENDC
.Exit
        reti                            ; return, enabling interrupts

INCLUDE "PATCH2.ASM"

; ---------------------------------------------------------------------------
;                       .::** PROGRAM START **::.
; ---------------------------------------------------------------------------

Start:

	ld		sp, SP_0        ; SP_0 located 1 byte beyond STACK_0

        ld              h, rROMB0 >> 8
        ld              [hl], BANK(MISC)  ; switch to bank where init code is

        call            GB_Reset        ; init everything needed at reset

        ; * let's investigate the rest of the initialization process !!!

        SETBANK_ROM     COLL

        call            STIC_CollInit   ; won't fit in MISC bank

        SETBANK_ROM     AY8914_Init

        call            AY8914_Init     ; won't fit in MISC bank

        SETBANK_ROM     BKG

        call            Bkg_Init        ; won't fit in MISC bank

        SETBANK_RAM     1

        ld              a, LCDC_WINON
        ldh             [rLCDC], a      ; enable display with window = ON

        IF              COLOR_TUNE == 1

        ld              hl, ColComp

        xor             a
        ld              [hl+], a        ; ColComp = 0
        ld              [hl+], a        ; ColNum = 0

        ENDC

        ; -------------------------------------------------------------------
        ;                        GetStackStart
        ; -------------------------------------------------------------------

        ld              hl, AddrHiToROMB0 + $80 - $10

        ld              c, 0            ; c = MaxPageSize

        ;

        ld              a, BANK(EMPTY)
.NotEquEnd
        ld              d, l            ; d = CurPageStart
.EquLoop
        cp              [hl]
        jr              nz, .EquEnd

        inc             l
        jr              nz, .EquLoop
.EquEnd
        ld              a, l
        sub             d               ; a = CurPageStart - CurPageEnd

        cp              c               ; is CurPageSize < MaxPageSize ?
        jr              c, .LessThan    ; yes, so don't replace it

        ld              c, a            ; c = MaxPageSize
        ld              b, d            ; b = MaxPageStart
.LessThan
        add             d               ; reached end of AddrHiToROMB0 ?
        jr              z, .Exit        ; yes, so ...

        ;

        ld              a, BANK(EMPTY)
.NotEquLoop
        cp              [hl]
        jr              z, .NotEquEnd

        inc             l
        jr              nz, .NotEquLoop
.Exit
        ld              a, b
        ld              [StackStart], a

        call            Copyright

        SETBANK_ROM     EXEC

        call            RT_Patch

        ei                              ; finally, enable interrupts
Reset:
        call            Intv_Reset      ; init CP1610 to its power-on state

.MainLoop

        ; -------------------------------------------------------------------
        ;                    .::** Main Loop **::.
        ; -------------------------------------------------------------------

        IF              SYNC_DISPLAY == 0

        ld              hl, VBlank_Ctr  ; hl -> VBlank_Ctr

        xor             a
.VBL
        cp              [hl]            ; is VBlank_Ctr = 0 ?
        jr              z, .VBL         ; yes, so wait for VBlankFunc to run

        ld              [hl+], a
        ld              [hl], a         ; set STIC_Display = 0

        ELSE

        ld              hl, Page_Ready

.Wait   ld              a, [hl]
        dec             a               ; is Page_Ready = 1 ?
        jr              z, .Wait        ; yes, so wait until VBlankFunc runs

        xor             a
        ld              [STIC_Display], a  ; set STIC_Display = 0

        ENDC

        call            Process         ; process 1/60 worth of instructions

        ld              a, [STIC_Display]     ; see if prg. enabled display
        ldh             [STIC_Display_Hi], a  ; and update it in High RAM

        ldh             a, [cROMB0]     ; a <- current ROM bank #

        push            af              ; save current ROM bank #

        SETBANK_ROM     COLL            ; switch to bank with collision code

        ; NOTE:  STIC_CollPrep must be called before STIC_Analyze, so it
        ;        can check the status of the ReCopy bit (in MO attribute
        ;        byte #1) before it is cleared.

        call            STIC_CollPrep     ; prepare COLLMO for routines below
        call            STIC_CollChkMO    ; flag collisions between MOs
        call            STIC_CollChkScrn  ; flag collisions w/ screen edges

        SETBANK_RAM     3               ; switch to WRAM bank #3

        call            STIC_CollChkBkg   ; flag collisions w/ background

        SETBANK_ROM     MISC            ; select bank where routines below are

        SETBANK_RAM     6               ; switch to WRAM bank #6

        call            GRAM_CheckFlags ; handle changes to GRAM cards

        SETBANK_RAM     3               ; switch to WRAM bank #3

        SETBANK_ROM     BKG

        call            ToggleGraphicsMode
        call            BktGetCols
        call            STIC_OverFlow

        SETBANK_RAM     5               ; switch to WRAM bank #5

        SETBANK_ROM     MISC

        call            STIC_Analyze    ; update LocalOAM with new STIC info.
        call            MO_GetColors    ; update MO_Colors w/ new RGB values

        ld              a, 1
        ld              [Page_Ready], a ; set Page_Ready = 1

        SETBANK_RAM     1               ; switch back to WRAM bank #1

        call            HandleControls

        pop             af              ; restore current ROM bank #

        ldh             [cROMB0], a     ; note bank # for ISRs
        ld              [rROMB0], a     ; switch back to previous ROM bank

        IF              COUNT_FRAMES == 1

        ld              hl, CurFrames
        inc             [hl]            ; inc CurFrames

        ENDC

        IF              COLOR_TUNE == 1

        call            ColorTune

        ENDC

        jp              .MainLoop

        ; -------------------------------------------------------------------
        ;    Handlers for writes to AY8914 registers (see ayRegToHandler)
        ; -------------------------------------------------------------------

AY8914_TonAPerLo:                       ; AY8914 Register #0 @ $01F0

        ld              a, [SCRATCHPAD_RAM + $0F4 * 2]

AY8914_TonAPerHi:                       ; AY8914 Register #4 @ $01F4

        and             $0F             ; keep period in 0..4095
        add             a, ayPerToAudFreq >> 9
        ld              h, a
        ld              a, [SCRATCHPAD_RAM + $0F0 * 2]
        ld              l, a

        add             hl, hl          ; hl -> ayPerToAudFreq[ TonAPer * 2 ]

        ld              a, BANK(ayPerToAudFreq)
        ld              [bc], a         ; switch to bank w/ ayPerToAudFreq

        ld              a, [hl+]
        ld              d, [hl]         ; da <- "gb" frequency value (0-2047)

        ld              hl, ayTonAFreq

        ld              [hl+], a        ; update LOW byte of ayTonAFreq

        ld              a, [hl]         ; a <- HIGH byte of (old) ayTonAFreq
        and             $80             ; isolate Initial (volume changed)
        or              d               ; set MSBs of new frequency value
        ld              [hl], a         ; update HIGH byte of ayTonAFreq

        ld              a, [c]          ; a <- current rROMB0
        ld              [bc], a         ; switch back to previous bank

        reti                            ; return, enabling interrupts

AY8914_TonBPerLo:                       ; AY8914 Register #1 @ $01F1

        ld              a, [SCRATCHPAD_RAM + $0F5 * 2]

AY8914_TonBPerHi:                       ; AY8914 Register #5 @ $01F5

        and             $0F             ; keep period in 0..4095
        add             a, ayPerToAudFreq >> 9
        ld              h, a
        ld              a, [SCRATCHPAD_RAM + $0F1 * 2]
        ld              l, a

        add             hl, hl          ; hl -> ayPerToAudFreq[ TonBPer * 2 ]

        ld              a, BANK(ayPerToAudFreq)
        ld              [bc], a         ; switch to bank w/ ayPerToAudFreq

        ld              a, [hl+]
        ld              d, [hl]         ; da <- "gb" frequency value

        ld              hl, ayTonBFreq

        ld              [hl+], a        ; update LOW byte of ayTonBFreq

        ld              a, [hl]         ; a <- HIGH byte of (old) ayTonBFreq
        and             $80             ; isolate Initial (volume changed)
        or              d               ; set MSBs of new frequency value
        ld              [hl], a         ; update HIGH byte of ayTonBFreq

        ld              a, [c]          ; a <- current rROMB0
        ld              [bc], a         ; switch back to previous bank

        reti                            ; return, enabling interrupts

AY8914_TonCPerLo:                       ; AY8914 Register #2 @ $01F2

        ld              a, [SCRATCHPAD_RAM + $0F6 * 2]

AY8914_TonCPerHi:                       ; AY8914 Register #6 @ $01F6

        and             $0F             ; keep period in 0..4095
        add             a, ayPerToAudFreq >> 9
        ld              h, a
        ld              a, [SCRATCHPAD_RAM + $0F2 * 2]
        ld              l, a

        add             hl, hl          ; hl -> ayPerToAudFreq[ TonCPer * 2 ]

        ld              a, BANK(ayPerToAudFreq)
        ld              [bc], a         ; switch to bank w/ ayPerToAudFreq

        ld              a, [hl+]
        ld              d, [hl]         ; da <- "gb" frequency value

        ld              hl, ayTonCFreq

        ld              [hl+], a        ; update LOW byte of ayTonCFreq

        ld              a, [hl]         ; a <- HIGH byte of (old) ayTonCFreq
        rrca                            ; CY <- Initial (volume changed)
        ld              a, d            ; a = new frequency value
        rla                             ; put Initial in bit #0
        ld              [hl], a         ; update HIGH byte of ayTonCFreq

        ld              a, [c]          ; a <- current rROMB0
        ld              [bc], a         ; switch back to previous bank

        reti                            ; return, enabling interrupts

AY8914_EnvPerLo:                        ; AY8914 Register #3 @ $01F3

        ldh             [ayEnvPerNew], a      ; update LOW byte of envPeriod

        reti                            ; return, enabling interrupts

AY8914_EnvPerHi:                        ; AY8914 Register #7 @ $01F7

        ldh             [ayEnvPerNew + 1], a  ; update HIGH of envPeriod

        reti                            ; return, enabling interrupts

AY8914_Mixer:                           ; AY8914 Register #8 @ $01F8

        ld              hl, ayEnabled

        xor             [hl]
        and             $3F             ; isolate mixer bits (0-5)
        xor             [hl]            ; and set them to the new mixer bits
        ld              [hl], a         ; update bits 0-5 of ayEnabled

        reti                            ; return, enabling interrupts

AY8914_NoiPer:                          ; AY8914 Register #9 @ $01F9

        ld              l, a
        ld              h, ayPerToAudPoly >> 8  ; hl -> ayPerToAudPoly
                                                ;       [ NoiPer ]

        ld              a, [hl]         ; a <- polynomial based on period
        ldh             [ayAudPoly], a  ; store it out to ayAudPoly

        reti                            ; return, enabling interrupts

AY8914_ChnAVol:                         ; AY8914 Register #11 @ $01FB

        ld              hl, ayEnabled
        set             6, [hl]         ; indicate that ChnA is NOT enveloped

        and             $3F

        cp              16                ; is it >= 16 (enveloped) ?
        jr              c, .NotEnveloped  ; no, so leave ayEnabled alone

        res             6, [hl]         ; indicate that ChnA IS enveloped
        ld              a, 15           ; enveloped, so use maximum volume

.NotEnveloped

        ld              l, a            ; hl -> ayVolToAudEnv[ ChnAVol ]

        ldh             a, [rAUD1ENV]   ; a <- ChnA's cur. volume (unless 0)
        cp              [hl]            ; is it the same as the new volume ?
        jr              z, .VolSame     ; yes, but may have changed 0 -> 15

        ld              a, [hl]         ; a <- output byte for rAUDxENV

        IF              CLICK_REDUCTION == 1

        ; if the new volume is zero, update rAUDTERM instead of rAUD1ENV

        or              a               ; is the new volume = 0 ?
        jr              z, .VolZero     ; if volume = 0, use rAUDTERM instead

        ENDC

        ldh             [rAUD1ENV], a   ; update rAUD1ENV (bad for NO$GMB)

        ld              l, ayTonAFreq & $FF + 1
        set             7, [hl]         ; set Initial (make volume change)

.VolSame

        IF              CLICK_REDUCTION == 1

        ld              l, ayMuted & $FF
        res             0, [hl]         ; make sure TonA not muted
        res             3, [hl]         ; make sure NoiA not muted

        reti                            ; return, enabling interrupts

.VolZero

        ld              l, ayMuted & $FF
        set             0, [hl]         ; mute TonA
        set             3, [hl]         ; mute NoiA

        ENDC

        reti                            ; return, enabling interrupts

AY8914_ChnBVol:                         ; AY8914 Register #12 @ $01FC

        ld              hl, ayEnabled
        set             7, [hl]         ; indicate that ChnB is NOT enveloped

        and             $3F

        cp              16                ; is it >= 16 (enveloped) ?
        jr              c, .NotEnveloped  ; no, so leave ayEnabled alone

        res             7, [hl]         ; indicate that ChnB IS enveloped
        ld              a, 15           ; enveloped, so use maximum volume

.NotEnveloped

        ld              l, a            ; hl -> ayVolToAudEnv[ ChnBVol ]

        ldh             a, [rAUD2ENV]   ; a <- ChnB's cur. volume (unless 0)
        cp              [hl]            ; is it the same as the new volume ?
        jr              z, .VolSame     ; yes, but may have changed 0 -> 15

        ld              a, [hl]         ; a <- output byte for rAUDxENV

        IF              CLICK_REDUCTION == 1

        ; if the new volume is zero, update rAUDTERM instead of rAUD2ENV

        or              a               ; is the new volume = 0 ?
        jr              z, .VolZero     ; if volume = 0, use rAUDTERM instead

        ENDC

        ldh             [rAUD2ENV], a   ; update rAUD2ENV (bad for NO$GMB)

        ld              l, ayTonBFreq & $FF + 1
        set             7, [hl]         ; set Initial (make volume change)

.VolSame

        IF              CLICK_REDUCTION == 1

        ld              l, ayMuted & $FF
        res             1, [hl]         ; make sure TonB not muted
        res             4, [hl]         ; make sure NoiB not muted

        reti                            ; return, enabling interrupts

.VolZero

        ld              l, ayMuted & $FF
        set             1, [hl]         ; mute TonB
        set             4, [hl]         ; mute NoiB

        ENDC

        reti                            ; return, enabling interrupts

AY8914_ChnCVol:                         ; AY8914 Register #13 @ $01FD

        ld              hl, ayEnabled + 1
        set             0, [hl]         ; indicate that ChnC is NOT enveloped

        and             $3F

        cp              16                ; is it >= 16 (enveloped) ?
        jr              c, .NotEnveloped  ; no, so leave ayEnabled alone

        res             0, [hl]         ; indicate that ChnC IS enveloped
        ld              a, 15           ; enveloped, so use maximum volume

.NotEnveloped

        ld              l, a            ; hl -> ayVolToAudEnv[ ChnAVol ]

        ld              a, [hl]         ; a <- output byte for rAUDxENV
        swap            a
        or              [hl]            ; put volume in LOW and HIGH nybbles

        ld              l, ayChnCVol & $FF
        cp              [hl]            ; is it the same as the old volume ?
        jr              z, .VolSame     ; yes, but may have changed 0 -> 15

        IF              CLICK_REDUCTION == 1

        ; if the new volume is zero, update rAUDTERM instead of Wave RAM

        or              a               ; is the new volume = 0 ?
        jr              z, .VolZero     ; if volume = 0, use rAUDTERM instead

        ENDC

        ld              [hl-], a        ; update ayChnCVol

        set             0, [hl]         ; set Initial (make volume change)

.VolSame

        IF              CLICK_REDUCTION == 1

        ld              l, ayMuted & $FF
        res             2, [hl]         ; make sure TonC not muted
        res             5, [hl]         ; make sure NoiC not muted

        reti                            ; return, enabling interrupts

.VolZero

        ld              l, ayMuted & $FF
        set             2, [hl]         ; mute TonC
        set             5, [hl]         ; mute NoiC

        ENDC

        reti                            ; return, enabling interrupts

AY8914_HandRight:                       ; AY8914 Register #14 @ $01FE

        ld              a, $FF
        ld              [SCRATCHPAD_RAM + $0FE * 2], a

        reti

AY8914_HandLeft:                        ; AY8914 Register #15 @ $01FF

        ld              a, $FF
        ld              [SCRATCHPAD_RAM + $0FF * 2], a

        reti

InvalidCode_1::

        ld              sp, HF_WinOff

        ld              [hbfRoutine], sp  ; hbfRoutine <- hbfWinOff

        ld              sp, HF_M2 + 1
        pop             hl
        ld              sp, hl

InvalidCode::

        di                              ; don't allow ISRs to mess up timing

        ld              [R7], sp        ; update [R7]

        ld              sp, SP_0
        pop             hl
        ld              sp, hl          ; restore SP from SP_0 (use STACK_0)

        ; set the palette by modifying PALTAB_1 (updated in VBlankFunc)

        ld              a, 3
        ldh             [cSVBK], a
        ldh             [rSVBK], a      ; switch to WRAM bank #3

        ld              hl, icPAL
        ld              de, PALTAB_0 + 2 * PALTAB_PALSIZE
        ld              c, 4 * 2 * 3    ; 4 colors @ 2 bytes ea., 3 palettes
.SetPal
        ld              a, [hl+]
        ld              [de], a
        inc             e
        dec             c
        jr              nz, .SetPal

        ld              a, 1
        ldh             [cSVBK], a
        ldh             [rSVBK], a      ; switch to WRAM bank #1

        ld              [Page_Ready], a ; make sure PALTAB_0 gets updated

        ; clear the background, setting all tiles to the upper-tile bank

        ld              a, $0A                      ; set Attribute Byte
        ld              bc, ( $00 - 32 ) << 8 + 32  ; set Tile Base, RepChar
        ld              de, 32 * 32                 ; set #/iterations
        ld              hl, _SCRN1                  ; set Destination

        call            RepChar         ; clear _SCRN1

        ; finally, let's write out the relevant information

        ld              de, ErrorMsgs   ; set Source Address

        ld              hl, _SCRN1 + 2 * 32
        call            WriteStr                 ;  Intellivision for
        ld              hl, _SCRN1 + 3 * 32
        call            WriteStr                 ;            GameBoy
        ld              hl, _SCRN1 + 4 * 32
        call            WriteStr                 ;               v1.0
        ld              hl, _SCRN1 + 13 * 32
        call            WriteStr                 ; By Carl Mueller, Jr.
        ld              hl, _SCRN1 + 14 * 32
        call            WriteStr                 ;  Copyright @ 2000
        ld              hl, _SCRN1 + 16 * 32
        call            WriteStr                 ; Please Reset GameBoy

        ld              a, $0C          ; red (palette #4), upper-tile bank

        ld              hl, _SCRN1 + 6 * 32
        call            WriteStr                 ;   Invalid OpCode

        xor             a
        ld              hl, _SCRN1 + 8 * 32 + 1
        call            WriteReg                 ; R0=$XXYY

        ld              a, 1
        ld              hl, _SCRN1 + 9 * 32 + 1
        call            WriteReg                 ; R1=$XXYY

        ld              a, 2
        ld              hl, _SCRN1 + 10 * 32 + 1
        call            WriteReg                 ; R2=$XXYY

        ld              a, 3
        ld              hl, _SCRN1 + 11 * 32 + 1
        call            WriteReg                 ; R3=$XXYY

        ld              a, 4
        ld              hl, _SCRN1 + 8 * 32 + 11
        call            WriteReg                 ; R4=$XXYY

        ld              a, 5
        ld              hl, _SCRN1 + 9 * 32 + 11
        call            WriteReg                 ; R5=$XXYY

        ld              a, 6
        ld              hl, _SCRN1 + 10 * 32 + 11
        call            WriteReg                 ; R6=$XXYY

        ld              a, 7
        ld              hl, _SCRN1 + 11 * 32 + 11
        call            WriteReg                 ; R7=$XXYY

        ; V-Blank and/or H-Blank interrupts are no doubt pending, so
        ; let's wait for the V-Blank period before re-enabling interrupts

.Wait_VB

        ldh             a, [rSTAT]      ; a <- LCDC Status
        and             $03
        cp              STATF_VB        ; are we in V-Blank ?
        jr              nz, .Wait_VB    ; no, so wait until we are

        xor             a
        ldh             [STIC_Display_Hi], a  ; display window only

        ei
                                        ; let the interrupts commence
.Wait_Release

        call            ReadJoy

        ld              a, [JoyDown]       ; a <- buttons currently held down
        or              a                  ; any held down ?
        jr              nz, .Wait_Release  ; yes, so wait until released

.Wait_Press

        call            ReadJoy

        ld              a, [JoyDown]                  ; a <- "downed" buttons
        cp              PADF_SELECT | PADF_A | PADF_B ; reset combo pressed ?
        jr              nz, .Wait_Press               ; nope, so keep looping

        di

        jp              Start

WriteReg:

        ld              [RegNo], a      ; set current register #

        ld              a, $0B          ; grey (palette #3), upper-tile bank
        ld              de, RegStr      ; set Source Address

        call            WriteStr        ; "R =$"

        ld              de, -3
        add             hl, de          ; position HL to fill register no.

        ; get the register number, put it in E, and call WriteHex

        ld              a, [RegNo]      ; a <- current register #
        ld              e, a            ; e = 4-bit value to display
        ld              a, $0B          ; grey (palette #3), upper-tile bank
        ld              c, 1            ; set #/digits to display = 1

        call            WriteHex        ; "Rx=$"

        inc             hl
        inc             hl              ; skip "=$"

        ; get the value stored in register, put it in DE, and call WriteHex

        push            hl              ; save position on _SCRN1

        ld              a, [RegNo]      ; a <- current register #
        add             a, a
        ld              l, a
        ld              h, R0 >> 8      ; hl -> Rx

        cp              7 << 1          ; is RegNo = 7 ?

        ld              a, [hl+]
        ld              d, [hl]
        ld              e, a            ; de <- [Rx]

        jr              nz, .SkipConv   ; no, so don't convert to "CP1610"

        srl             d
        rr              e

        dec             de

        ldh             a, [R7_Conv]
        add             d
        ld              d, a            ; de = [R7] (CP1610)

.SkipConv

        pop             hl              ; restore position on _SCRN1

        ld              a, $0B          ; grey (palette #3), upper-tile bank
        ld              c, 4            ; set #/digits to display = 4

        jp              WriteHex        ; "Rx=$XXYY"

INCLUDE "INTR.ASM"          ; VBlankFunc, HBlankFunc, UpdAttrib, UpdOneAttrib
;INCLUDE "RTPAT.ASM"

; **
; IH_515D:        Cycles BACKGROUND animations.
;
; Notes:          Also known as "BKGCYC" in DRACULA.SRC.
;
; Parameters:     * BC_RR = RomRecList      * BC_C1 = Convert #1
;                 * BC_RS = RomRecToSQPC    * BC_SR = SQPCToRomRec
;                 * BC_EA = EndAddr         * BC_CA = ContAddr
;
;                 * Must be set in advance.
;
; Modifies:       af, bc, de, hl
;

rIH_515D::

        ld              sp, SP_1        ; use STACK_1 for IH_515D

rBC_RR: ld              hl, SCRATCHPAD_RAM + $05D * 2  ; hl -> RomRecList
rBC_C1: ld              b, ( $4000 >> 1 - $5000 ) >> 8

rBC_Loop:

        ; get RomRecAddr, add RomRecOfs, and get NumPics

        ld              e, l
        ld              d, h            ; de -> RomRecAddr

        srl             d
        rr              e               ; de = RomRecAddr >> 1

        ld              a, [hl+]
        inc             l
        ld              h, [hl]         ; h <- HIGH of RomRecAddr (CP1610)
        ld              l, a            ; l <- LOW of RomRecAddr (CP1610)

        or              h               ; is RomRecAddr = 0 ?

        ld              c, 2            ; * outside loop ?
        add             hl, bc
        add             hl, hl          ; hl -> RomRec @ NumPics (Real)

        ld              c, [hl]         ; c <- NumPics

rBC_RS: ld              hl, ( SYSTEM_RAM + $0F1 * 2 ) - ( SCRATCHPAD_RAM + $05D * 2 ) >> 1 + 1
        add             hl, de          ; hl -> SQPC

        jr              z, rBC_SR       ; yes, so don't sequence animation

        ; ^ 23 cycles

        ; IMAGIC "SQPC"                                    nnnn  = PicNum
        ;                                                  Rrrrr = DlyRefresh
        ; 15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00  ddddd = Delay
        ;  m  ?  d  d  d  d  d  R  r  r  r  r  n  n  n  n  m     = Minus

        ld              a, [hl-]
        ld              e, [hl]         ; e <- rrrrnnnn (LOW byte of SQPC)
        ld              d, a            ; d <- m?dddddR (HIGH byte of SQPC)

        and             $3E             ; isolate Delay in bits 1-5
        jr              z, rBC_Zero     ; if ZY = 1, Delay was already = 0

        sub             $02             ; dec Delay
        jr              nz, rBC_NotZero ; if ZF = 0, Delay was in 2..31

        ; advance PicNum, make sure it's in 0..NumPics - 1, and update it

        ld              a, e            ; a = rrrrnnnn (LOW byte of SQPC)
        and             $0F             ;   = ....nnnn

        inc             a               ; inc PicNum

        bit             7, d            ; is Minus = 1 ?
        jr              z, .Cont1       ; no, so leave the PicNum inc'd

        sub             2               ; dec PicNum
        jr              nc, .Cont1      ; if CY = 0, PicNum was > 0

        ld              a, c            ; a = NumPics
        dec             a               ; conv NumPics to MaxPic
.Cont1
        cp              c               ; is PicNum < NumPics ?
        jr              c, .Cont2       ; yes, so don't wrap to zero

        xor             a               ; set PicNum = 0
.Cont2
        xor             e
        and             $0F             ; preserve PicNum
        xor             e               ; set PicNum

        ld              [hl], a         ; * store out LOW byte of SQPC

rBC_Zero:                               ; Delay was already = 0

        ; set Delay = DelayRefresh, and update Delay

        ld              a, d            ; a = m?dddddR (HIGH byte of SQPC)
        rrca                            ;   = Rm?ddddd (CY = R)

        ld              a, e            ; a = rrrrnnnn (LOW byte of SQPC)
        rra                             ;   = Rrrrrnnn
        rrca                            ;   = nRrrrrnn
        rrca                            ;   = nnRrrrrn

        ; ^ 6 cycles

rBC_NotZero:                            ; Delay was in 2..31

        xor             d
        and             $3E             ; preserve DelayRefresh
        xor             d               ; set Delay = DelayRefresh

        inc             l               ; * re-pos @ HIGH byte of SQPC

        ld              [hl], a         ; store out HIGH byte of SQPC

rBC_SR: ld              de, ( SCRATCHPAD_RAM + $05D * 2 ) >> 1 - ( SYSTEM_RAM + $0F1 * 2 ) + 1

        add             hl, de
        add             hl, hl          ; hl -> RomRecList

        ld              a, l
rBC_EA: cp              ( SCRATCHPAD_RAM + $067 * 2 ) & $FF
        jr              nz, rBC_Loop

        ld              bc, rROMB0 + cROMB0 & $FF

        di

rBC_CA: ld              sp, $42AA       ; return to DRACULA @ $5155

        reti

rIH_515D_End::

; **
; IH_52B3:        Add Velocities to Positions (both X & Y).
;
; Notes:          Also known as "ADDVEL" in DRACULA.SRC.
;
; Parameters:     * AV_O1 = ObjUpd1         * AV_XA = XAdj
;                 * AV_XP = XPos            * AV_O2 = ObjUpd2
;                 * AV_YA = YAdj            * AV_CA = ContAddr
;
;                 * AV_P1 = PosToFrac1      * AV_P2 = PosToFrac2
;                 * AV_DP = DcmlPt            AV_PA = PosAdj
;
;                 * Must be set in advance.
;
; Modifies:       af, bc, de, hl
;

rIH_52B3::

        ei

        ld              sp, SP_1        ; use STACK_1 for IH_52B3

rAV_O1: ld              a, [SCRATCHPAD_RAM + $004 * 2]  ; a <- ObjUpd
        rrca                                            ; is ObjUpd.0 = 1 ?

        ld              a, 0            ; assume ObjUpd.0 = 1 (XAdj = 0)

        jr              c, rAV_XAdj     ; yes, so set XAdj = 0

rAV_XA: ld              a, [SCRATCHPAD_RAM + $009 * 2]  ; a <- XAdj
rAV_XAdj:
        ld              [AV_PA + 1], a  ; set "PosAdj" to XAdj

rAV_XP: ld              hl, SYSTEM_RAM + $0125 * 2 + 1  ; hl -> XPos

        ld              bc, $FF << 8 + 8  ; c = 8   (Count)
                                          ; b = $FF (XPosMask)

        call            AV_Loop

rAV_O2: ld              a, [SCRATCHPAD_RAM + $004 * 2]  ; a <- ObjUpd
        rrca                                            ; is ObjUpd.0 = 1 ?

        ld              a, 0            ; assume ObjUpd.0 = 0 (YAdj = 0)

        jr              nc, rAV_YAdj    ; no, so set YAdj = 0

rAV_YA: ld              a, [SCRATCHPAD_RAM + $00D * 2]  ; a <- YAdj
rAV_YAdj:
        ld              [AV_PA + 1], a  ; set "PosAdj" to YAdj

        ld              bc, $7F << 8 + 8  ; c = 8   (Count)
                                          ; b = $7F (YPosMask)

        call            AV_Loop

        ld              bc, rROMB0 + cROMB0 & $FF

        di

rAV_CA: ld              sp, $45BA       ; return to DRACULA @ $52DD

        ret

rAV_Loop:

        ld              a, [hl-]        ; a <- HIGH byte of X/YPos

        and             $40             ; is X/YPos.14 (Link) = 1 ?

        push            hl              ; save ^X/YPos @ LOW byte

        jr              z, rAV_NoLink   ; no, so object not linked w/ prev.

        ; X/YPos.14 = 1, so cur. object is "linked" with prev. object

        dec             hl              ; pos @ HIGH byte of prev. X/YPos
        dec             l               ; pos @ LOW byte of prev. X/YPos

        ld              a, [hl]         ; a <- LOW byte of prev. X/YPos

rAV_P1: ld              de, ( ( SCRATCHPAD_RAM + $025 * 2 ) - ( SYSTEM_RAM + $125 * 2 ) << 1 ) & $FFFF + 6

        add             hl, hl
        add             hl, de          ; hl -> X/YFracVel @ X/YVel

        add             [hl]            ; add cur. X/YVel to prev. X/YPos

        pop             hl              ; rest. ^X/YPos @ LOW byte

        ld              e, [hl]         ; e <- cur. MO's X/YPos

        jr              rAV_UpdPos      ; update cur. MO's X/YPos

rAV_NoLink:

rAV_P2: ld              de, ( ( SCRATCHPAD_RAM + $025 * 2 ) - ( SYSTEM_RAM + $125 * 2 ) << 1 ) & $FFFF + 2

        add             hl, hl
        add             hl, de          ; hl -> X/YFracVel @ X/YVel

        ld              d, a            ; set HIGH byte of X/YVel = 0

rAV_DP: ld              a, [SCRATCHPAD_RAM + $016 * 2]  ; a <- DcmlPt
;        or              a                               ; is DcmlPt = 0 ?
        ld              e, a                            ; save DcmlPt in e

        ld              a, [hl-]        ; a <- X/YVel

        bit             7, a            ; is X/YVel.7 = 1 (Sign) ?
        jr              z, .NoSign      ; no, so X/YVel is not signed

        dec             d               ; d = $FF
.NoSign
        inc             e
        dec             e               ; is DcmlPt = 0 ?
        jr              z, .Skip        ; yes, so leave X/YVel as is
.Shift
        add             a, a
        rl              d               ; shift X/YVel left once

        dec             e               ; is DcmlPt = 0 ?
        jr              nz, .Shift      ; no, so keep shifting left
.Skip
        dec             l               ; pos @ X/YFrac

        add             [hl]            ; add LOW byte of X/YVel to X/YFrac
        ld              [hl], a         ; update X/YFrac

        pop             hl              ; rest. ^X/YPos @ LOW byte

        ld              a, [hl+]        ; a <- X/YPos
        ld              e, a            ; save a copy of X/YPos in e

        adc             d               ; add HIGH byte of X/YVel to X/YPos

        bit             7, [hl]         ; is X/YPos.15 = 1 ?
        jr              nz, rAV_WithBkg ; yes, so object moves w/ background

rAV_PA: sub             0               ; sub X/YAdj from X/YVel
rAV_WithBkg:
        dec             l               ; re-pos ^X/YPos @ LOW byte
rAV_UpdPos:
        and             b               ; isolate X/YPos (and w/ X/YPosMask)

        xor             e
        and             b               ; preserve X/YPos
        xor             e               ; set X/YPos

        ld              [hl+], a        ; update X/YPos

        inc             hl              ; skip HIGH byte of cur. X/YPos
        inc             l               ; skip LOW byte of next X/YPos

        dec             c               ; dec Count
        jr              nz, rAV_Loop

        ret

rIH_52B3_End::

ST_Y7:  ld              hl, MO + MO_RECSIZE * 7
        jr              STIC_YPosition

ST_Y6:  ld              hl, MO + MO_RECSIZE * 6
        jr              STIC_YPosition

ST_Y5:  ld              hl, MO + MO_RECSIZE * 5
        jr              STIC_YPosition

ST_Y4:  ld              hl, MO + MO_RECSIZE * 4
        jr              STIC_YPosition

ST_Y3:  ld              hl, MO + MO_RECSIZE * 3
        jr              STIC_YPosition

ST_Y2:  ld              hl, MO + MO_RECSIZE * 2
        jr              STIC_YPosition

ST_Y1:  ld              hl, MO + MO_RECSIZE * 1
        jr              STIC_YPosition

ST_Y0:  ld              hl, MO + MO_RECSIZE * 0

STIC_YPosition:                         ; STIC Register #8-#15 @ $0008-$000F

        ; $08 + n : Y position register, moving object #n.
        ;
        ;   15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
        ;    ?  ?  ?  ? ym xm ys yf  r  y  y  y  y  y  y  y

        ld              a, [hl]         ; a <- MO attribute byte #0

        xor             d
        and             $F3             ; clear YM and XM bits
        xor             d               ; set XM and YM bits
        or              $80             ; set ReAnalyze

        ld              [hl+], a        ; store out new MO attribute byte #0

        inc             l               ; skip X-Coord.

        ld              a, e            ; a = LOW byte of Y Position reg.
        and             $7F             ; isolate new Y-Coord.
        ld              [hl+], a        ; store out new Y-Coord.

        xor             e               ; isolate YRES in bit #7
        swap            a               ; put YRES in bit #3
        xor             d
        and             $FC             ; clear YSIZE and YFULL
        xor             d               ; set YSIZE and YFULL

        xor             [hl]            ; changed bits will be set
        and             $0B             ; did YRES, YSIZE, or YFULL change ?
        jr              z, .Exit        ; no, so just return

        xor             [hl]            ; set YRES, YSIZE, and/or YFULL
        or              $80             ; set ReCopy
        ld              [hl], a         ; store out new MO attribute byte #1

        ; YS and/or YF may have changed, so object may need color shading

        rr              l               ; l = ...mmm.1
        rr              l               ;   = 1...mmm.
        rr              l               ;   = 11...mmm (MO_FgCols @ MO #)

        and             $03             ; isolate YSIZE & YFULL in bits #0-#1
        dec             a               ; set bit #2 if neither set

        xor             [hl]
        and             $04
        xor             [hl]            ; bit #2 = YS << 1 + YS == 0
        or              $01             ; make sure ReCopy set

        ld              [hl], a         ; update MO_FgCols[ MO # ]
.Exit
        reti                            ; return, enabling interrupts

SECTION_39D3_End::

        PRINTT      "- SECTION_39D3: "
        PRINTV      1581 - ( SECTION_39D3_End - SECTION_39D3_Start )
        PRINTT      " bytes still available.\n"

;HOME_END:
;
;        PRINTT  "- INTVGB: "
;        PRINTV  HOME_END + 1
;        PRINTT  " bytes consumed by entire HOME section.\n"

SECTION "Sound_Code",CODE,BANK[4]

; ---------------------------------------------------------------------------
; Data in ROM that needs to be aligned on special multiples should go here
; ---------------------------------------------------------------------------

ayPerToAudFreq::                ; $4000-$5FFF, aligned on multiples of 512

        INCBIN  "PERIODS.BIN"   ; conversion table for periods 0-4095

ayEnabledToAudTerm::            ; $6000-$61FF, aligned on multiples of 256

CTR         SET     0

            REPT    512         ; iterate thru possible combos of 9 bits

AUDTERM     SET     0

TONA_ENA    SET     ( ( CTR & $0001 ) >> 0 ) ^ 1
TONB_ENA    SET     ( ( CTR & $0002 ) >> 1 ) ^ 1
TONC_ENA    SET     ( ( CTR & $0004 ) >> 2 ) ^ 1

NOIA_ENA    SET     ( ( CTR & $0008 ) >> 3 ) ^ 1
NOIB_ENA    SET     ( ( CTR & $0010 ) >> 4 ) ^ 1
NOIC_ENA    SET     ( ( CTR & $0020 ) >> 5 ) ^ 1

CHNA_ENV    SET     ( ( CTR & $0040 ) >> 6 ) ^ 1
CHNB_ENV    SET     ( ( CTR & $0080 ) >> 7 ) ^ 1
CHNC_ENV    SET     ( ( CTR & $0100 ) >> 8 ) ^ 1

            ; SO1 - Sound OutPut #1 (Left Channel)

            IF      ( TONA_ENA == 1 ) & ( CHNA_ENV != 1 )
AUDTERM     SET     AUDTERM | $01  ; Output Sound #1 to SO1
            ENDC

            IF      ( TONB_ENA == 1 ) & ( CHNB_ENV != 1 )
AUDTERM     SET     AUDTERM | $02  ; Output Sound #2 to SO1
            ENDC

            IF      ( TONC_ENA == 1 ) & ( CHNC_ENV != 1 )
AUDTERM     SET     AUDTERM | $04  ; Output Sound #3 to SO1
            ENDC

            IF      ( ( NOIA_ENA == 1 ) & ( CHNA_ENV != 1 ) ) | ( ( NOIB_ENA == 1 ) & ( CHNB_ENV != 1 ) ) | ( ( NOIC_ENA == 1 ) & ( CHNC_ENV != 1 ) )
AUDTERM     SET     AUDTERM | $08  ; Output Sound #4 to SO1
            ENDC

            ; SO2 - Sound Output #2 (Right Channel)

            IF      TONA_ENA & CHNA_ENV == 1
AUDTERM     SET     AUDTERM | $10  ; Output Sound #1 to SO2
            ENDC

            IF      TONB_ENA & CHNB_ENV == 1
AUDTERM     SET     AUDTERM | $20  ; Output Sound #2 to SO2
            ENDC

            IF      TONC_ENA & CHNC_ENV == 1
AUDTERM     SET     AUDTERM | $40  ; Output Sound #3 to SO2
            ENDC

            IF      ( NOIA_ENA & CHNA_ENV ) | ( NOIB_ENA & CHNB_ENV ) | ( NOIC_ENA & CHNC_ENV ) == 1
AUDTERM     SET     AUDTERM | $80  ; Output Sound #4 to SO2
            ENDC

            DB      AUDTERM

CTR         SET     CTR + 1

            ENDR

ayEnvTypeToHandler::            ; $6200-$62FF, aligned on multiples of 256

        REPT            8               ; avoid AND operation on ayEnvType

        DW              TF_EnvEnd_0     ; CONT: 0 ATAK: 0 ALTR: 0 HOLD: 0
        DW              TF_EnvEnd_1     ; CONT: 0 ATAK: 0 ALTR: 0 HOLD: 1
        DW              TF_EnvEnd_2     ; CONT: 0 ATAK: 0 ALTR: 1 HOLD: 0
        DW              TF_EnvEnd_3     ; CONT: 0 ATAK: 0 ALTR: 1 HOLD: 1
        DW              TF_EnvEnd_4     ; CONT: 0 ATAK: 1 ALTR: 0 HOLD: 0
        DW              TF_EnvEnd_5     ; CONT: 0 ATAK: 1 ALTR: 0 HOLD: 1
        DW              TF_EnvEnd_6     ; CONT: 0 ATAK: 1 ALTR: 1 HOLD: 0
        DW              TF_EnvEnd_7     ; CONT: 0 ATAK: 1 ALTR: 1 HOLD: 1
        DW              TF_EnvEnd_8     ; CONT: 1 ATAK: 0 ALTR: 0 HOLD: 0
        DW              TF_EnvEnd_9     ; CONT: 1 ATAK: 0 ALTR: 0 HOLD: 1
        DW              TF_EnvEnd_A     ; CONT: 1 ATAK: 0 ALTR: 1 HOLD: 0
        DW              TF_EnvEnd_B     ; CONT: 1 ATAK: 0 ALTR: 1 HOLD: 1
        DW              TF_EnvEnd_C     ; CONT: 1 ATAK: 1 ALTR: 0 HOLD: 0
        DW              TF_EnvEnd_D     ; CONT: 1 ATAK: 1 ALTR: 0 HOLD: 1
        DW              TF_EnvEnd_E     ; CONT: 1 ATAK: 1 ALTR: 1 HOLD: 0
        DW              TF_EnvEnd_F     ; CONT: 1 ATAK: 1 ALTR: 1 HOLD: 1

        ENDR

CurToPer::                         ; $6300-$6377, aligned on multiples of 256

CUR_FRAMES  SET     0
            REPT    120            ; up to 198% supported

PERCENTAGE  SET     ( 100 * CUR_FRAMES ) / 60

;            PRINTT  "Percentage ({CUR_FRAMES}/60): {PERCENTAGE}%\n"

TENS        SET     PERCENTAGE / 10
ONES        SET     PERCENTAGE + -1 * TENS * 10

;            PRINTT  "Tens: {TENS}\n"
;            PRINTT  "Ones: {ONES}\n"

            IF      TENS > 9
TENS        SET     TENS + -10
            ENDC

            DB      ( TENS << 4 + ONES ) & $FF

CUR_FRAMES  SET     CUR_FRAMES + 1
            ENDR

ayVolToAudVol_ROM::             ; $6300-$630F, no special alignment

        DB              $0F, $0F, $0F, $0F, $0F, $0F, $07, $07
        DB              $07, $07, $17, $17, $27, $37, $57, $77

ayPerToAudPoly_ROM::            ; $6310-$632F, must follow ayVolToAudVol_ROM

        DB              $45, $02, $05, $07, $15, $16, $17, $24
        DB              $25, $25, $26, $26, $27, $34, $34, $34
        DB              $35, $35, $35, $36, $36, $36, $36, $37
        DB              $37, $37, $44, $44, $44, $44, $44, $45

ayVolToAudEnv_ROM::             ; $6330-$633F, must follow ayPerToAudPoly_ROM

;        DB              $00, $10, $20, $30, $40, $50, $60, $70
;        DB              $80, $90, $A0, $B0, $C0, $D0, $E0, $F0

;        DB              $00, $00, $00, $10, $10, $10, $10, $20
;        DB              $20, $30, $40, $50, $70, $80, $B0, $F0

;        DB              $00, $00, $10, $10, $10, $10, $20, $20
;        DB              $30, $40, $50, $60, $80, $B0, $F0, $F0

         DB              $00, $10, $10, $10, $10, $20, $20, $30
         DB              $40, $50, $60, $80, $B0, $F0, $F0, $F0

ayRegToHandler_ROM::            ; $6340-$636D, must follow ayVolToAudEnv_ROM

            DW          ET_Exit           ; $01E9
            DW          ET_Exit           ; $01EA
            DW          ET_Exit           ; $01EB
            DW          ET_Exit           ; $01EC
            DW          ET_Exit           ; $01ED
            DW          ET_Exit           ; $01EE
            DW          ET_Exit           ; $01EF (see AP_ScratchPad)

            DW          AY8914_TonAPerLo  ; AY8914 Register #0 @ $01F0
            DW          AY8914_TonBPerLo  ; AY8914 Register #1 @ $01F1
            DW          AY8914_TonCPerLo  ; AY8914 Register #2 @ $01F2
            DW          AY8914_EnvPerLo   ; AY8914 Register #3 @ $01F3
            DW          AY8914_TonAPerHi  ; AY8914 Register #4 @ $01F4
            DW          AY8914_TonBPerHi  ; AY8914 Register #5 @ $01F5
            DW          AY8914_TonCPerHi  ; AY8914 Register #6 @ $01F6
            DW          AY8914_EnvPerHi   ; AY8914 Register #7 @ $01F7
            DW          AY8914_Mixer      ; AY8914 Register #8 @ $01F8
            DW          AY8914_NoiPer     ; AY8914 Register #9 @ $01F9
            DW          AY8914_EnvType    ; AY8914 Register #10 @ $01FA
            DW          AY8914_ChnAVol    ; AY8914 Register #11 @ $01FB
            DW          AY8914_ChnBVol    ; AY8914 Register #12 @ $01FC
            DW          AY8914_ChnCVol    ; AY8914 Register #13 @ $01FD
            DW          AY8914_HandRight  ; AY8914 Register #14 @ $01FE
            DW          AY8914_HandLeft   ; AY8914 Register #15 @ $01FF

SndInitData::                   ; $636E-$6384, must follow ayPerToAudPoly_ROM

        IF              CLICK_REDUCTION == 1

;        DB              $00, $80, $F0, $00, $80  ; Sound #1, $FF10-$FF14
;        DB              $00, $80, $F0, $00, $80  ; Sound #2, $FF15-$FF19
;        DB              $80, $00, $20, $00, $80  ; Sound #3, $FF1A-$FF1E
;        DB              $00, $00, $F0, $00, $80  ; Sound #4, $FF1F-$FF23

        ;                                     ^ THIS IS THE PROBLEM !?!?!?!

        DB              $00, $80, $08, $00, $00  ; Sound #1, $FF10-$FF14
        DB              $00, $80, $08, $00, $00  ; Sound #2, $FF15-$FF19
        DB              $80, $00, $20, $00, $00  ; Sound #3, $FF1A-$FF1E
        DB              $00, $00, $08, $00, $00  ; Sound #4, $FF1F-$FF23

        DB              $77, $00, $80            ; Sound Control, $FF24-$FF26

        ELSE

        DB              $00, $80, $00, $00, $80  ; Sound #1, $FF10-$FF14
        DB              $00, $80, $00, $00, $80  ; Sound #2, $FF15-$FF19
        DB              $80, $00, $20, $00, $80  ; Sound #3, $FF1A-$FF1E
        DB              $00, $00, $00, $00, $80  ; Sound #4, $FF1F-$FF23

        DB              $77, $00, $80            ; Sound Control, $FF24-$FF26

        ENDC

        INCLUDE         "SOUND.ASM"     ; TimerFunc, AY8914_Init, etc.

SECTION "Background_Code",CODE,BANK[2]

; ---------------------------------------------------------------------------
; Data in ROM that needs to be aligned on special multiples should go here
; ---------------------------------------------------------------------------

Case2Tab::                         ; $4000-$4FFF, aligned on multiples of 4096
IF      NEED_PRIORITIES == 1
        INCBIN  "CAS2TABP.BIN"
ELSE
        INCBIN  "CAS2TAB.BIN"
ENDC

CNToBDD::                          ; $5000-$51FF, aligned on multiples of 256

            ; CNToBDD - page #1

CTR         SET     0
            REPT    256

            ; ??n??NNN

CARDNUM     SET     ( CTR & $07 ) << 5 + ( ( CTR ^ ( CARD >> 8 ) ) >> 3 ) & $1F
OUTB        SET     ( BDD_TileNdx + CARDNUM + ( 32 - 20 ) * ( CARDNUM / 20 ) ) & $FF

;            IF      CTR == $0
;            PRINTT  "LOW -> CardNum: {CARDNUM} OutB: {OUTB}\n"
;            ENDC

            DB      OUTB

CTR         SET     CTR + 1
            ENDR

            ; CNToBDD - page #2

CTR         SET     0
            REPT    256

            ; ??n??NNN

CARDNUM     SET     ( CTR & $07 ) << 5 + ( ( CTR ^ ( CARD >> 8 ) ) >> 3 ) & $1F
OUTB        SET     ( BDD_TileNdx + CARDNUM + ( 32 - 20 ) * ( CARDNUM / 20 ) ) >> 8

;            IF      CTR == $0
;            PRINTT  "HIGH -> CardNum: {CARDNUM} OutB: {OUTB}\n"
;            ENDC

            DB      OUTB

CTR         SET     CTR + 1
            ENDR

SqGrafx:                           ; $5200-$52FF, fixed here

CTR         SET     0
            REPT    256

            IF      CTR & $80 == 0

OUTB        SET     $80            ; !1234...

            IF      CTR & $40 == 0

            ; Sq12ToAddr: ..222111

SQ1COL      SET     ( CTR >> 0 ) & $07
SQ2COL      SET     ( CTR >> 3 ) & $07

            IF      SQ1COL != 7
OUTB        SET     OUTB | $40     ; flag that square #1 is solid
            ENDC

            IF      SQ2COL != 7
OUTB        SET     OUTB | $20     ; flag that square #2 is solid
            ENDC

            ELSE

            ; Sq34ToAddr: .!444333

SQ3COL      SET     ( CTR >> 0 ) & $07  ; ( ( CTR >> 6 ) & $03 ) | ( ( CTR << 2 ) & $04 )
SQ4COL      SET     ( CTR >> 3 ) & $07  ; ( ( CTR >> 3 ) & $04 ) | ( ( CTR >> 1 ) & $03 )

            IF      SQ3COL != 7
OUTB        SET     OUTB | $10     ; flag that square #3 is solid
            ENDC

            IF      SQ4COL != 7
OUTB        SET     OUTB | $08     ; flag that square #4 is solid
            ENDC

            ENDC

            ELSE

            ; SqGrafx: !1234...

            IF      CTR & $07 < 4

            ; do squares 1 & 2 (in rows 0-3)

LSQSOLID    SET     ( CTR >> 6 ) & $01
RSQSOLID    SET     ( CTR >> 5 ) & $01

            ELSE

            ; do squares 3 & 4 (in rows 4-7)

LSQSOLID    SET     ( CTR >> 4 ) & $01
RSQSOLID    SET     ( CTR >> 3 ) & $01

            ENDC

OUTB        SET     0

            IF      LSQSOLID == 1
OUTB        SET     OUTB | $F0
            ENDC

            IF      RSQSOLID == 1
OUTB        SET     OUTB | $0F
            ENDC

            ENDC

            DB      OUTB

CTR         SET     CTR + 1
            ENDR

MasterLevel::                      ; $5300-$53FF, aligned on multiples of 256
        INCBIN  "MASTRLVL.BIN"

Case4Tab::                         ; $5400-$55FF, aligned on multiples of 512
        INCBIN  "CASE4TAB.BIN"

SquareGrafx::                      ; $5600-$65FF, no special alignment
        INCBIN  "CLRDSQUA.BIN"

UniqueTab::                        ; $6600-$67FF, aligned on multiples of 256
        INCBIN  "UNQECLRS.BIN"

Case1Tab::                         ; $6800-$68FF, aligned on multiples of 1024

IF      NEED_PRIORITIES == 1
        INCBIN  "CAS1TABP.BIN"
ELSE
        INCBIN  "CAS1TAB.BIN"
ENDC

BKG::
        INCLUDE "BKG.ASM"

; ---------------------------------------------------------------------------
;                    .::** End of HOME Section **::.
; ---------------------------------------------------------------------------

SECTION "Misc_Code",CODE,BANK[1]

; ---------------------------------------------------------------------------
; Data in ROM that needs to be aligned on special multiples should go here
; ---------------------------------------------------------------------------

OBJPAL:
        INCBIN  "OBJPAL.BIN"    ; @ $4000-$47FF (fixed at this address)

DoubleNybble_Lo:                ; @ $4800-$48FF (aligned on multiples of 256)

        REPT    16
        DB      $00             ; $0 = 00000000b
        DB      $03             ; $1 = 00000011b
        DB      $0C             ; $2 = 00001100b
        DB      $0F             ; $3 = 00001111b
        DB      $30             ; $4 = 00110000b
        DB      $33             ; $5 = 00110011b
        DB      $3C             ; $6 = 00111100b
        DB      $3F             ; $7 = 00111111b
        DB      $C0             ; $8 = 11000000b
        DB      $C3             ; $9 = 11000011b
        DB      $CC             ; $A = 11001100b
        DB      $CF             ; $B = 11001111b
        DB      $F0             ; $C = 11110000b
        DB      $F3             ; $D = 11110011b
        DB      $FC             ; $E = 11111100b
        DB      $FF             ; $F = 11111111b
        ENDR

DoubleNybble_Hi:                ; @ $4900-$49FF (aligned on multiples of 256)

        REPT    16
        DB      $00             ; $0 = 00000000b
        ENDR

        REPT    16
        DB      $03             ; $1 = 00000011b
        ENDR

        REPT    16
        DB      $0C             ; $2 = 00001100b
        ENDR

        REPT    16
        DB      $0F             ; $3 = 00001111b
        ENDR

        REPT    16
        DB      $30             ; $4 = 00110000b
        ENDR

        REPT    16
        DB      $33             ; $5 = 00110011b
        ENDR

        REPT    16
        DB      $3C             ; $6 = 00111100b
        ENDR

        REPT    16
        DB      $3F             ; $7 = 00111111b
        ENDR

        REPT    16
        DB      $C0             ; $8 = 11000000b
        ENDR

        REPT    16
        DB      $C3             ; $9 = 11000011b
        ENDR

        REPT    16
        DB      $CC             ; $A = 11001100b
        ENDR

        REPT    16
        DB      $CF             ; $B = 11001111b
        ENDR

        REPT    16
        DB      $F0             ; $C = 11110000b
        ENDR

        REPT    16
        DB      $F3             ; $D = 11110011b
        ENDR

        REPT    16
        DB      $FC             ; $E = 11111100b
        ENDR

        REPT    16
        DB      $FF             ; $F = 11111111b
        ENDR

GROM_8:                         ; @ $4A00-$51FF (aligned on multiples of 256)
        INCBIN  "GROM_8.BIN"    ; mapped to $3000-$37FF on the real INTV

;CardToTileTab:                  ; @ $5200-$53FF (aligned on multiples of 512)
;
;BASEADR         SET         BACKTAB >> 8 + 2
;
;CTR_1           SET         0
;                REPT        256
;
;CTR_2           SET         ( CTR_1 - BASEADR ) & $FF
;
;CARD_NO         SET         ( CTR_2 & $07 ) << 5 + CTR_2 >> 3
;
;                DW          _SCRN0 + CARD_NO + ( 32 - 20 ) * ( CARD_NO / 20 )
;
;CTR_1           SET         CTR_1 + 1
;                ENDR
;
;CSTab:                          ; @ $5400-$541F (aligned on multiples of 512)
;
;        DW      $0000, $00F0, $F000, $F0F0      ; 0-3
;        DW      $000F, $00FF, $F00F, $F0FF      ; 4-7
;        DW      $0F00, $0FF0, $FF00, $FFF0      ; 8-11
;        DW      $0F0F, $0FFF, $FF0F, $FFFF      ; 12-15
;
;PAD_5   DS      $E0

MISC:
        INCLUDE "COPYVRAM.ASM"  ; includes CopyToVRAM
        INCLUDE "INIT.ASM"      ; includes InitBackTab, InitPalTab
        INCLUDE "MISC.ASM"
        INCLUDE "CONTROLS.ASM"
        INCLUDE "IH.ASM"        ; IMAGIC Helper

;        INCLUDE "CPU_RAM.ASM"

stRegToHandler_ROM:

            DW          ST_X0, ST_X1, ST_X2, ST_X3
            DW          ST_X4, ST_X5, ST_X6, ST_X7  ; STIC Register $00-$07

            DW          ST_Y0, ST_Y1, ST_Y2, ST_Y3
            DW          ST_Y4, ST_Y5, ST_Y6, ST_Y7  ; STIC Register $08-$0F

            DW          ST_A0, ST_A1, ST_A2, ST_A3
            DW          ST_A4, ST_A5, ST_A6, ST_A7  ; STIC Register $10-$17

            DW          ST_RI, ST_RI, ST_RI, ST_RI
            DW          ST_RI, ST_RI, ST_RI, ST_RI  ; STIC Register $18-$1F

            DW          STIC_DisplayEnable          ; STIC Register $20
            DW          STIC_GraphicsMode           ; STIC Register $21

            DW          ST_RI, ST_RI, ST_RI, ST_RI
            DW          ST_RI, ST_RI                ; STIC Register $22-$27

            DW          ST_C0, ST_C1, ST_C2, ST_C3  ; STIC Register $28-$2B

            DW          STIC_BorderColor            ; STIC Register $2C

            DW          ST_RI, ST_RI, ST_RI         ; STIC Register $2D-$2F

            DW          STIC_HorizontalDelay        ; STIC Register $30
            DW          STIC_VerticalDelay          ; STIC Register $31
            DW          STIC_BlockOut               ; STIC Register $32

            REPT        77
            DW          ST_RI                       ; STIC Register $33-$7F
            ENDR

BitCntToMask_ROM:

COUNT       SET     1
            REPT    16

BITMASK     SET     0
            REPT    COUNT

BITMASK     SET     BITMASK << 1
BITMASK     SET     BITMASK | $01
            ENDR

            DW      BITMASK

COUNT       SET     COUNT + 1
            ENDR

; ---------------------------------------------------------------------------
;                    .::** End of MISC Section **::.
; ---------------------------------------------------------------------------

SECTION "Exec_ROM",CODE,BANK[5]

EXEC:                           ; $4000-$5FFF, aligned on multiples of 512
IF      GAME_SPLITS > 0
        INCBIN  "EXEC.IG1"
ELSE
EPAD_1  DS      $2000
ENDC

GROM_16:                        ; $6000-$6FFF, aligned on multiples of 512
        INCBIN  "GROM_16.BIN"

BKTCODE:                        ; $7000-$71EF, aligned on multiples of 512
IF      GAME_SPLITS > 0
        INCBIN  "BKTCODE.IG1"
ELSE
EPAD_2  DS      $01D0
ENDC

EPAD_0  DS      $30

EMPTY:                          ; $7200-$73FF, aligned on multiples of 512
        REPT    256 * 2
        DB      $FF
        ENDR

HDToHndlr:                      ; $7400-$75FF, aligned on multiples of 512
        INCLUDE "GETHAND.INC"

; **
; EXEC_11AB:  Copy the STIC collision registers to [$107-$10E] and the
;             background and border collision bits (combined) to [$10F]
;             and [$110].
;
; Notes:      Called from EXEC_1154 only.
;

EXEC_11AB::

        ld              hl, STIC + $018 * 2
        ld              de, SCRATCHPAD_RAM + $007 * 2

        ld              b, $80          ; bit #7 = end-of-registers marker
.Loop
        ld              a, [hl+]        ; a <- LOW byte of collision reg.

        ld              [de], a         ; store out to [$107..$10E]
        inc             e
        inc             e

        ld              a, [hl+]        ; a <- HIGH byte of collision reg.

        rrca                            ; CY <- collision w/ bkg ON bits
        rr              c
        rrca                            ; CY <- collision w/ screen edges
        rr              b

        jr              nc, .Loop

        ld              l, e
        ld              h, d            ; hl -> SCRATCHPAD_RAM[ $00F * 2 ]

        ld              a, c
        ld              [hl+], a        ; [$10F] <- collision w/ bkg ON bits
        inc             l
        ld              [hl], b         ; [$110] <- collision w/ screen edges

        ld              bc, rROMB0 + cROMB0 & $FF

        di                              ; disable interrupts to restore stack

        ld              sp, $4272       ; return @ $1139

        RETURN          0               ; return, enabling interrupts

; **
; EXEC_1154:   Update the STIC registers with the MO records stored at
;              [$31D-$35E], or copy the STIC's collision registers to
;              [$107-$10E] (also [$10F] and [$110] for the background and
;              border collision data) depending upon the value in R1.
;
; Parameters:  if R1 = 2, the STIC registers for MO #s 0 - 3 are updated.
;              if R1 = 1, the STIC registers for MO #s 4 - 7 are updated.
;              if R1 = 0, the STIC's collision registers are copied to
;                $107 - $10E, $10F, and $110.
;
; Modifies:    All registers.
;

EXEC_1154::

        ld              sp, SP_1        ; use STACK_1 for EXEC_1154

        ei                              ; make sure interrupts enabled

        ld              a, [R1]         ; a <- LOW byte of [R1]
        or              a               ; is [R1] = 0 ?
        jr              z, EXEC_11AB    ; yes, so copy collision registers

        dec             a               ; is [R1] = 1 ?
        jr              nz, .Exit       ; no, so nothing to do

        ld              hl, UDB_Flag

        cp              [hl]            ; is UDB_Flag = 0 ?
        jr              nz, .UDB_Update ; * no, so update UDB_CardNdx

        ld              c, a            ; start with MO # 0

.Upd_Loop

        ld              a, c            ; a = MO #
        swap            a               ;   = MO # * 16
        add             $01D << 1 + 1   ;   = MO # * 16 + $01D << 1 + 1

        ld              l, a
        ld              h, ( SYSTEM_RAM + $100 * 2 ) >> 8

        ; first handle the update to the STIC's X-Position register

        ld              b, l            ; save MO record @ HIGH of Attribute

        ld              a, [hl+]        ; a <- HIGH byte of Attribute word
        rra
        rrca
        rrca                            ; put XS, V, and C in bits 0-2
        and             $07             ; * and isolate them
        ld              d, a

        ld              a, [hl+]        ; a <- LOW byte of COLLTAB address

        or              [hl]            ; is COLLTAB = 0 ?
        ld              a, c            ; a = MO #
        jr              z, .OffScreen   ; yes, so put MO off screen (X = 255)

        inc             l               ; pos @ LOW byte of X-Position 8.8
        inc             l               ; pos @ HIGH byte of X-Position 8.8

        ld              e, [hl]         ; e <- HIGH byte of X-Position 8.8

        ld              hl, SCRATCHPAD_RAM + $004 * 2

        cp              [hl]            ; is MO in use (# in 0..[$104] - 1)
        jr              c, .OnScreen    ; yes, so keep X-position (on-screen)
.OffScreen
        ld              e, $FF          ; position off screen (set X = 255)
.OnScreen
        call            .STIC_Write     ; STIC X-Position <- de

        ; now handle the update to the STIC's Y-Position register

        ld              l, b
        ld              h, ( SYSTEM_RAM + $0100 * 2 ) >> 8

        ld              a, [hl-]        ; a <- HIGH byte of Attribute word
        ld              b, [hl]         ; b <- LOW byte of Attribute word

        ld              de, 7
        add             hl, de          ; pos MO record @ HIGH of Y-Pos.

        and             e               ; isolate YM, XM, and YS in bits 0-2
        ld              d, a            ; save them in d

        ld              a, b            ; a = LOW byte of Attribute word
        add             a, a            ; put YFULL in CY
        rl              d               ; d = HIGH byte of STIC Y-Position
        add             a, a            ; put YRES in CY

        ld              a, [hl]         ; a <- HIGH byte of Y-Position 8.8
        rla                             ; put YRES in bit #0
        rrca                            ; put YRES in bit #7
        ld              e, a            ; e = LOW byte of STIC Y-Position

        ld              a, c            ; a = MO #
        add             $08             ; pos @ Y-Position register

        call            .STIC_Write     ; STIC Y-Position <- de

        ; now handle the update to the STIC's Attribute register

        ld              l, c                 ; l = MO #
        sla             l                    ;   = MO # * 2
        ld              h, UDB_CardNdx >> 8  ; hl -> UDB_CardNdx[ MO # * 2 ]

        ld              a, b            ; a = LOW byte of Attribute word
        and             $07             ; isolate fff in bits 0-2
        or              [hl]            ; put LSBs of CardNdx in bits 3-7
        ld              e, a            ; e = LOW byte of STIC Attribute

        inc             l               ; pos @ MSB of CardNdx

        ld              a, b            ; a = LOW byte of Attribute word
        and             $30             ; isolate p and F in bits 4-5
        sub             ( GRAM_8 >> 8 - $08 )  ; set GRAM toggle (forced on)
        add             [hl]                   ; put MSB of CardNdx in bit #0
        ld              d, a            ; d = LOW byte of STIC Attribute

        ld              a, c            ; a = MO #
        add             $10             ; pos @ MO's Attribute register

        call            .STIC_Write     ; update Attribute register

        inc             c               ; advance to next MO #

        bit             3, c            ; is MO # = 8 ?
        jr              z, .Upd_Loop    ; no, so keep looping

        ; finally, clear the STIC collision registers for MO #s 0-7

        ld              hl, STIC + $0018 * 2
        xor             a
.Clr_Loop
        ld              [hl+], a        ; clear LOW byte of collision reg.
        ld              [hl+], a        ; clear HIGH byte of collision reg.
        dec             c
        jr              nz, .Clr_Loop

        ld              bc, rROMB0 + cROMB0 & $FF
.Exit
        di                              ; disable interrups to restore stack

        ld              sp, $4272       ; return @ $1139

        RETURN          0               ; return, enabling interrupts

.UDB_Update

        dec             [hl]            ; set UDB_Flag = 0

        call            UDB_Update      ; UDB_CardNdx <- 8 GRAM offsets

        ld              c, 0            ; start with MO # 0

        jr              .Upd_Loop

.STIC_Write

        ld              h, STIC >> 8    ; hl -> STIC[ c * 2 ]

        add             a, a
        ld              l, a

        ld              a, e            ; a = LOW byte of "new" register
        cp              [hl]            ; same as LOW byte of old register ?
        ld              [hl+], a        ; store out LOW of "new" register
        ld              a, d            ; a = HIGH byte of "new" register
        jr              nz, .Diff       ; no, so have to process the write

        cp              [hl]            ; same as HIGH byte of old register ?
        ret             z               ; yes, so don't bother to process it
.Diff
        ld              [hl-], a        ; store out HIGH of "new" register

        ld              h, stRegToHandler >> 8

        ld              a, [hl+]
        ld              h, [hl]
        ld              l, a

        jp              [hl]            ; jump to appropriate STIC handler

; **
; EXEC_11C0:  Copy the bitmaps of an MO pair, specified by bits 2-4 of
;	      [$111] (with double resolution bits in 0-1 for the even
;	      and odd MO, respectively), with the picture number of the
;	      even MO in [$112] and the picture number of the odd MO
;	      in [$113].
;
; Parameters: None
;
; Modifies:   All registers.
;

EXEC_11C0::

        ld              sp, SP_1        ; use STACK_1 for EXEC_11C0

        ; first, check if our UDB info is up to date, if not call UDB_Update

        ld              a, [UDB_Flag]
        or              a               ; is UDB_Flag = 0 (clear CY) ?
        jr              nz, .UDB_Update ; no, so need to update UDB_CardNdx
.Start

        ; while in WRAM #1, copy a few SCRATCHPAD_RAM locations to WRAM #0

        ld              hl, SCRATCHPAD_RAM + $013 * 2

        ld              a, [hl-]        ; a <- [$113] (odd MO's pic #)
        dec             l
        ld              [ehSP113], a    ; store it to WRAM bank #0 location

        ld              e, [hl]         ; e <- [$112] (even MO's pic #)
        dec             l
        dec             l

        ld              a, [hl]         ; a <- [$111] (MO # + res. bits)
        rra                             ; put even MO's YRES bit in CY
        ld              [ehSP111], a    ; store it to WRAM bank #0 location

        ; ^ 21 cycles

        ; now, copy the bitmap data for the even numbered MO (0, 2, 4, or 6)

        rl              d               ; put YRES bit for even MO in bit #0
        and             $0E             ; isolate MO # in bits 1-3
        ld              l, a            ; and put it in l

        ld              a, 4            ; a = #/WRAM bank where OBJGRAM is
        ldh             [cSVBK], a      ; [cSVBK] <- WRAM bank # (for ISRs)
        ldh             [rSVBK], a      ; [rSVBK] <- WRAM bank #

        ld              a, e            ; a <- even MO's picture #

        or              a               ; is even MO's picture # = 0 ?
        call            nz, .Handle_MO  ; no, so MO needs to be handled

        ; now, copy the bitmap data for the odd numbered MO (1, 3, 5, or 7)

        ld              a, [ehSP111]    ; a <- MO # + res. bits

        rrca                            ; put YRES bit for odd MO in CY
        rl              d               ; put YRES for even MO in bit #0
        inc             a               ; make MO # odd (1, 3, 5, or 7)
        add             a, a            ; put MO # in bits 1-3
        ld              l, a            ; and put it in l

        ld              a, [ehSP113]    ; a <- odd MO's picture #

        or              a               ; is odd MO's picture # = 0 ?
        call            nz, .Handle_MO  ; no, so MO needs to be handled

        ld              a, 1            ; a = #/WRAM bank where R0, etc. is
        ldh             [cSVBK], a      ; [cSVBK] <- WRAM bank # (for ISRs)
        ldh             [rSVBK], a      ; [rSVBK] <- WRAM bank #

        ld              bc, rROMB0 + cROMB0 & $FF

        di                              ; disable interrupts to restore stack

        ld              sp, $4292       ; return to $1149

        RETURN          0               ; return, enabling interrupts

.UDB_Update

        ; this may occur literally only once during the entire game, so it's
        ; better to do a relative jump (above) instead of a call w/ condition

        xor             a
        ld              [UDB_Flag], a   ; set UDB_Flag = 0

        call            UDB_Update      ; update our "local" UDB information

        ; NOTE: CF must = 0 at this point (see UDB_Update)

        jr              .Start

.Handle_MO

        ; a = picture # (1-255)
        ; l = MO # * 2

        ; multiply the picture # by 8 and add it to OBJGRAM's base address

        dec             a               ; a = NNNnnnn (PicNum in 0..254)
        rlca                            ;   = NNnnnnN
        rlca                            ;   = NnnnnNN
        rlca                            ;   = nnnnNNN
        ld              b, a            ; save a copy in b

 	and		$F8		;   = nnnn... (PicNum's LSBs)
        ld		c, a

 	xor		b		;   = ....NNN (PicNum's MSBs)
        add		OBJGRAM >> 8
        ld		b, a  		; bc -> OBJGRAM[ PicNum * 8 ]

        ; get the GRAM offset from UDB and set up GRAM_8 & GRAM_16 pointers

        ld		h, UDB_CardNdx >> 8  ; hl -> UDB_CardNdx[ MO # * 2 ]

        ld              a, [hl+]
        ld              h, [hl]
        ld              l, a            ; hl -> GRAM_8

        add             a, a            ; put bit #4 of CardNdx in CY
        ld              e, a            ; e = LOW byte of GRAM_16 address

        ld              a, h
        rla                             ; put bit #4 of CardNdx in bit #0
        add             ( GRAM_16 - GRAM_8 << 1 ) >> 8 & $FF

        bit             0, d            ; is YRES = 1 ?

        ld              d, a            ; de -> GRAM_16

        jr              nz, .DblCopy    ; yes, so jump to "double" copy rt.

        ; -------------------------------------------------------------------
        ;                  SngCopy Routines (when YRES = 0)
        ; -------------------------------------------------------------------

        ; first, test if the image already exists at this loc. in GRAM

LN_NUM	SET		0		; set LineNum = 0
	REPT		7		; do the first 7 lines (0-6)

LN      EQUS            STRSUB( "{LN_NUM}", 2, 1 )
LABEL	EQUS		".SngCopy_{LN}"

        ld              a, [bc]         ; a <- bitmap data from OBJGRAM
        inc             c

        cp              [hl]            ; same byte already stored in GRAM_8 ?
        jr              nz, LABEL       ; no, so jump to real copy loop

        inc             l               ; advance GRAM_8
        inc             e
        inc             e               ; advance GRAM_16

LN_NUM	SET		LN_NUM + 1	; inc LineNum
        PURGE		LABEL, LN
        ENDR

        ld              a, [bc]         ; a <- bitmap data from OBJGRAM

        cp              [hl]            ; same byte already stored in GRAM_8 ?
        jr              nz, .SngCopy_7  ; no, so jump to real copy loop

        ; image already stored in GRAM, so no reason to set GRAM_Flags, etc.

        ret

LN_NUM	SET		0		; set LineNum = 0
	REPT		7		; do the first 7 lines (0-6)

LN      EQUS            STRSUB( "{LN_NUM}", 2, 1 )
LABEL	EQUS		".SngCopy_{LN}"

        IF              LN_NUM != 0
        ld              a, [bc]         ; a <- bitmap data from OBJGRAM
        inc             c
        ENDC
LABEL
        ld	        [hl+], a	; store out bitmap data to GRAM_8
        ld              [de], a         ; store out bitmap data to GRAM_16
        inc             e
        inc             e

LN_NUM	SET		LN_NUM + 1	; inc line number
        PURGE		LABEL, LN
        ENDR

        ld              a, [bc]         ; a <- bitmap data from OBJGRAM
.SngCopy_7
        ld	        [hl], a         ; store out bitmap data to GRAM_8
        ld              [de], a         ; store out bitmap data to GRAM_16

        ; stored out a NEW image to GRAM, so we need to set GRAM_Flags, etc.

        ld              a, e                      ; a = nnnn111.
        add             d
        sub             ( GRAM_16 >> 8 + 7 << 1 ) ;   = nnnn..NN

        ld              l, a
        ld              h, GRAM_Flags >> 8  ; hl -> GRAM flag for card

        ld              [hl], h         ; store out a non-zero value ($C5)

        IF              GRAM_MODIFIED == 1

        ld              l, h            ; hl -> GRAM_Modified ($C5C5)
        ld              [hl], l         ; store out a non-zero value

        ENDC

        ret

.DblCopy

        ; -------------------------------------------------------------------
        ;                  DblCopy Routines (when YRES = 1)
        ; -------------------------------------------------------------------

        ; first, test if the image already exists at this loc. in GRAM

LN_NUM  SET             0               ; set LineNum = 0
        REPT            15              ; do the first 15 lines (0-14)

LN      EQUS            STRSUB( "{LN_NUM}", 2, 1 )
LABEL	EQUS		".DblCopy_{LN}"

        ld              a, [bc]         ; a <- bitmap data from OBJGRAM
        inc             c

        cp              [hl]            ; same byte already stored in GRAM ?
        jr              nz, LABEL       ; no, so jump to real copy loop

        IF              LN_NUM !=7
        inc             l               ; advance GRAM_8
        inc             e
        inc             e               ; advance GRAM_16
        ELSE
        inc             hl              ; advance GRAM_8
        inc             e
        inc             de              ; advance GRAM_16
        ENDC

LN_NUM	SET		LN_NUM + 1	; inc LineNum
        PURGE		LABEL, LN
        ENDR

        ld              a, [bc]         ; a <- bitmap data from OBJGRAM

        cp              [hl]            ; same byte already stored in GRAM ?
        jr              nz, .DblCopy_F  ; no, so jump to real copy loop

        ; image already stored in GRAM, so no reason to set GRAM_Flags, etc.

        ret

LN_NUM	SET		0		; set LineNum = 0
        REPT            15              ; do the first 15 lines (0-14)

LN      EQUS            STRSUB( "{LN_NUM}", 2, 1 )
LABEL	EQUS		".DblCopy_{LN}"

        IF              LN_NUM != 0
        ld              a, [bc]         ; a <- bitmap data from OBJGRAM
        inc             c
        ENDC
LABEL
        ld	        [hl+], a	; store out bitmap data to GRAM_8
        ld              [de], a         ; store out bitmap data to GRAM_16
        IF              LN_NUM !=7
        inc             e
        inc             e               ; advance GRAM_16
        ELSE
        inc             e
        inc             de              ; advance GRAM_16
        ENDC

LN_NUM	SET		LN_NUM + 1	; inc LineNum
        PURGE		LABEL, LN
        ENDR

        ld              a, [bc]         ; a <- bitmap data from OBJGRAM
.DblCopy_F
        ld	        [hl], a         ; store out bitmap data to GRAM_8
        ld              [de], a         ; store out bitmap data to GRAM_16

        ; stored out a NEW image to GRAM, so we need to set GRAM_Flags, etc.

        ld              a, e                      ; a = nnnn111.
        add             d
        sub             ( GRAM_16 >> 8 + 7 << 1 ) ;   = nnnn..NN

        ld              l, a
        ld              h, GRAM_Flags >> 8  ; hl -> GRAM flag for 2nd card

        ld              [hl], h         ; store out a non-zero value ($C5)

        sub             $10
        sbc             0
        ld              l, a            ; hl -> GRAM flag for 1st card

        ld              [hl], h         ; store out a non-zero value ($C5)

        IF              GRAM_MODIFIED == 1

        ld              l, h            ; hl -> GRAM_Modified ($C5C5)
        ld              [hl], l         ; store out a non-zero value

        ENDC

        ret

EXEC_1358::                             ; finish up the job of EXEC_130B

        ld              a, [ehNewYPos]  ; a <- HIGH byte of NewYPos

        di                              ; disable interrupts to restore stack

        add             a, a            ; is NewYPos.15 = 1 ?
        jr              nc, EXEC_138B_1 ; no, so no illegal wrap on Y

        ; illegal wrap at the TOP or BOTTOM, so transfer control to $13C3

        add             a, a            ; put NewYPos.14 in CY

        ld              hl, R5          ; hl -> R5

        ld              a, $12
        ld              [hl+], a
        ld              [hl], a         ; [R5] <- $1212 (RetAddr)

        adc             a, ( $02 - $12 ) & $FF  ; a = 2 if BOTTOM, 3 if TOP

        ld              l, R0 & $FF     ; hl -> R0

        ld              [hl+], a        ; store out LOW byte of [R0]
        ld              [hl], 0         ; store out HIGH byte of [R0] (0)

        ld              sp, $4786       ; return to EXEC @ $13C3

        reti                            ; return, enabling interrupts

EXEC_138B::                             ; finish up the job of EXEC_130B

        ld              hl, ehNewYPos   ; hl -> HIGH byte of NewYPos

        di                              ; disable interrupts to restore stack

        ld              a, [SCRATCHPAD_RAM + $018 * 2]  ; a <- View_Y2

        cp              [hl]            ; is View_Y2 < NewYPos ?
        jr              c, EXEC_13A4_1  ; yes, so went off view at BOTTOM

        ld              a, [SCRATCHPAD_RAM + $019 * 2]  ; a <- View_Y1

        cp              [hl]            ; is View_Y1 > NewYPos ?
        jr              c, EXEC_138B_1  ; no, so didn't go off view at TOP
        jr              nz, EXEC_13A4_1 ; no, so didn't go off view at TOP

EXEC_138B_1::

        ld              sp, $4424       ; return to EXEC @ $1212

        reti                            ; return, enabling interrupts

EXEC_13A4::

        ld              a, [ehNewYPos]  ; a <- HIGH byte of NewYPos

        di                              ; disable interrupts to restore stack

        cp              104             ; is NewYPos in 0..103 ?
        jr              c, EXEC_138B_1  ; yes, so object is on-screen

        ; went off screen at the TOP or BOTTOM, so transfer control to $13C3

        cp              115             ; CY = 0 if if NewXPos in 115..255

EXEC_13A4_1::

        ld              hl, R5          ; hl -> R5

        ld              a, $12
        ld              [hl+], a
        ld              [hl], a         ; [R5] <- $1212 (RetAddr)

        ; if NewYPos in 104..114 = BOTTOM, if NewYPos in 115..255 = TOP

        sbc             a, $12 - $03    ; a = 2 if BOTTOM, 3 if TOP

        ld              l, R0 & $FF     ; hl -> R0

        ld              [hl+], a        ; store out LOW byte of [R0]
        ld              [hl], 0         ; store out HIGH byte of [R0] (0)

        ld              sp, $4786       ; return to EXEC @ $13C3

        reti                            ; return, enabling interrupts

        INCLUDE         "EXECHLPR.ASM"
        INCLUDE         "RTPAT.ASM"

SECTION "Cartridge_ROM_1",CODE,BANK[6]

GAME_1:

IF      GAME_SPLITS > 0

PRINTT  "{GAME_NAME}\n"

        INCBIN  "{GAME_NAME}.IG1"

ENDC

IF      GAME_SPLITS > 1

SECTION "Cartridge_ROM_2",CODE,BANK[7]

GAME_2:
        INCBIN  "{GAME_NAME}.IG2"

ENDC

IF      GAME_SPLITS > 2

SECTION "Cartridge_ROM_3",CODE,BANK[8]

GAME_3:
        INCBIN  "{GAME_NAME}.IG3"

ENDC

SECTION "Collision_Code",CODE,BANK[3]

; ---------------------------------------------------------------------------
; Data in ROM that needs to be aligned on special multiples should go here
; ---------------------------------------------------------------------------

; AttrTab -- Takes YM, XM, YR, XS, YS, and YF as "parameters" in bits 2-7
;            of the LOW byte and returns:
;
;              256-byte page #0:
;
;                Offset  Name         Output                 Dependencies
;
;                  0     Width        8 or 16                XS
;                  1     Height       4, 8, 16, 32, or 64    YR, YS, YF
;                  2     SrcOpCode    LD      A, [HL+/-]     YM
;                  3     DstOpCode    INC/DEC (D)E           YM, YR
;
;              256-byte page #1:
;
;                Offset  Name         Output                 Dependencies
;
;                  0     Reserved     Undefined              N/A
;                  1     EntryPoint   BP_TQ - BP_EP          YS, YF
;                                     BP_TQ - BP_EP - 2
;                  2     XMirrorTab   XMirrorTab >> 8        XM
;                                     ZShiftTab >> 8
;                  3     XSizeTab     ZShiftTabX >> 8        XS
;                                     ZShiftTab
;
;              256-byte page #2:
;
;                Offset  Name         Output                 Dependencies
;
;                  0     LoopPoint    BP_DblLoop - BP_JR - 2 YS, YF
;                                     DP_SngLoop - BP_JR - 2
;                                     BP_TQ - BP_JR - 1
;                  1     OpCode #1    LD      A, [HL+/-]     YM, YS, YF
;                                     NOP
;                                     LD      C, $XX
;                  2     OpCode #2    OR      [HL]           YS, YF
;                                     POP     HL
;                                     $XX
;                  3     OpCode #3    INC/DEC (H)L           YM, YR, YS, YF
;                                     LD      A, [HL+/-]
;

AttrTab::                       ; @ $4000-$42FF, aligned on multiples of 256

PAGE        SET     0
            REPT    3

ATTR        SET     0
            REPT    64

YM          SET     ( ( ATTR >> 5 ) & 1 ) ^ 1           ; * INVERTED
XM          SET     ( ATTR >> 4 ) & 1
YR          SET     ( ATTR >> 3 ) & 1
XS          SET     ( ( ATTR >> 2 ) & 1 ) ^ 1           ; * INVERTED
YS          SET     ( ATTR >> 1 ) & 1
YF          SET     ( ATTR >> 0 ) & 1

            ; AttrTab Page #0

            IF      PAGE == 0                           ; PAGE #0

            DB      8 + 8 * XS                          ; 0 = WIDTH
            DB      ( 4 + 4 * YR ) << ( YS << 1 + YF )  ; 1 = HEIGHT
            DB      $2A + $10 * YM                      ; 2 = LD A, [HL+/-]

            IF      YR == 0
            DB      $1C + $01 * YM                      ; 3 = INC/DEC  E
            ELSE
            DB      $13 + $08 * YM                      ; 3 = INC/DEC  DE
            ENDC

            ENDC

            ; AttrTab Page #1

            IF      PAGE == 1                           ; PAGE #1

            DB      0                                   ; 0 = RESERVED

            IF      YS << 1 + YF == 0                   ; 1 = ENTRY POINT
            DB      ( rBP_TQ + 0 ) - ( rBP_EP + 2 )
            ELSE
            DB      ( rBP_TQ + 2 ) - ( rBP_EP + 2 )
            ENDC

            IF      XM == 0                             ; 2 = X-MIRROR TABLE
            DB      ZShiftTab >> 8
            ELSE
            DB      XMirrorTab >> 8
            ENDC

            IF      XS == 0                             ; 3 = X-SIZE TABLE
            DB      ZShiftTab >> 8
            ELSE
            DB      ZShiftTabX >> 8
            ENDC

            ENDC

            ; AttrTab Page #2

            IF      PAGE == 2

            IF      YS << 1 + YF == 0                   ; 0 = LOOP POINT
            DB      ( rBP_DblLoop ) - ( rBP_LP + 2 )
            ELSE
            IF      YS << 1 + YF == 1
            DB      ( rBP_TQ + 1 ) - ( rBP_LP + 2 )
            ELSE
            DB      ( rBP_SngLoop ) - ( rBP_LP + 2 )
            ENDC
            ENDC

            IF      YS << 1 + YF == 0                   ; 1-3 = OPCODES

            DB      $2A + $10 * YM                      ; 1 = LD  A, [HL+/-]
            DB      $B6                                 ; 2 = OR  [HL]
            IF      YR == 0
            DB      $2C + $01 * YM                      ; 3 = INC/DEC L
            ELSE
            DB      $23 + $08 * YM                      ; 3 = INC/DEC HL
            ENDC

            ELSE

            IF      YS << 1 + YF == 1

            DB      $00                                 ; 1 = NOP
            DB      $E1                                 ; 2 = POP HL
            DB      $2A + $10 * YM                      ; 3 = LD  A, [HL+/-]

            ELSE

            DB      $0E                                 ; 1 = LD  C, $XX
            DB      2 << YF                             ; 2 = DB  $XX
            DB      $2A + $10 * YM                      ; 3 = LD  A, [HL+/-]

            ENDC
            ENDC

            ENDC

ATTR        SET     ATTR + 1
            ENDR

PAGE        SET     PAGE + 1
            ENDR

XMirrorTab::                    ; @ $4300-$43FF, aligned on multiples of 256
        INCBIN  "XMTAB.BIN"

RowStartTab::                   ; @ $4400-$47FF, aligned on multiples of 512
        INCBIN  "ROWSTTAB.BIN"

RepLineTab::                    ; @ $4800-$48FF, aligned on multiples of 256
        INCBIN  "RPLNTAB.BIN"

MoXToBitMask::                  ; @ $4900-$49FF, aligned on multiples of 256
        INCBIN  "BITMASK.BIN"

;GROM_8:                         ; @ $4A00-$51FF, fixed here
        INCBIN  "GROM_8.BIN"    ; mapped to $3000-$37FF on the real INTV

;SqGrafx:                        ; @ $5200-$52FF, fixed here

CTR         SET     0
            REPT    256

            IF      CTR & $80 == 0

OUTB        SET     $80            ; !1234...

            IF      CTR & $40 == 0

            ; Sq12ToAddr: ..222111

SQ1COL      SET     ( CTR >> 0 ) & $07
SQ2COL      SET     ( CTR >> 3 ) & $07

            IF      SQ1COL != 7
OUTB        SET     OUTB | $40     ; flag that square #1 is solid
            ENDC

            IF      SQ2COL != 7
OUTB        SET     OUTB | $20     ; flag that square #2 is solid
            ENDC

            ELSE

            ; Sq34ToAddr: .!444333

SQ3COL      SET     ( CTR >> 0 ) & $07  ; ( ( CTR >> 6 ) & $03 ) | ( ( CTR << 2 ) & $04 )
SQ4COL      SET     ( CTR >> 3 ) & $07  ; ( ( CTR >> 3 ) & $04 ) | ( ( CTR >> 1 ) & $03 )

            IF      SQ3COL != 7
OUTB        SET     OUTB | $10     ; flag that square #3 is solid
            ENDC

            IF      SQ4COL != 7
OUTB        SET     OUTB | $08     ; flag that square #4 is solid
            ENDC

            ENDC

            ELSE

            ; SqGrafx: !1234...

            IF      CTR & $07 < 4

            ; do squares 1 & 2 (in rows 0-3)

LSQSOLID    SET     ( CTR >> 6 ) & $01
RSQSOLID    SET     ( CTR >> 5 ) & $01

            ELSE

            ; do squares 3 & 4 (in rows 4-7)

LSQSOLID    SET     ( CTR >> 4 ) & $01
RSQSOLID    SET     ( CTR >> 3 ) & $01

            ENDC

OUTB        SET     0

            IF      LSQSOLID == 1
OUTB        SET     OUTB | $F0
            ENDC

            IF      RSQSOLID == 1
OUTB        SET     OUTB | $0F
            ENDC

            ENDC

            DB      OUTB

CTR         SET     CTR + 1
            ENDR

        ; shifting tables with XSIZE = 0

RShiftTab::                     ; @ $5300-$59FF, aligned on multiples of 256
        INCBIN  "RSHIFT.BIN"    ; shifts to the RIGHT by 1-7 bits

ZShiftTab::                     ; @ $5A00-$5AFF, aligned on multiples of 256
        INCBIN  "ZSHIFT.BIN"    ; shifts by 0 (when no shifting is necessary)

LShiftTab::                     ; @ $5B00-$61FF, aligned on multiples of 256
        INCBIN  "LSHIFT.BIN"    ; shifts to the LEFT by 1-7 bits

        ; shifting tables with XSIZE = 1

RShiftTabX::                    ; @ $6200-$68FF, aligned on multiples of 256
        INCBIN  "RSHIFTX.BIN"   ; shifts to the RIGHT by 1-7 bits

ZShiftTabX::                    ; @ $6900-$69FF, aligned on multiples of 256
        INCBIN  "ZSHIFTX.BIN"   ; shifts by 0 (when no shifting is necessary)

LShiftTabX::                    ; @ $6A00-$78FF, aligned on multiples of 256
        INCBIN  "LSHIFTX.BIN"   ; shifts to the LEFT by 1-15 bits

CardYx8ToCard::                 ; @ $7900-$7AFF, aligned on multiples of 256

            ; 1st 256-byte page - LOW byte of CARD @ DataAddr

CTR         SET     0
            REPT    256

CARD_Y      SET     ( CTR >> 3 ) & $1F

            DB      ( CARD + CARD_Y * 20 * CARD_RECSIZE ) & $FF

CTR         SET     CTR + 1
            ENDR

            ; 2nd 256-byte page - HIGH byte of BACKTAB address @ data address

CTR         SET     0
            REPT    256

CARD_Y      SET     ( CTR >> 3 ) & $1F

            DB      ( CARD + CARD_Y * 20 * CARD_RECSIZE ) >> 8

CTR         SET     CTR + 1
            ENDR

COLL:
        INCLUDE "COLL.ASM"

; ---------------------------------------------------------------------------
;         .::** Start of WRAM Bank 0 ($C000-$CFFF) declarations **::.
; ---------------------------------------------------------------------------

SECTION "Work_RAM_Bank_0",BSS[$C000]

GRAM_16         DS    64 * 16      ; $C000-$C3FF, aligned on multiples of 256

TileInUseBy     DS    256          ; $C400-$C4FF, aligned on multiples of 256

GRAM_Flags      DS    64 * 4       ; $C500-$C5FF, fixed here

MO              DS    MO_SIZE      ; $C600-$C643, aligned on multiples of 256

STACK_0         DS    $7C - 2      ; $C644-$C6BD, no special alignment
SP_0            DS    2            ; $C6BE-$C6BF, no special alignment

MO_FgCols       DS    8            ; $C6C0-$C6C7, fixed here (see STIC_Write)
MO_BgCols       DS    8            ; $C6C8-$C6CF, fixed here (see STIC_Write)
MO_CsCols       DS    8            ; $C6D0-$C6D7, fixed here (see STIC_Write)

STACK_1         DS    $28 - 2      ; $C6D8-$C6FD, no special alignment
SP_1            DS    2            ; $C6FE-$C6FF, no special alignment

ayVolToAudVol   DS    16           ; $C700-$C70F, aligned on multiples of 256

vbfRegs:                           ; * OK ?
tmfRegs         DS    4 * 2        ; used to store AF, BC, DE, and HL

TimerFunc       DS    rTF_EnvEnd_0 - TimerFunc_ROM
TF_EnvEnd_0     DS    rTF_EnvEnd_1 - rTF_EnvEnd_0
TF_EnvEnd_1     DS    rTF_EnvEnd_2 - rTF_EnvEnd_1
TF_EnvEnd_2     DS    rTF_EnvEnd_3 - rTF_EnvEnd_2
TF_EnvEnd_3     DS    rTF_EnvEnd_4 - rTF_EnvEnd_3
TF_EnvEnd_4     DS    rTF_EnvEnd_5 - rTF_EnvEnd_4
TF_EnvEnd_5     DS    rTF_EnvEnd_6 - rTF_EnvEnd_5
TF_EnvEnd_6     DS    rTF_EnvEnd_7 - rTF_EnvEnd_6
TF_EnvEnd_7     DS    rTF_EnvEnd_9 - rTF_EnvEnd_7
TF_EnvEnd_9     DS    rTF_EnvEnd_F - rTF_EnvEnd_9
TF_EnvEnd_F     DS    rTF_EnvEnd_C - rTF_EnvEnd_F
TF_EnvEnd_C     DS    rTF_EnvEnd_D - rTF_EnvEnd_C
TF_EnvEnd_D     DS    rTF_EnvEnd_A - rTF_EnvEnd_D
TF_EnvEnd_A     DS    rTF_EnvEnd_E - rTF_EnvEnd_A
TF_EnvEnd_E     DS    rTF_EnvEnd_B - rTF_EnvEnd_E
TF_EnvEnd_B     DS    rTF_EnvEnd_8 - rTF_EnvEnd_B
TF_EnvEnd_8     DS    rTF_Exit - rTF_EnvEnd_8
TF_Exit         DS    rTF_M2 - rTF_Exit
TF_M2           DS    TimerFunc_ROM_End - rTF_M2

GH_STACK        DS    4 * 2        ; Esee EXEC_1523
DS_STACK        DS    3 * 2        ; Esee HOME_17D6

ehProcList      DS    6 * 4        ; Esee HOME_17D6
ehProcExtra     DS    1            ; Esee UDB_Update
ehProcCount     DS    1            ; Esee HOME_17D6

WINPAL_1        DS    2 * 4 * 2    ; Esee HBlankFunc & VBlankFunc

WRB0PAD_2       DS    $E6 - ( TimerFunc_ROM_End - TimerFunc_ROM + 4 * 2 + 3 * 2 + 6 * 4 + 1 + 1 + 2 * 4 * 2 )

ayAudTerm       DS    1            ; $C7FE  used by AY8914_Update &
ayMask          DS    1            ; $C7FF          TimerFunc

UDB_CardNdx     DS    8 * 2        ; $C800-$C80F, aligned on multiples of 256

COLLMO          DS    COLLMO_SIZE  ; $C810-$C8B8, must reside on one page

; ---------------------------------------------------------------------------
;                .::** Background variables & code **::.
; ---------------------------------------------------------------------------

;AP_BW           DS    3

BktWrite        DS    rBW_BW - BktWrite_ROM
BW_BW           DS    rBW_SP - rBW_BW
BW_SP           DS    BktWrite_ROM_End - rBW_SP

SrchPal1        DS    rS1_ColLoop - SrchPal1_ROM
S1_ColLoop      DS    rS1_BG - rS1_ColLoop
S1_BG           DS    rS1_CN - rS1_BG
S1_CN           DS    rS1_PI - rS1_CN
S1_PI           DS    rS1_CS - rS1_PI
S1_CS           DS    rS1_BH - rS1_CS
S1_BH           DS    rS1_BL - rS1_BH
S1_BL           DS    SrchPal1_ROM_End - rS1_BL

;SrchPal1        DS    rS1_ColLoop - SrchPal1_ROM
;S1_ColLoop      DS    rS1_BG - rS1_ColLoop
;S1_BG           DS    rS1_UpdTileNfo - rS1_BG
;S1_UpdTileNfo   DS    rS1_CN - rS1_UpdTileNfo
;S1_CN           DS    rS1_PI - rS1_CN
;S1_PI           DS    rS1_B1 - rS1_PI
;S1_B1           DS    rS1_BH - rS1_B1
;S1_BH           DS    rS1_BL - rS1_BH
;S1_BL           DS    rS1_CS - rS1_BL
;S1_CS           DS    SrchPal1_ROM_End - rS1_CS

SrchPal2        DS    rS2_PalLoop - SrchPal2_ROM
S2_PalLoop      DS    rS2_BG - rS2_PalLoop
S2_BG           DS    rS2_FG - rS2_BG
S2_FG           DS    rS2_ML - rS2_FG
S2_ML           DS    rS2_MH - rS2_ML
S2_MH           DS    rS2_GB - rS2_MH
S2_GB           DS    rS2_CN - rS2_GB
S2_CN           DS    rS2_B1 - rS2_CN
S2_B1           DS    rS2_PI - rS2_B1
S2_PI           DS    rS2_B2 - rS2_PI
S2_B2           DS    rS2_CP - rS2_B2
S2_CP           DS    rS2_BH - rS2_CP
S2_BH           DS    rS2_BL - rS2_BH
S2_BL           DS    rS2_FH - rS2_BL
S2_FH           DS    rS2_FL - rS2_FH
S2_FL           DS    SrchPal2_ROM_End - rS2_FL

SrchPal4        DS    rS4_S1 - SrchPal4_ROM
S4_S1           DS    rS4_B1 - rS4_S1
S4_B1           DS    rS4_L1 - rS4_B1
S4_L1           DS    rS4_H1 - rS4_L1
S4_H1           DS    rS4_S2 - rS4_H1
S4_S2           DS    rS4_B2 - rS4_S2
S4_B2           DS    rS4_L2 - rS4_B2
S4_L2           DS    rS4_H2 - rS4_L2
S4_H2           DS    rS4_S3 - rS4_H2
S4_S3           DS    rS4_B3 - rS4_S3
S4_B3           DS    rS4_L3 - rS4_B3
S4_L3           DS    rS4_H3 - rS4_L3
S4_H3           DS    rS4_S4 - rS4_H3
S4_S4           DS    rS4_B4 - rS4_S4
S4_B4           DS    rS4_BC - rS4_B4
S4_BC           DS    rS4_PalNext1 - rS4_BC
S4_PalNext1     DS    rS4_CN - rS4_PalNext1
S4_CN           DS    rS4_CI - rS4_CN
S4_CI           DS    rS4_PI - rS4_CI
S4_PI           DS    rS4_H4 - rS4_PI
S4_H4           DS    rS4_L4 - rS4_H4
S4_L4           DS    SrchPal4_ROM_End - rS4_L4

; ---------------------------------------------------------------------------
; .::** Collision detection variables & code (see STIC_FlagCollisions) **::.
; ---------------------------------------------------------------------------

STIC_CollChkMO  DS    rCM_DstChkC - STIC_CollChkMO_ROM
CM_DstChkC      DS    rCM_WH - rCM_DstChkC
CM_WH           DS    rCM_XP - rCM_WH
CM_XP           DS    rCM_YP - rCM_XP
CM_YP           DS    rCM_B1 - rCM_YP
CM_B1           DS    rCM_B2 - rCM_B1
CM_B2           DS    rCM_DA - rCM_B2
CM_DA           DS    rCM_RC - rCM_DA
CM_RC           DS    rCM_NC - rCM_RC
CM_NC           DS    rCM_DC - rCM_NC
CM_DC           DS    rCM_DB - rCM_DC
CM_DB           DS    rCM_SC - rCM_DB
CM_SC           DS    rCM_SB - rCM_SC
CM_SB           DS    STIC_CollChkMO_ROM_End - rCM_SB

PixelPercise    DS    rPP_SI - PixelPercise_ROM
PP_SI           DS    rPP_DI - rPP_SI
PP_DI           DS    rPP_SA - rPP_DI
PP_SA           DS    rPP_SX - rPP_SA
PP_SX           DS    rPP_SS - rPP_SX
PP_SS           DS    rPP_SR - rPP_SS
PP_SR           DS    rPP_DD - rPP_SR
PP_DD           DS    rPP_DA - rPP_DD
PP_DA           DS    rPP_DX - rPP_DA
PP_DX           DS    rPP_DS - rPP_DX
PP_DS           DS    rPP_DR - rPP_DS
PP_DR           DS    rPP_JR - rPP_DR
PP_JR           DS    PixelPercise_ROM_End - rPP_JR

CardXx8         DS    1
CardYx8         DS    1
CardWidth       DS    1

BkgPixelPercise DS    rBP_HC - BkgPixelPercise_ROM
BP_HC           DS    rBP_CS - rBP_HC
BP_CS           DS    rBP_MD - rBP_CS
BP_MD           DS    rBP_MI - rBP_MD
BP_MI           DS    rBP_RI - rBP_MI
BP_RI           DS    rBP_EP - rBP_RI
BP_EP           DS    rBP_MC - rBP_EP
BP_MC           DS    rBP_TQ - rBP_MC
BP_TQ           DS    rBP_MX - rBP_TQ
BP_MX           DS    rBP_MS - rBP_MX
BP_MS           DS    rBP_LP - rBP_MS
BP_LP           DS    rBP_RL - rBP_LP
BP_RL           DS    rBP_VL - rBP_RL
BP_VL           DS    rBP_VH - rBP_VL
BP_VH           DS    BkgPixelPercise_ROM_End - rBP_VH

; ---------------------------------------------------------------------------
;             .::** VBlankFunc & HBlankFunc variables **::.
; ---------------------------------------------------------------------------

WINPAL_0        DS    8 * 2        ;  2 palettes at 8 bytes ea.

hbfHL           DS    2            ;  \
hbfDE           DS    2            ;   > used to store af, de, and hl
hbfRoutine      DS    2            ;  /

HBlankFunc:                        ; 

HF_WinOff       DS    rHF_WinOn - rHF_WinOff
HF_WinOn        DS    rHF_DO - rHF_WinOn
HF_DO           DS    rHF_M2 - rHF_DO
HF_M2           DS    HBlankFunc_ROM_End - rHF_M2

;vbfRegs         DS    4 * 2        ; used to store af, bc, de, and hl

VBlankFixUp     DS    rVF_M2 - VBlankFixUp_ROM
VF_M2           DS    VBlankFixUp_ROM_End - rVF_M2

; ---------------------------------------------------------------------------
;           .::** JoyPad variables (see ReadJoy, MainLoop) **::.
; ---------------------------------------------------------------------------

JoyPress        DS      1          ; \
JoyDown         DS      1          ;  \
JoyRelease      DS      1          ;  / used by ReadJoy
JoyUp           DS      1          ; /

Select_Used     DS      1

; ---------------------------------------------------------------------------
;                 .::** Miscellaneous variables **::.
; ---------------------------------------------------------------------------

GB_Type         DS    1            ; $01 = GB/SGB, $FF = GBP, $11 = GBC
VBlank_Ctr      DS    1            ;  see VBlankFunc, MainLoop
STIC_Display    DS    1            ;  see HBlankFunc, MainLoop               d
STIC_Mode_New   DS    1            ; see STIC_Write
LY_WinOff       DS    1            ; row at which the window may go back off
Save_BC         DS    2            ; used by MO_CopyToVRAM

Page_Ready      DS    1            ; 
CurFrames       DS    1            ; 
TotFrames       DS    1            ; 

UDB_Flag        DS    1            ; E

ehSP111         DS    1            ; Esee EXEC_11C0
ehSP113         DS    1            ; Esee EXEC_11C0
ehNewYPos       DS    1            ; Esee EXEC_130B
ehKeyClicks     DS    1            ; Esee EXEC_1523

STIC_Analyzed   DS    1            ;  set to 1 upon completion of STIC_Analyze
                                   ;  and cleared by VBlankFunc (lets it know
                                   ;  when it can update rSCX & rSCY registers)
rSCY_New        DS    1            ;  new rSCY (updated in VBlankFunc)
rSCX_New        DS    1            ;  new rSCX (updated in VBlankFunc)
rSCX_Old        DS    1            ;  old rSCX (in case LocalOAM not ready)

;ColComp         DS    1            ; E
;ColNum          DS    1            ; 
;ColHigh         DS    1            ; 
;ColLow          DS    1            ; 
;ColRed          DS    1            ; 
;ColGreen        DS    1            ; 
;ColBlue         DS    1            ;  see COLTUNE.ASM

StackStart      DS    1

SECTION "GRAM_8_Bit",BSS[$CE00]

GRAM_8          DS    64 * 8       ; $CE00-$CFFF, fixed here (see POKE, etc.)
                                   ; 512 8-bit locations,  mapped $3800-$39FF

; ---------------------------------------------------------------------------
;         .::** Start of WRAM Bank 1 ($D000-$DFFF) declarations **::.
; ---------------------------------------------------------------------------

SECTION "Work_RAM_Bank_1",BSS[$D000]

SYSTEM_RAM         DS   352 * 2   ; $D000-$D2BF, aligned on multiples of 512

WRB1PAD_0          DS   $18

sSaveRow           DS   20 * 2

R0                 DS   2         ; R0 = Primary Acculumator
R1                 DS   2         ; R1,
R2                 DS   2         ; R2,
R3                 DS   2         ; R3 = Data Counters
R4                 DS   2         ; R4,
R5                 DS   2         ; R5 = Data Counters w/ auto-increment
R6                 DS   2         ; R6 = Stack Pointer
R7                 DS   2         ; R7 = Program Counter

AddrHiToROMB0      DS   256 - 16  ; $D310-$D3FF  fixed here (see PEEK, etc.)
AddrHiToReal       DS   256       ; $D400-$D4FF  fixed here (see PEEK, etc.)

ayPerToAudPoly     DS   256       ; $D500-$D5FF Enoise period to polynomial

SCRATCHPAD_RAM     DS   256 * 2   ; $D600-$D7FF, aligned on multiples of 512

ayVolToAudEnv      DS   16        ; $D800-$D80F E"real" vols for AY8914 vols

ayTonAFreq         DS   2         ; $D810  TonA's GB frequency value (0-2047)
ayTonBFreq         DS   2         ; $D812  TonB's GB frequency value (0-2047)
ayTonCFreq         DS   2         ; $D814  TonC's GB frequency value (0-2047)
ayChnCVol          DS   1         ; $D816  Used to upd. Wave RAM for Snd #3

ayMuted            DS   1         ; $D817  used to override mixer enable bits
ayEnabled          DS   2         ; $D818  envelope & mixer enable bits

IH_50BB::          DS   rOS_PB - rIH_50BB
OS_PB::            DS   rOS_RR - rOS_PB
OS_RR::            DS   rOS_RA - rOS_RR
OS_RA::            DS   rOS_C1 - rOS_RA
OS_C1::            DS   rOS_C2 - rOS_C1
OS_C2::            DS   rOS_CA - rOS_C2
OS_CA::            DS   rIH_50BB_End - rOS_CA

IH_50D7::          DS   rIH_50D7_End - rIH_50D7

IH_515D::          DS   rBC_RR - rIH_515D
BC_RR::            DS   rBC_C1 - rBC_RR
BC_C1::            DS   rBC_RS - rBC_C1
BC_RS::            DS   rBC_SR - rBC_RS
BC_SR::            DS   rBC_EA - rBC_SR
BC_EA::            DS   rBC_CA - rBC_EA
BC_CA::            DS   rIH_515D_End - rBC_CA

WRB1PAD_1          DS   $B8 - ( rIH_50D7_End - rIH_50BB + rIH_515D_End - rIH_515D )

ayRegToHandler     DS   23 * 2    ; $D8D2-$D8FF Ejump table for AY8914 writes
                                  ; (must follow SCRATCHPAD_RAM "high")

WRB1PAD_2          DS   3 * 256

STIC               DS   64 * 2    ; $DC00-$DC7F, aligned on multiples of 512
                                  ; 64  14-bit locations, mapped $0000-$003F

IH_52B3            DS  rAV_O1 - rIH_52B3
AV_O1::            DS  rAV_XA - rAV_O1
AV_XA::            DS  rAV_XP - rAV_XA
AV_XP::            DS  rAV_O2 - rAV_XP
AV_O2::            DS  rAV_YA - rAV_O2
AV_YA::            DS  rAV_CA - rAV_YA
AV_CA::            DS  rAV_Loop - rAV_CA
AV_Loop::          DS  rAV_P1 - rAV_Loop
AV_P1::            DS  rAV_P2 - rAV_P1
AV_P2::            DS  rAV_DP - rAV_P2
AV_DP::            DS  rAV_PA - rAV_DP
AV_PA::            DS  rIH_52B3_End - rAV_PA

WRB1PAD_3          DS   $80 - ( rIH_52B3_End - rIH_52B3 )

stRegToHandler     DS   256       ; $DD00-$DDFF, must follow STIC

BitCntToMask       DS   16 * 2    ;

OS_STACK           DS   2 * 2
CC_STACK           DS   4 * 2

IH_5357::          DS   rUS_XP - rIH_5357
US_XP::            DS   rUS_SB - rUS_XP
US_SB::            DS   rUS_SC - rUS_SB
US_SC::            DS   rUS_Exit - rUS_SC
US_Exit::          DS   rUS_CA - rUS_Exit
US_CA::            DS   rUS_UpdSTIC - rUS_CA
US_UpdSTIC::       DS   rIH_5357_End - rUS_UpdSTIC

IH_51E2::          DS   rGC_CI - rIH_51E2
GC_CI::            DS   rGC_CA - rGC_CI
GC_CA::            DS   rIH_51E2_End - rGC_CA

IH_514A::          DS   rCA_RR - rIH_514A
CA_RR::            DS   rCA_C1 - rCA_RR
CA_C1::            DS   rCA_RS - rCA_C1
CA_RS::            DS   rCA_SR - rCA_RS
CA_SR::            DS   rCA_EA - rCA_SR
CA_EA::            DS   rCA_CA - rCA_EA
CA_CA::            DS   rIH_514A_End - rCA_CA

WRB1PAD_4          DS   256 - ( 16 * 2 + 2 * 2 + 4 * 2 + rIH_5357_End - rIH_5357 + rIH_51E2_End - rIH_51E2 + rIH_514A_End - rIH_514A )

GcToIc             DS   gcNumConstants * 2
HandDelay          DS   icNumConstants * 2

IH_50E0::          DS   rCC_XP - rIH_50E0
CC_XP::            DS   rCC_CD - rCC_XP
CC_CD::            DS   rCC_C1 - rCC_CD
CC_C1::            DS   rCC_IM - rCC_C1
CC_IM::            DS   rCC_CI - rCC_IM
CC_CI::            DS   rCC_CA - rCC_CI
CC_CA::            DS   rCC_RA - rCC_CA
CC_RA::            DS   rCC_C2 - rCC_RA
CC_C2::            DS   rIH_50E0_End - rCC_C2

IH_5128::          DS   rIH_5128_End - rIH_5128

Status            DS    1         ; $C6 Ebit #3-0 = Sign, Zero, Overflow & Carry
SDBD              DS    1         ; $C7 E1 = set, 0 = not set
RegNo             DS    1         ; $D1 Eused by InvalidCode

SECTION "High RAM",HRAM

; ---------------------------------------------------------------------------
;          .::** High RAM at $FF80 - $FFFE (127 bytes) **::.
; ---------------------------------------------------------------------------

; MO_Colors MUST be at $FF80 (see MO_SetColors, STIC_Write)

MO_Colors         DS    8 * 3 * 2 ; $80  used by CopyToOAM_Hi
BC_Active         DS    1         ; $B0  used by VBlankFunc, STIC_Write
BC_OCPS           DS    1         ; $B1  used by VBlankFunc, STIC_Analyze
BrdrColRGB        DS    2         ; $B2  used by VBlankFunc, STIC_Write

CopyToOAM         DS    CopyToOAM_ROM_End - CopyToOAM_ROM  ; $B4 Ecalled by VBlankFunc

STIC_Mode         DS    1         ; $C6 E0 = CS, 1 = FG/BG
STIC_Display_Hi   DS    1         ; $C7 Edisplay state (1=enabled, 0=disabled)

;Status            DS    1         ; $C6 Ebit #3-0 = Sign, Zero, Overflow & Carry
;SDBD              DS    1         ; $C7 E1 = set, 0 = not set

R7_Conv           DS    1         ; $C8 Ecomputed by Jump -- converts R7 from
                                  ;       a "real" address to a "CP1610" address

ayEnvPerNew       DS    2         ; $C9 
ayGetEnv          DS    1         ; $CB 
ayEnvType         DS    1         ; $CC 

ColorStack        DS    4         ; $CD Eused by STIC_Write, CS_BktWrite

;RegNo             DS    1         ; $D1 Eused by InvalidCode
ayAudPoly         DS    1         ; $D1 Epolynomial based on noise period

cSVBK             DS    1         ; $D2 Ecurrent RAM bank #
cROMB0            DS    1         ; $D3 Ecurrent ROM bank #, fixed here

ColorToRGB        DS    16 * 2    ; $D4 Eused by Bkt_Analyze

ayEnvVol          DS    1         ; $F4 
ayEnvInc          DS    1         ; $F5 
ayEnvCtr          DS    2         ; $F6 
ayEnvPerCur       DS    2         ; $F8 
ayEnvEndHandler   DS    2         ; $FA 



