; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

        ; CARD
        ;
        ;       0    LOW byte of DataAddr
        ;       1    HIGH byte of DataAddr
        ;       2    UniqueColors + Flags            rcannuuu
        ;       3    Single/PairColors  / Sq12Col    33222111
        ;       4    Single/PairInUseBy / Sq34Col    .!444333
        ; /\ -> 6    ColorIndices                    ....ffbb / 33221100 (*)
        ; || -> 5    PalNdx                          11ppp..1
        ; |_ -> 7    TileNdx

        ; TILE
        ;
        ;       0   HIGH byte of MakeUpNdx   1101NNnn
        ;       1   LOW byte of MakeUpNdx    nnnnffbb
        ; |- -> 2   MakeUpBank + Queue       ?q???bbb
        ; || -> 3   ColorIndices             ....ffbb
        ; || -> 4   LOW byte of DataAddr
        ; || -> 5   HIGH byte of DataAddr
        ; ||    6   LOW byte of VRAMAddr
        ; \/    7   HIGH byte of VRAMAddr

; **
; Bkg_Init:
;

Bkg_Init:

        ; initialize MakeUp_GROM (in WRAM bank #2)

        ld              a, 2            ; a = #/WRAM bank with MakeUp_GROM
        ldh             [rSVBK], a      ; switch to WRAM bank #2

        xor             a
        ld              bc, 256 * 16
        ld              hl, MakeUp_GROM

        call            MemFill         ; clear MakeUp_GROM

        ; initialize ML_BOTTOM and ML_TOP (in WRAM bank #3)

        ld              a, 5            ; a = #/WRAM bank with MOVELIST
        ldh             [rSVBK], a      ; switch to WRAM bank #5

        ld              hl, ML_BOTTOM

        xor             a
        ld              [hl+], a        ; set ML_BOTTOM = 0
        ld              [hl], a         ; set ML_TOP = 0

        ; initialize PALTAB_1 (InUseBy) (in WRAM bank #3)

        ld              a, 3            ; a = #/WRAM bank with CARD
        ldh             [rSVBK], a      ; switch to WRAM bank #3

;        ld              hl, CurToPer_ROM
;        ld              de, CurToPer
;        ld              c, 120

;        call            MemCopy         ; copy CurToPer_ROM to CurToPer

        xor             a
        ld              hl, PALTAB_1    ; hl -> PALTAB_1 (InUseBy fields)
        ld              c, 8 * 4 * 2

        call            MemFill         ; clear PALTAB_1

        ; initialize BDD_TileNdx and BDD_TileAttr (in WRAM bank #3)

        ld              bc, 32 * 12
        ld              hl, BDD_TileNdx

        call            MemFill         ; clear 12 rows of 32 tiles

        ld              bc, 32 * 12
        ld              hl, BDD_TileAttr

        call            MemFill         ; clear 12 rows of 32 tiles

        ; clear BACKTAB area of SYSTEM_RAM (set all locations to GROM #0)

;        ld              hl, SYSTEM_RAM
;        ld              bc, 240 * 2     ; 240 locations @ 2 bytes ea.
;
;        call            MemFill

        ; initialize Single/PairInUseBy (in WRAM bank #3)

        dec             a                  ; a = $FF
        ld              hl, SingleInUseBy  ; hl -> SingleInUseBy
        ld              c, 64

        call            MemFill         ; clear SingleInUseBy w/ $FF

        ld              hl, PairInUseBy
        ld              bc, 256

        call            MemFill         ; clear PairInUseBy w/ $FF

        ; initialize CARD structure (in WRAM bank #3)

        ld              a, $81
        ld              hl, CARD
        ld              bc, PALTAB_DMY0 & $FF00

        ld              de, 240 << 8 + $FF  ; do 240 cards
.CardLoop
        ld              [hl], c
        inc             l
        ld              [hl], c         ; set CARD's DataAddr to GROM #0
        inc             l

        ld              [hl+], a        ; set UniqueColors + Flags = $81

        ld              [hl+], a        ; set Single/PairColors  = $81
        ld              [hl], b         ; set Single/PairInUseBy = $D1
        inc             l

        inc             l               ; skip ColorIndices
        ld              [hl+], a        ; set PalNdx = $81
        ld              [hl], e         ; set TileNdx = 255
        inc             hl

        dec             d
        jr              nz, .CardLoop

        dec             a                     ; = ld a, $80
        ld              [BktGetCols_Flag], a  ; need to get new color indices
        ld              [BktGetCols_Stop], a  ; see BktGetCols

        ; initialize MakeUp_GRAM (in WRAM bank #6)

        ld              a, 6            ; a = #/WRAM bank with MakeUp_GRAM
        ldh             [rSVBK], a      ; switch to WRAM bank #6

        xor             a
        ld              hl, MakeUp_GRAM
        ld              bc, 64 * 16

        call            MemFill         ; clear MakeUp_GRAM

        ; initialize CL_BOTTOM and CL_TOP (in WRAM bank #6)

        ld              hl, CL_BOTTOM

        ld              [hl-], a        ; set CL_BOTTOM = 0
        ld              [hl], a         ; set CL_TOP = 0

        ; initialize TILE structure (in WRAM bank #6)

        ld              hl, TILE
        ld              bc, MakeUp_DMY

        ; set TILE records #0-#127 to use VRAM tiles #256-#383 ($9000-$97FF)

        ld              de, _VRAM + 256 * 16 + GRAM_8 >> 8 - $40
.TileLoop1
        ld              [hl], b
        inc             l
        ld              [hl], c         ; set MakeUpNdx = MakeUp_DMY
        inc             l

        ld              [hl], e         ; set MakeUpBank + Queue = $8E
        inc             l

        inc             l               ; skip ColorIndices
        inc             l               ; skip LOW byte of DataAddr
        inc             l               ; skip HIGH byte of DataAddr

        ld              [hl+], a        ; store out LOW byte of VRAMAddr
        ld              [hl], d         ; store out HIGH byte of VRAMAddr
        inc             hl

        add             $10             ; advance to next VRAM address
        jr              nc, .TileLoop1

        inc             d               ; inc HIGH byte of VRAM address

        bit             3, d            ; initialized 128 tiles ?
        jr              z, .TileLoop1   ; no, so keep looping

        ; set TILE records #128-#255 to use VRAM tiles #128-#255 ($8800-$8FFF)

        ld              d, ( _VRAM + 128 * 16 ) >> 8
.TileLoop2
        ld              [hl], b
        inc             l
        ld              [hl], c         ; set MakeUpNdx = MakeUp_DMY
        inc             l

        ld              [hl], e         ; set MakeUpBank + Queue = $8E
        inc             l

        inc             l               ; skip ColorIndices
        inc             l               ; skip LOW byte of DataAddr
        inc             l               ; skip HIGH byte of DataAddr

        ld              [hl+], a        ; store out LOW byte of VRAMAddr
        ld              [hl], d         ; store out HIGH byte of VRAMAddr
        inc             hl

        add             $10             ; advance to next VRAM address
        jr              nc, .TileLoop2

        inc             d               ; inc HIGH byte of VRAM address

        bit             3, d            ; initialized 256 tiles ?
        jr              nz, .TileLoop2  ; no, so keep looping

        ; initialize TileInUseBy (in WRAM bank #0)

        ld              hl, TileInUseBy
        ld              bc, 255

        call            MemFill         ; clear TileInUseBy #0-#254

        ; reserve tiles 193-199 & 249-255 (odd only) for own selfish purposes

        dec             a               ; = ld a, 255

        ld              c, 4            ; reserve 4 tiles
.Reserve1
        ld              [hl-], a        ; set tile's InUseBy = 255
        dec             l               ; skip even numbered tiles

        dec             c
        jr              nz, .Reserve1

        ld              c, 4            ; reserve 4 tiles
        ld              l, 193          ; hl -> TileInUseBy[ 193 ]
.Reserve2
        ld              [hl+], a        ; set tile's InUseBy = 255
        inc             l               ; skip even numbered tiles

        dec             c
        jr              nz, .Reserve2

        ; initialize GRAM_Flags (in WRAM bank #0)

        ld              hl, GRAM_Flags
        xor             a
.FlagsLoop
        ld              [hl+], a
        inc             a
        jr              nz, .FlagsLoop

        ; Copy BktWrite, SrchPal1, SrchPal2, & SrchPal4 to WRAM bank #0

        ld              hl, BktWrite_ROM
        ld              de, BktWrite
        ld              bc, SrchPal4_ROM_End - BktWrite_ROM

        call            MemCopy         ; copy SrchPal1, 2, & 4 to Low RAM

        ; init upper VRAM tiles #128-#383 to all possible ColoredSquares

        ld              a, 1
        ldh             [rVBK], a       ; switch to bank 1 / upper-tiles

        ld              hl, SquareGrafx
        ld              de, _VRAM + 128 * 16
        ld              bc, 256 * 16

        call            MemCopy

        xor             a
        ldh             [rVBK], a       ; switch to bank 0 / lower-tiles

        ;

        ld              bc, 8 * 2             ; 8 lines @ 2 bytes ea.
        ld              hl, _VRAM + 249 * 16  ; hl -> tile #249's bitmap data

        call            MemFill

        dec             c               ; = ld c, $FF

        ld              hl, _VRAM + 251 * 16  ; hl -> tile #251's bitmap data

        REPT            8

        ld              [hl], c         ; store out LSB of color #0 ($FF)
        inc             l
        ld              [hl+], a        ; store out MSB of color #0 ($00)

        ENDR

        ld              hl, _VRAM + 253 * 16  ; hl -> tile #253's bitmap data

        REPT            8

        ld              [hl+], a        ; store out LSB of color #0 ($00)
        ld              [hl], c         ; store out MSB of color #0 ($FF)
        inc             l

        ENDR

        ld              a, c
        ld              c, 8 * 2              ; 8 lines @ 2 bytes ea.
        ld              hl, _VRAM + 255 * 16  ; hl -> tile #255's bitmap data

        call            MemFill

        ret

; **
; BktGetCols:     Get new color indices for all cards that have their
;                 Re-analyze bit set (bit #7 of UniqueColors + Flags),
;                 either as a result of a "deadlocked" palette or as
;                 a "victim" of a CS-advance bit change.
;
; Parameters:     [rSVBK] must already be set to WRAM bank #3.
;
; Modifies:       All registers.
;

BktGetCols:

        ld              hl, BktGetCols_Flag

        ld              a, [hl]         ; a <- BktGetCols_Flag
        or              a               ; is it <> 0 ?
        ret             z               ; no, so nothing to do

        xor             a
        ld              [hl], a         ; set BktGetCols_Flag = 0

        ldh             a, [STIC_Mode]
        or              a               ; are we in CS mode ?
        jp              nz, .FGBG       ; no, so get colors in FG/BG mode

        ; ------------------------------------------------------------------
        ;                      COLOR STACK MODE
        ; ------------------------------------------------------------------

.CS_Refresh
        ld              de, -1 * CARD_RECSIZE
.CS_Loop
        add             hl, de

        ld              a, [hl]         ; a <- UniqueColors + Flags
        add             a, a            ; is Re-analyze (bit #7) set ?
        jr              nc, .CS_Loop    ; no, so keep checking cards

        ret             z               ; hit BktGetCols_Stop (checked 240)

;        push            af
;
;        ld              a, l
;        cp              ( CARD + ( 20 * 1 + 3 ) << 3 + 2 ) & $00FF
;        jr              nz, .Ahead
;
;        ld              a, h
;        cp              ( CARD + ( 20 * 1 + 3 ) << 3 + 2 ) >> 8
;        jr              nz, .Ahead
;
;        ld              b, b            ; BREAK !!!
;.Ahead
;        pop             af

        ; Re-analyze (bit #7) is set, now direct it to the proper handler

        push            hl              ; save CARD @ UniqueColors + Flags

        add             a, a              ; ColoredSquares (bit #6) set ?
        jr              c, .CS_SrchPal4   ; yes, so re-analyze w/ SrchPal4

        res             7, [hl]         ; reset Re-analyze bit

        bit             2, a              ; is UniqueColors = 1 ?
        jr              nz, .CS_SrchPal1  ; yes, so re-analyze w/ SrchPal1

        ; -------------------------------------------------------------------
        ;       Get new colors using SrchPal2 (COLOR STACK MODE)
        ; -------------------------------------------------------------------

        ; * S2_ML = MakeUpNdx LOW       * S2_MH = MakeUpNdx HIGH
        ; * S2_GB = GraphicBase         * S2_CN = CardNum

        dec             l               ; pos @ HIGH byte of DataAddr
        dec             l               ; pos @ LOW byte of DataAddr

        ; extract CardNum from CARD address

        ld              a, h            ; a = 11.11NNN (NNN = CardNum's MSBs)
        xor             l               ;   = ??n??NNN
        ld              [S2_CN + 1], a  ; store out "CardNum"

        ; extract MakeUpNdx and GraphicBase from card's DataAddr

        ld              a, [hl+]        ; a <- LOW byte of DataAddr

        add             a, a            ; a = nnnn.... (CY = N)
        ld              [S2_ML + 1], a  ; store out "MakeUpNdx LOW"

        ld              a, [hl+]        ; a <- HIGH byte of DataAddr

        ld              de, ( MakeUp_GROM - GRAM_8 << 1 ) & $FF00 + ( ( GRAM_8 - MakeUp_GROM >> 1 ) >> 8 ) & $00FF

        rla                             ; put N in bit #0
        jr              c, .CS_InGRAM   ; if CY = 1, DataAddr >= $8000

        ld              de, ( MakeUp_GROM - GROM_8 << 1 ) & $FF00 + ( ( GROM_8 - MakeUp_GROM >> 1 ) >> 8 ) & $00FF
.CS_InGRAM
        add             d               ; a = HIGH byte of MakeUpNdx
        ld              [S2_MH + 1], a  ; store out "MakeUpNdx HIGH"

        ld              a, e            ; a = "GraphicBase"
        ld              [S2_GB + 1], a  ; store out "GraphicBase"

        ; finally, get UniqueColors + Flags and FGCol from CARD

        ld              d, 2            ; d = UniqueColors

        ld              a, [hl+]        ; a <- UniqueColors + Flags

        and             $18             ; a = ...nn...
        rrca                            ;   = ....nn..
        rrca                            ;   = .....nn.
        rrca                            ;   = ......nn

        add             ColorStack & $FF
        ld              c, a

        ld              a, [c]          ; a <- BGCol << 1

        ld              c, [hl]         ; c <- Old Single/PairColors

        xor             c
        and             $1E             ; preserve Ffff
        xor             c               ; a = fffbBbbF

        ld              [hl+], a        ; store out New Single/PairColors

        ld              b, [hl]         ; b <- Old Single/PairInUseBy
        ld              [hl], PairInUseBy >> 8  ; *
        inc             l

        ld              e, a            ; e = New PairColors

        call            BktGetCols_CS   ; reanalyze in CS-mode using SrchPal2

        pop             hl              ; restore CARD @ UniqueColors + Flags

        jr              .CS_Refresh

.CS_SrchPal4

        ; -------------------------------------------------------------------
        ;        Get new colors using SrchPal4 (COLORED SQUARES MODE)
        ; -------------------------------------------------------------------

        ; * S4_CN = CardNum

        ; extract CardNum from card's CARD address

        ld              a, h            ; a = 11.11NNN (NNN = CardNum's MSB)
        xor             $02             ;   = 11.11N?N
        xor             l               ;   = ??n??N?N

        ld              [S4_CN + 1], a  ; store out "CardNum"

        inc             l               ; skip UniqueColors + Flags

        ld              e, [hl]         ; e <- FGCol / Sq12Col
        inc             l
        ld              d, [hl]         ; d <- BGCol / Sq34Col

        call            BktGetCols_CLSQ

        pop             hl              ; restore CARD @ UniqueColors + Flags

        jr              .CS_Refresh

.CS_SrchPal1

        ; -------------------------------------------------------------------
        ;       Get new colors using SrchPal1 (COLOR STACK MODE)
        ; -------------------------------------------------------------------

        ; * S1_CN = CardNum

        ; extract CardNum from card's CARD address

        ld              a, h            ; a = 11.11NNN (NNN = CardNum's MSB)
        xor             $02             ;   = 11.11N?N
        xor             l               ;   = ??n??N?N

        ld              [S1_CN + 1], a  ; store out "CardNum"

        ld              d, 1            ; d = UniqueColors

        ld              a, [hl+]        ; a <- UniqueColors + Flags

        and             $18             ; a = ...nn...
        rrca                            ;   = ....nn..
        rrca                            ;   = .....nn.
        rrca                            ;   = ......nn (CSNdx)

        add             ColorStack & $FF
        ld              c, a            ; C -> ColorStack[ CSNdx ]

        ld              a, [c]          ; a <- BGCol << 1

        ld              c, [hl]         ; c <- Old Single/PairColors
        ld              [hl+], a        ; store out New Single/PairColors

        ld              b, [hl]         ; b <- Old Single/PairInUseBy
        ld              [hl], SingleInUseBy >> 8  ; *
        inc             l

        ld              e, a            ; e = New SingleColors

        call            BktGetCols_CSGROM0

        pop             hl              ; restore CARD @ UniqueColors + Flags

        jp              .CS_Refresh

.FGBG
        ; ------------------------------------------------------------------
        ;                 FOREGROUND / BACKGROUND MODE
        ; ------------------------------------------------------------------

.FGBG_Refresh
        ld              de, -1 * CARD_RECSIZE
.FGBG_Loop
        add             hl, de

        ld              a, [hl]         ; a <- UniqueColors + Flags
        add             a, a            ; is Re-analyze (bit #7) set ?
        jr              nc, .FGBG_Loop  ; no, so keep checking cards

        ret             z               ; hit BktGetCols_Stop (checked 240)

        ; card's Re-analyze (bit #7) is set

        res             7, [hl]         ; reset ReAnalyze bit

        push            hl              ; save CARD @ UniqueColors + Flags

        and             1 << 1              ; is UniqueColors = 1 ?
        jr              nz, .FGBG_SrchPal1  ; yes, so reanalyze w/ SrchPal1

        ; -------------------------------------------------------------------
        ;    Get new colors using SrchPal2 (FOREGROUND / BACKGROUND MODE)
        ; -------------------------------------------------------------------

        ; * S2_ML = MakeUpNdx LOW       * S2_MH = MakeUpNdx HIGH
        ; * S2_GB = GraphicBase         * S2_CN = CardNum

        dec             l               ; pos @ HIGH byte of DataAddr
        dec             l               ; pos @ LOW byte of DataAddr

        ; extract CardNum from card's CARD address

        ld              a, h            ; a = 11.11NNN (NNN = CardNum's MSBs)
        xor             l               ;   = ??n??NNN
        ld              [S2_CN + 1], a  ; store out "CardNum"

        ; extract MakeUpNdx and GraphicBase from card's DataAddr

        ld              a, [hl+]        ; a <- LOW byte of DataAddr

        add             a, a            ; a = nnnn.... (CY = N)
        ld              [S2_ML + 1], a  ; store out "MakeUpNdx LOW"

        ld              a, [hl+]        ; a <- HIGH byte of DataAddr

        ld              de, ( MakeUp_GROM - GRAM_8 << 1 ) & $FF00 + ( ( GRAM_8 - MakeUp_GROM >> 1 ) >> 8 ) & $00FF

        rla                             ; put N in bit #0
        jr              c, .FGBG_InGRAM ; if CY = 1, DataAddr >= $8000

        ld              de, ( MakeUp_GROM - GROM_8 << 1 ) & $FF00 + ( ( GROM_8 - MakeUp_GROM >> 1 ) >> 8 ) & $00FF
.FGBG_InGRAM
        add             d               ; a = HIGH byte of MakeUpNdx
        ld              [S2_MH + 1], a  ; store out "MakeUpNdx HIGH"

        ld              a, e            ; a = "GraphicBase"
        ld              [S2_GB + 1], a  ; store out "GraphicBase"

        ; finally, get UniqueColors + Flags, FGCol, and BGCol from CARD

        ld              d, 2            ; d = UniqueColors
        inc             l               ; skip UniqueColors + Flags

        ld              c, [hl]         ; c <- Old PairColors
        inc             l
        ld              b, [hl]         ; b <- Old PairInUseBy
        ld              [hl], PairInUseBy >> 8  ; *
        inc             l

        ld              e, c            ; e = Old PairColors = New PairColors

        call            BktGetCols_FGBG

        pop             hl              ; restore CARD @ UniqueColors + Flags

        jr              .FGBG_Refresh

.FGBG_SrchPal1

        ; -------------------------------------------------------------------
        ;    Get new colors using SrchPal1 (FOREGROUND / BACKGROUND MODE)
        ; -------------------------------------------------------------------

        ; * S1_CN = CardNum

        ; extract CardNum from card's CARD address

;        ld              a, h            ; a = 11.11NNN (NNN = CardNum's MSB)
        xor             h               ;   = 11.11N?N
        xor             l               ;   = ??n??N?N

        ld              [S1_CN + 1], a  ; store out "CardNum"

        ld              d, 1            ; d = UniqueColors
        inc             l               ; skip UniqueColors + Flags

        ld              c, [hl]         ; c <- Old Single/PairColors
        inc             l
        ld              b, [hl]         ; b <- Old Single/PairInUseBy
        ld              [hl], SingleInUseBy >> 8  ; *
        inc             l

        ld              e, c            ; e = Old PairColors = New PairColors

        call            BktGetCols_FGBGGROM0

        pop             hl              ; restore CARD @ UniqueColors + Flags

        jr              .FGBG_Refresh

; **
; BktWrite_CLSQ:
;
; Parameters:     hl = SYSTEM_RAM address
;                 de = BACKTAB attribute word
;
; Modifies:       All registers.
;

BktWrite_CLSQ::

        ; setup a pointer to CARD so we can start updating it

        add             hl, hl          ; hl = SYSTEM_RAM << 1 + CardNum << 2
        add             hl, hl          ;    = SYSTEM_RAM << 2 + CardNum << 3

        ld              a, h
        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              h, a            ; hl -> CARD[ CardNum * 8 ]

        ; extract CardNum from CARD address

        xor             l               ; a = ??n??NNN (NNN = CardNum's MSBs)
        ld              [S4_CN + 1], a  ; store out "CardNum"

        ; calculate "DataAddr" for card (used only for collision checking)

        ld              a, d            ; a = ??4!0443 (HIGH of BACKTAB word)

        xor             e
        and             $3F             ; preserve all bits but 6-7
        xor             e               ;   = 334!.443

        rlca                            ;   = 34!.4433
        rlca                            ;   = 4!.44333

        bit             7, a            ; is MSB of Sq4Col (bit #7) set ?
        jr              z, .NoSq4MSB    ; no, so leave bit #5 reset

        xor             $A0             ;   = .!444333
.NoSq4MSB
        ld              d, a            ; save Sq34Col in d

        ld              a, e            ; a = 33222111 (LOW of BACKTAB word)

        and             $3F             ;   = ..222111

        ld              c, a
        ld              b, SqGrafx >> 8 ; bc -> Sq12ToAddr[ ..222111 ]

        ld              a, [bc]         ; a = !12.....
        ld              [hl], a         ; save in LOW byte of DataAddr

        ld              c, d            ; bc -> Sq34ToAddr[ .!444333 ]

        ld              a, [bc]         ; a = !..34...
        or              [hl]            ;   = !1234...

        ld              [hl+], a        ; store out LOW byte of DataAddr
        ld              [hl], b         ; store out HIGH byte of DataAddr
        inc             l

        inc             l               ; skip UniqueColors + Flags

        ld              c, [hl]         ; c <- Old Sngl/PairColors / Sq12Col
        ld              [hl], e         ; store New Sngl/PairColors / Sq12Col

        inc             l

        ld              b, [hl]         ; b <- Old Sngl/PairInUseBy / Sq34Col
        ld              [hl], d         ; store New Sngl/PairInUseBy / Sq34Col

        ; "de-register" card from Single/PairInUseBy (freed when reaches 0)

        ld              a, [bc]         ; a <- Sngl/PairInUseBy
        dec             a
        ld              [bc], a         ; de-register from Sngl/PairInUseBy

; **
; BktGetCols_CLSQ:  Get new color indices for squares #1-#4, using
;                   the current color on the color stack whenever
;                   a square's color number is equal to 7.
;
; Parameters:       e  = 33222111 (Sq12Col)
;                   d  = .!444333 (Sq34Col)
;                   hl = CARD @ Single/PairInUseBy / Sq34Col
;
; Modifies:         All registers.
;
; Notes:            Called by BktGetCols and part of BktWrite_CLSQ.
;

BktGetCols_CLSQ:

        dec             l               ; skip Single/PairInUseBy / Sq34Col
        dec             l               ; skip Single/PairColors / Sq12Col

        push            hl              ; save CARD @ UniqueColors + Flags

        ld              a, [hl]         ; a <- UniqueColors + Flags

        and             $18             ; a = ...nn...
        rrca                            ;   = ....nn..
        rrca                            ;   = .....nn.
        rrca                            ;   = ......nn (CSNdx)

        add             ColorStack & $FF
        ld              c, a            ; c -> ColorStack[ CSNdx ]

        ld              a, [c]          ; a <- ColNum << 1
        ld              b, a            ; save it in b

        ; extract Square #1's color from BACKTAB word

        ld              hl, Case4Tab + $0E

        ld              a, e            ; a = 33222111 (LOW of BACKTAB word)
        add             a, a            ;   = 3222111.
        and             l               ;   = ....111. (Sq1Col)

        cp              l               ; is Sq1Col = 7 ?
        jr              nz, .Sq1Col06   ; no, so don't replace it

        inc             h               ; set "BackBit" (bit #0)

        ld              a, b            ; a <- ColorStack[ CSNdx ]
.Sq1Col06
        add             ColorToRGB & $FF
        ld              c, a            ; c -> ColorToRGB[ Sq1Col * 2 ]

        ld              a, [c]          ; a <- LOW byte of Sq1ColRGB
        ld              [S4_L1 + 1], a  ; store out "Low #1"
        ld              [S4_S1 + 1], a  ; store out "Search #1"

        inc             c               ; pos @ HIGH byte of Sq1ColRGB

        ld              a, [c]          ; a <- HIGH byte of Sq1ColRGB
        ld              [S4_H1 + 1], a  ; store out "High #1"

        ld              a, h
        ld              [S4_B1 + 1], a  ; store out "BackBit #1"

        ; extract Square #2's color from BACKTAB word

        ld              h, Case4Tab >> 8

        ld              a, e            ; a = 33222111 (LOW of BACKTAB word)
        rrca                            ;   = 13322211
        rrca                            ;   = 11332221
        and             l               ;   = ....222. (Sq2Col)

        cp              l               ; is Sq2Col = 7 ?
        jr              nz, .Sq2Col06   ; no, so don't replace it

        inc             h               ; set "BackBit" (bit #0)

        ld              a, b            ; a <- ColorStack[ CSNdx ]
.Sq2Col06
        add             ColorToRGB & $FF
        ld              c, a            ; c -> ColorToRGB[ Sq2Col * 2 ]

        ld              a, [c]          ; a <- LOW byte of Sq2ColRGB
        ld              [S4_L2 + 1], a  ; store out "Low #2"
        ld              [S4_S2 + 1], a  ; store out "Search #2"

        inc             c               ; pos @ HIGH byte of Sq2ColRGB

        ld              a, [c]          ; a <- HIGH byte of Sq2ColRGB
        ld              [S4_H2 + 1], a  ; store out "High #2"

        ld              a, h
        ld              [S4_B2 + 1], a  ; store out "BackBit #2"

        ; extract Square #3's color from BACKTAB word

        ld              h, Case4Tab >> 8

        ld              a, d            ; a = .!444333 (Sq34Col)
        add             a, a            ;   = !444333.
        and             l               ;   = ....333. (Sq3Col)

        cp              l               ; is Sq3Col = 7 ?
        jr              nz, .Sq3Col06   ; no, so don't replace it

        inc             h               ; set "BackBit" (bit #0)

        ld              a, b            ; a <- ColorStack[ CSNdx ]
.Sq3Col06
        add             ColorToRGB & $FF
        ld              c, a            ; c -> ColorToRGB[ Sq3Col * 2 ]

        ld              a, [c]          ; a <- LOW byte of Sq3ColRGB
        ld              [S4_L3 + 1], a  ; store out "Low #3"
        ld              [S4_S3 + 1], a  ; store out "Search #3"

        inc             c               ; pos @ HIGH byte of Sq3ColRGB

        ld              a, [c]          ; a <- HIGH byte of Sq3ColRGB
        ld              [S4_H3 + 1], a  ; store out "High #3"

        ld              a, h
        ld              [S4_B3 + 1], a  ; store out "BackBit #3"

        ; extract Square #4's color from BACKTAB word

        ld              h, Case4Tab >> 8

        ld              a, d            ; a = .!444333 (Sq34Col)
        rrca                            ;   = 3.!44433
        rrca                            ;   = 33.!4443
        and             l               ;   = ....444.

        cp              l               ; is Sq4Col = 7 ?
        jr              nz, .Sq4Col06   ; no, so don't replace it

        inc             h               ; set "BackBit" (bit #0)

        ld              a, b            ; a <- ColorStack[ CSNdx ]
.Sq4Col06
        add             ColorToRGB & $FF
        ld              c, a            ; c -> ColorToRGB[ Sq4Col * 2 ]

        ld              a, [c]          ; a <- LOW byte of Sq4ColRGB
        ld              [S4_L4 + 1], a  ; store out "Low #4"
        ld              [S4_S4 + 1], a  ; store out "Search #4"

        inc             c               ; pos @ HIGH byte of Sq4ColRGB

        ld              a, [c]          ; a <- HIGH byte of Sq3ColRGB
        ld              [S4_H4 + 1], a  ; store out "High #4"

        ld              a, h
        ld              [S4_B4 + 1], a  ; store out "BackBit #4"

        ; -------------------------------------------------------------------
        ;                Handle a change in Advance bit
        ; -------------------------------------------------------------------

        ; check if Advance was previously set (never set in ColoredSquares)

        pop             hl              ; restore CARD @ UniqueColors + Flags

        ld              a, [hl+]        ; a <- UniqueColors + Flags

        bit             5, a            ; was Advance (bit #5) set ?
        jr              z, .AdvSame     ; no, so CSNdx is valid

        ; Advance changed (reset), so decrement _this_ card's CSNdx

        ld              c, $98          ; c = 1..11...
        add             c               ; dec _this_ card's CSNdx

        dec             l               ; re-pos @ UniqueColors + Flags

        push            hl              ; save CARD @ UniqueColors + Flags

        ld              [hl+], a        ; update UniqueColors + Flags

        ; now decrement CSNdx and set ReAnalyze for any remaining cards

        ld              de, CARD_RECSIZE - 1
.AdvLoop
        add             hl, de          ; advance to next card

        ld              b, [hl]         ; b <- UniqueColors + Flags

        ld              a, b
        and             $18             ; a = ...nn... (nn = OldCSNdx)
        add             c               ;   = 1.?NN... (NN = NewCSNdx)

        xor             b
        and             c               ; preserve Re-analyze and NewCSNdx
        xor             b               ; restore all other bits

        ld              [hl+], a        ; update UniqueColors + Flags

        and             e               ; is UniqueColors = 0 ?
        jr              nz, .AdvLoop    ; no, so card # still in 1..239

        pop             hl              ; restore CARD @ UniqueColors + Flags

        ld              a, [hl+]        ; a <- UniqueColors + Flags
.AdvSame

        ; -------------------------------------------------------------------
        ;           "De-register" card's color and tile indices
        ; -------------------------------------------------------------------

        ; prepare to "de-register" card's previous color and tile indices

        inc             l               ; skip Single/PairColors / Sq12Col
        inc             l               ; skip Single/PairInUseBy / Sq34Col

        and             $07             ; a = .....uuu (UniqueColors)
        ld              d, a            ; save UniqueColors in d

        ld              a, [hl+]        ; a <- ColorIndices
        rlca                            ; put 1st ColorIndex in bits 1-2

        ld              c, [hl]         ; c <- $C0 + ( PalNdx + 1 ) << 3

        push            hl              ; save CARD @ PaletteIndex

        ; "de-register" colors that card previously used from PALTAB_1

        ld              h, PALTAB_1 >> 8
.DeRegLoop
        ld              e, a            ; save ColorIndices in e
        and             $06             ; a = 11111cc1
        or              c               ;   = 11pppcc1

        ld              l, a            ; hl -> PALTAB_1 (InUseBy)

        dec             [hl]            ; "de-register" color from PALTAB_1

        ld              a, e
        rrca
        rrca                            ; advance to next ColorIndex

        dec             d               ; dec UniqueColors
        jr              nz, .DeRegLoop

        ; prepare a few final variables for our transfer to SrchPal4

        ld              l, PALTAB_1 & $FF  ; hl -> PALTAB_1 (InUseBy)

        ld              a, $FF          ; *
        ld              [S4_BC + 1], a  ; set "BestCase" = 255

        jp              SrchPal4        ; continue the process in RAM

; **
; BktWrite_CSGROM0:
;
; Parameters:       hl = SYSTEM_RAM address
;                   d  = HIGH byte of BACKTAB word
;
; Modifies:         All registers.
;

BktWrite_CSGROM0::

        ; setup pointer to CARD, so we can begin to update the record

        add             hl, hl          ; hl = SYSTEM_RAM << 1 + CardNum << 2
        add             hl, hl          ;    = SYSTEM_RAM << 2 + CardNum << 3

        ld              a, h
        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              h, a            ; hl -> CARD[ CardNum * 8 ]

        ; extract CardNum from the CARD address

        xor             l               ; a = ??n??NNN (CardNum's MSBs)
        ld              [S1_CN + 1], a  ; store it out to "CardNum"

        ;            COLOR STACK / COLORED SQUARES MODE
        ;
        ;       15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
        ;        ?  ?  a cF  t  N  N  n  n  n  n  n  n  f  f  f

        ; set CARD's DataAddr = GROM #0 (already determined)

        xor             a
        ld              [hl+], a        ; store out LOW byte of DataAddr
        ld              [hl+], a        ; store out HIGH byte of DataAddr

        ; check if "Advance" (bit #13 of BACKTAB word) changed

        ld              b, [hl]         ; b <- UniqueColors + Flags

        ld              a, b            ; a = rcannuuu (a = OldAdv)
        xor             d
        and             $20             ; did Advance (bit #5) change ?
        jr              z, .AdvSame     ; no, so CSNdx is valid

        ; new Advance, so inc / dec _this_ card's CSNdx and toggle Advance

        and             d               ; a = ..a..... (a = NewAdv)
        rrca                            ;   = ...a....
        add             $98             ;   = 1.??1...
        ld              c, a            ; save "inc / dec CSNdx" byte in c

        add             b               ; inc / dec CSNdx for _this_ card

        xor             b
        and             $18             ; preserve NewCSNdx in bits 3-4
        xor             b               ; restore all other bits
        xor             $20             ; toggle Advance (bit #5)

        ; now inc / dec CSNdx and set ReAnalyze for any remaining cards

        push            hl              ; save CARD @ UniqueColors + Flags

        ld              [hl+], a        ; update UniqueColors + Flags

        ld              de, CARD_RECSIZE - 1  ; e must = $07 (used as a mask)
.AdvLoop
        add             hl, de          ; pos CARD @ next card's record

        ld              b, [hl]         ; b <- UniqueColors + Flags

        ld              a, b            ; a = rcannuuu
        and             $18             ;   = ...nn... (nn = old CSNdx)
        add             c               ;   = 1??NN... (NN = new CSNdx)

        xor             b
        and             $98             ; preserve ReAnalyze and CSNdx
        xor             b               ; restore all other bits

        ld              [hl+], a        ; update UniqueColors + Flags

        and             e               ; was UniqueColors = 0 ?
        jr              nz, .AdvLoop    ; no, so card # still in 1..239

        pop             hl              ; restore CARD @ UniqueColors + Flags

        ld              b, [hl]         ; b <- UniqueColors + Flags

.AdvSame

        ld              a, b            ; a = rcannuuu
        and             $38             ;   = ..ann...
        inc             a               ;   = ..ann..1 (UniqueColors = 1)

        ld              [hl+], a        ; set UniqueColors = 1, clear Flags

        and             $18             ;   = ...nn...
        rrca                            ;   = ....nn..
        rrca                            ;   = .....nn.
        rrca                            ;   = ......nn (CSNdx)

        add             ColorStack & $FF
        ld              c, a            ; c -> ColorStack[ CSNdx ]

        ld              a, [c]          ; a <- BGCol << 1

        ld              e, a            ; put a copy of BGCol << 1 in e

        ld              c, [hl]         ; c <- Old Single/PairColors / Sq12Col
        ld              [hl+], a        ; -> New Single/PairColors / Sq12Col

        ld              a, b            ; a = UniqueColors + Flags
        and             $07             ; isolate UniqueColors in bits 0-2
        ld              d, a            ; and save it in d

        ld              b, [hl]                   ; <- Old Single/PairInUseBy
                                                  ;    / Sq34Col
        ld              [hl], SingleInUseBy >> 8  ; -> New Single/PairInUseBy
                                                  ;    / Sq34Col
        inc             l

; **
; BktGetCols_CSGROM0:  Get a new color index for the background color.
;
; Parameters:          c  = Old Single/PairColors
;                      b  = Old Single/PairInUseBy
;                      e  = New Single/PairColors (BGCol << 1)
;                      d  = UniqueColors
;                      hl = CARD @ PaletteIndex
;
; Modifies:            All registers.
;
; Notes:               Called by BktGetCols and part of BktWrite_CSGROM0
;

BktGetCols_CSGROM0::

        ; "de-register" card from Single/PairInUseBy (freed when reaches 0)

        ld              a, [bc]         ; a <- Single/PairInUseBy
        dec             a
        ld              [bc], a         ; de-register from Single/PairInUseBy

        ; "de-register" cards previous ColorIndices from PALTAB_1

        ld              a, [hl+]        ; a <- ColorIndices
        rlca                            ; put 1st ColorIndex in bits 1-2

        ld              c, [hl]         ; c <- $C1 + PalNdx << 3

        push            hl              ; save CARD @ PaletteIndex

        ld              h, PALTAB_1 >> 8
.DeRegLoop
        ld              b, a            ; save ColorIndices in b
        and             $06             ; a = .....cc.
        or              c               ;   = 11pppcc1

        ld              l, a            ; hl -> PALTAB_1 (InUseBy)

        dec             [hl]            ; de-register index from PALTAB_1

        ld              a, b
        rrca
        rrca                            ; put next ColorIndex in bits 1-2

        dec             d               ; de-registered all colors ?
        jr              nz, .DeRegLoop  ; no, so keep looping

        ; check SingleInUseBy to see if BGColor has already been found

        ld              l, e            ; hl -> SingleInUseBy[ ...bBbb. ]

        inc             [hl]            ; register card w/ SingleInUseBy
        jr              z, .SearchPal   ; 1st card, so need to search palette

        inc             l               ; hl -> SingleIndices[ ...bBbb1 ]

        ld              a, [hl]         ; a <- 11pppbb1
        ld              e, a            ; save a copy of SingleIndices in e
        ld              l, a            ; hl -> BG's PALTAB_1 (InUseBy)

        inc             [hl]            ; register card w/ BG's color slot

        ; restore card and update PaletteIndex & ColorIndices

        pop             hl              ; restore CARD @ PaletteIndex

        ld              [hl-], a        ; store out PaletteIndex
        ld              [hl], d         ; store out ColorIndices (0)

        ; update Background Display Data for tile indices

        ld              hl, S1_CN + 1   ; *

        ld              l, [hl]
        ld              h, CNToBDD >> 8 ; hl -> CNToBDD[ CardNum ]

        ld              c, [hl]
        inc             h
        ld              b, [hl]         ; bc -> BDD_TileNdx

        and             $C7             ; a = 11...bb1 (TileNdx)

        ld              [bc], a         ; store out TileNdx to BDD_TileNdx

        ; update Background Display Data for tile attributes

        xor             e               ; a = ..ppp...
        rrca                            ;   = ...ppp..
        rrca                            ;   = ....ppp.
        rrca                            ;   = .....ppp (PalNdx)

        set             1, b            ; bc -> BDD_TileAttr

        ld              [bc], a         ; store out PalNdx to BDD_TileAttr

        ret

.SearchPal

        ld              a, e            ; a = ....bBbb. (New SingleColors)
        inc             a               ; pos @ SingleIndices
        ld              [S1_CS + 1 ], a ; store out to "ColorSingle"

        add             ColorToRGB & $FF - 1
        ld              c, a            ; c -> ColorToRGB[ BGCol * 2 ]

        ld              a, [c]          ; a <- LOW byte of BGColRGB
        ld              [S1_BG + 1], a  ; store it out to "BackGround"
        ld              [S1_BL + 1], a  ; store it out to "BackGround LOW"

        inc             c               ; pos @ HIGH byte of BGColRGB

        ld              a, [c]          ; a <- HIGH byte of BGColRGB
        ld              [S1_BH + 1], a  ; store it out to "BackGround HIGH"

        ld              bc, Case1Tab + 1
        ld              de, 4 << 12 + $80FF  ; d = BestCase (Level 4 = worst)
        ld              l, d                 ; hl -> PALTAB_1 (InUseBy)

        jp              S1_ColLoop      ; search palette for one color

; **
; BktWrite_CS:
;
; Parameters:     hl = SYSTEM_RAM address
;                 de = BACKTAB attribute word
;
; Modifies:       All registers.
;

BktWrite_CS:

;        ld              a, l
;        cp              ( SYSTEM_RAM + ( 20 * 11 + 19 ) << 1 ) & $00FF
;        jr              nz, .Ahead
;
;        ld              a, h
;        cp              ( SYSTEM_RAM + ( 20 * 11 + 19 ) << 1 ) >> 8
;        jr              nz, .Ahead
;
;        ld              b, b            ; BREAK !!!
;.Ahead
        ld              bc, $03 << 8 + ( ( GRAM_8 - MakeUp_GROM >> 1 ) >> 8 ) & $00FF

        bit             3, d            ; is GRAM toggle = 1 ?
        jr              nz, .InGRAM     ; yes, so skip GROM-specific checks

        ; GRAM toggle reset, so check 1st if it's using ColoredSquares mode

        bit             4, d               ; is F (of Ffff) set ?
        jp              nz, BktWrite_CLSQ  ; yes, so process as ColoredSquares

        ; card comes from GROM, so check if it's using GROM #0 (Ffff ignored)

        ld              a, e            ; a = nnnnnfff (LOW of BACKTAB word)

        xor             d
        and             $F8             ; preserve bits 0-4 of CardNdx
        xor             d               ;   = nnnnnNNN

        jp              z, BktWrite_CSGROM0

        ld              bc, $0F << 8 + ( ( GROM_8 - MakeUp_GROM >> 1 ) >> 8 ) & $00FF

.InGRAM

        ; setup a pointer to CARD, so we can begin to update the record

        add             hl, hl          ; hl = SYSTEM_RAM << 1 + CardNum << 2
        add             hl, hl          ;    = SYSTEM_RAM << 2 + CardNum << 3

        ld              a, h
        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              h, a            ; hl -> CARD[ CardNum * 8 ]

        ; extract CardNum from the CARD address

        xor             l               ; a = ??n??NNN (NNN = CardNum's MSBs)
        ld              [S2_CN + 1], a  ; store it out to "CardNum"

        ;            COLOR STACK / COLORED SQUARES MODE
        ;
        ;       15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
        ;        ?  ?  a cF  t  N  N  n  n  n  n  n  n  f  f  f

        ; extract DataAddr, MakeUpNdx, and GraphicBase from BACKTAB word

        ld              a, e            ; a = Nnnnnfff
        and             $F8             ;   = Nnnnn...
        ld              [hl+], a        ; store out LOW byte of DataAddr

        add             a, a            ;   = nnnn.... (CY = N)
        ld              [S2_ML + 1], a  ; store out "MakeUpNdx LOW"

        ld              a, d              ; a = ??a..nnn / ??aF!??n
        rla                               ;   = ?a..nnnN / ?aF!??nN
        and             b                 ;   = ....nnnN / ......nN
        add             MakeUp_GROM >> 8  ;   = 11.1nnnN / 11.1..nN

        ld              [S2_MH + 1], a  ; store out "MakeUpNdx HIGH"

        rra                             ; a = .11.1nnn / .11.1..n
        add             c               ; add "GraphicBase"
        ld              [hl+], a        ; store out HIGH byte of DataAddr

        ld              a, c            ; a = LOW byte of GRAM or GROM
        ld              [S2_GB + 1], a  ; store out "GraphicBase"

        ; check if "Advance" (bit #13 of BACKTAB word) changed

        ld              b, [hl]         ; b <- UniqueColors + Flags

        ld              a, b            ; a = rcannuuu (a = OldAdv)
        xor             d
        and             $20             ; did Advance (bit #5) change ?
        jr              z, .AdvSame     ; no, so CSNdx is valid

        ; new Advance, so inc / dec _this_ card's CSNdx and toggle Advance

        and             d               ; a = ..a..... (a = NewAdv)
        rrca                            ;   = ...a....
        add             $98             ;   = 1.??1...
        ld              c, a            ; save "inc / dec CSNdx" byte in c

        add             b               ; inc / dec CSNdx for _this_ card

        xor             b
        and             $18             ; preserve NewCSNdx in bits 3-4
        xor             b               ; restore all other bits
        xor             $20             ; toggle Advance (bit #5)

        ; now inc / dec CSNdx and set ReAnalyze for any remaining cards

        push            hl              ; save CARD @ UniqueColors + Flags
        push            de              ; save BACKTAB attribute word

        ld              [hl+], a        ; update UniqueColors + Flags

        ld              de, CARD_RECSIZE - 1  ; e must = $07 (used as a mask)
.AdvLoop
        add             hl, de          ; pos CARD @ next card's record

        ld              b, [hl]         ; b <- UniqueColors + Flags

        ld              a, b            ; a = rcannuuu
        and             $18             ;   = ...nn... (nn = old CSNdx)
        add             c               ;   = 1??NN... (NN = new CSNdx)

        xor             b
        and             $98             ; preserve ReAnalyze and CSNdx
        xor             b               ; restore all other bits

        ld              [hl+], a        ; update UniqueColors + Flags

        and             e               ; was UniqueColors = 0 ?
        jr              nz, .AdvLoop    ; no, so card # still in 1..239

        pop             de              ; restore BACKTAB attribute word
        pop             hl              ; restore CARD @ UniqueColors + Flags

        ld              b, [hl]         ; b <- UniqueColors + Flags

.AdvSame

        ; see if we can avoid searching the palette for new color indices

        ld              a, b            ; a = rcannuuu
        and             $C5             ; is   ( ReAnalyze = 1 )
                                        ;   or ( ColoredSquares = 1 )
                                        ;   or ( UniqueColors = 1 ) TRUE ?
        jr              nz, .GetColors  ; yes, so just extract FG & BG colors

        ; extract AND compare FG & BG colors from BACKTAB word w/ old values

        ld              a, [hl+]        ; a <- UniqueColors + Flags

        and             $18             ;   = ...nn...
        rrca                            ;   = ....nn..
        rrca                            ;   = .....nn.
        rrca                            ;   = ......nn (CSNdx)

        add             ColorStack & $FF
        ld              c, a            ; c -> ColorStack[ CSNdx ]

        ld              a, e            ; a = nnnnnfff
        and             $07             ;   = .....fff
        add             a, a            ;   = ....fff.

        xor             d
        and             $EF             ; preserve every bit except #4
        xor             d               ; a = ...Ffff. (FGCol << 1)

        ld              e, a            ; save FGCol << 1 in e

        ld              a, [c]          ; a <- BGCol << 1

        swap            e               ; e = fff....F *
        or              e               ; a = fffbBbbF

        cp              [hl]            ; is "new" FBGCol = old FBGCol ?
        jr              nz, .NewFBC     ; no, so need new color indices

        ; FG & BG colors are the same, so don't get new color indices

        ld              a, $C9          ; a = OpCode for RET
        ld              [S2_B1], a      ; store out "Bypass #1" (YES)

        inc             l               ; pos @ Single/PairInUseBy / Sq34Col
        inc             l               ; pos @ ColorIndices

        ld              a, [hl+]        ; a <- ColorIndices

        inc             l               ; pos @ PaletteIndex

        push            hl              ; save CARD @ TileNdx

        ld              l, [hl]         ; l <- TileNdx (1-255)
        ld              h, TileInUseBy >> 8

        dec             [hl]            ; de-register tile from TileInUseBy

        jp              S2_ML           ; continue @ "MakeUpNdx LOW"

.GetColors

        xor             b               ; a = ..ann.?.
        or              $02             ;   = ..ann.1. (UniqueColors = 2)

        ld              [hl+], a        ; set UniqueColors = 2, clear Flags

        ; extract FG and BG colors values from BACKTAB word (do not compare)

        and             $18             ;   = ...nn...
        rrca                            ;   = ....nn..
        rrca                            ;   = .....nn.
        rrca                            ;   = ......nn (CSNdx)

        add             ColorStack & $FF
        ld              c, a            ; c -> ColorStack[ CSNdx ]

        ld              a, e            ; a = nnnnnfff
        and             $07             ;   = .....fff
        add             a, a            ;   = ....fff.

        xor             d
        and             $EF             ; preserve every bit except #4
        xor             d               ; a = ...Ffff. (FGCol << 1)

        ld              e, a            ; save FGCol << 1 in e

        ld              a, [c]          ; a <- BGCol << 1

        swap            e               ; e = fff....F *
        or              e               ; a = fffbBbbF
.NewFBC
        ld              e, a

        ld              c, [hl]         ; <- Old Single/PairColors / Sq12Col
        ld              [hl+], a        ; -> New Single/PairColors / Sq12Col

        ld              a, b            ; a = UniqueColors + Flags
        and             $07             ; isolate UniqueColors in bits 0-2
        ld              d, a            ; and save it in d

        ld              b, [hl]                ; <- Old Single/PairInUseBy
                                               ;    / Sq34Col
        ld              [hl], PairInUseBy >> 8 ; -> New Single/PairInUseBy
                                               ;    / Sq34Col
        inc             l

; **
; BktGetCols_CS:  Get new color indices for both the foreground
;                 and background colors.
;
; Parameters:     b  = Old Single/PairInUseBy / Sq34Col
;                 c  = Old Single/PairColors / Sq12Col
;                 d  = UniqueColors
;                 e  = New SingleColors
;                 hl = CARD @ PalNdx
;
; Modifies:       All registers.
;
; Notes:          Called by BktGetCols and part of BktWrite_CS.
;

BktGetCols_CS:

        ; "de-register" card from Single/PairInUseBy (freed if it reaches 0)

        ld              a, [bc]
        dec             a
        ld              [bc], a         ; de-reg. from Single/PairInUseBy

        ld              a, [hl+]        ; a <- ColorIndices
        rlca                            ; put 1st ColorIndex in bits 1-2

        ld              c, [hl]         ; c <- $C0 + ( PalNdx + 1 ) << 3
        inc             l

        push            hl              ; save CARD @ TileNdx

        ; "de-register" card from TileInUseBy (free up tile if it reaches 0)

        ld              l, [hl]         ; l <- TileNdx (1-255)
        ld              h, TileInUseBy >> 8

        dec             [hl]            ; de-register card from TileInUseBy

        ; "de-register" card's previous ColorIndices from PALTAB_1

        ld              h, PALTAB_1 >> 8
.DeRegLoop
        ld              b, a            ; save ColorIndices in b
        and             $06             ; a = .....cc.
        or              c               ;   = 11pppcc1

        ld              l, a            ; hl -> PALTAB_1 (InUseBy)

        dec             [hl]            ; de-register index from PALTAB_1

        ld              a, b
        rrca
        rrca                            ; put next ColorIndex in bits 1-2

        dec             d               ; de-registered all colors ?
        jr              nz, .DeRegLoop  ; no, so keep looping

        ; check if valid color & palette indices exist for this color pair

        ld              l, e
        inc             h               ; hl -> PairInUseBy[ fffbBbbF ]

        inc             [hl]            ; "register" card with PairInUseBy

        jr              z, .SearchPal   ; if 1st card, need to search palette

        ; found valid color & palette indices for this color pair

        xor             a               ; a = OpCode for NOP
        ld              [S2_B1], a      ; store out "Bypass #1" (NO)
        ld              a, $C9          ; a = OpCode for RET
        ld              [S2_B2], a      ; store out "Bypass #2" (YES)

        ; -------------------------------------------------------------------
        ;                  Update palette information
        ; -------------------------------------------------------------------

        inc             h               ; hl -> PairIndices[ fffbBbbF ]

        ld              a, [hl]         ; a <- 1pppffbb

        ld              c, a            ; save a copy of ColorIndices in c

        rrca                            ; a = b1pppffb
        or              $C1             ;   = 11pppff1

        ld              l, a
        ld              h, PALTAB_1 >> 8  ; hl -> FG's PALTAB_1

        and             $F9             ; a = 11ppp..1

        ld              b, a            ; save $C1 + PaletteIndex << 3 in b

        ld              [S2_PI + 1], a  ; store out "PaletteIndex"

        ; register card w/ PALTAB_1 with indices pulled from PairIndices

        ld              a, c            ; a = 1pppffbb
        rlca                            ;   = pppffbb1
        and             $06             ;   = .....bb.
        or              b               ;   = 11pppbb1

        inc             [hl]            ; register card w/ FG's color slot

        ld              l, a            ; l = LOW byte of BG's PALTAB_1

        inc             [hl]            ; register card w/ BG's color slot

        ld              a, c            ; a = 1pppffbb
        and             $0F             ;   = ....ffbb (ColorIndices)

        ld              h, TileInUseBy >> 8 ; *

        jp              S2_ML           ; continue @ "MakeUpNdx LOW"

.SearchPal

        ; did NOT find valid color & palette indices for this color pair

        xor             a               ; a = OpCode for NOP
        ld              [S2_B1], a      ; store out "Bypass #1" (NO)
        ld              [S2_B2], a      ; store out "Bypass #2" (NO)

        ld              a, e            ; a = fffbBbbF

        ld              [S2_CP + 1], a  ; store out to "ColorPair"

        and             $1E             ;   = ...bBbb. (BGCol << 1)

        add             ColorToRGB & $FF
        ld              c, a            ; c -> ColorToRGB[ BGCol * 2 ]

        ld              a, [c]          ; a <- LOW byte of BGColRGB
        ld              [S2_BG + 1], a  ; store it out to "BackGround"
        ld              [S2_BL + 1], a  ; store it out to "BackGround LOW"

        ld              d, a            ; save a copy of LOW of BGColRGB in d

        inc             c               ; pos @ HIGH byte of BGColRGB

        ld              a, [c]          ; a <- HIGH byte of BGColRGB
        ld              [S2_BH + 1], a  ; store it out to "BackGround HIGH"

        ld              a, e            ; a = fffbBbbF
        swap            a               ;   = BbbFfffb
        or              $E1             ;   = 111Ffff1 ($E1 + FGCol << 1)

        add             ( ColorToRGB - $E1 ) & $FF  ; CY = 1
        ld              c, a                        ; c -> ColorToRGB
                                                    ;      [ FGCol * 2 ]

        ld              a, [c]          ; a <- LOW byte of FGColRGB
        ld              [S2_FL + 1], a  ; store it out to "ForeGround LOW"

        sbc             d               ; a =   LOW byte of FGColRGB
                                        ;     - LOW byte of BGColRGB
                                        ;     - 1

        ld              [S2_FG + 1], a  ; store it out to "ForeGround"

        inc             c               ; pos @ HIGH byte of BGColRGB

        ld              a, [c]          ; a <- HIGH byte of FGColRGB
        ld              [S2_FH + 1], a  ; store it out to "ForeGround HIGH"

        ld              hl, PALTAB_0         ; hl -> PALTAB_0 (RGB)
        ld              de, 7 << 12 + $80FF  ; d = BestCase (Level 7 = worst)

        jp              S2_PalLoop

; **
; ToggleGraphicsMode:
;

ToggleGraphicsMode:

        ld              a, [STIC_Mode_New]

        ld              hl, STIC_Mode

        cp              [hl]            ; has graphics mode changed ?
        ret             z               ; no, so nothing to do

        ld              [hl], a         ; store out new mode to STIC_Mode

        ; format of PairInUseBy/Indices differs between modes, so invalidate

        ld              a, $FF
        ld              hl, PairInUseBy
        ld              bc, 256

        call            MemFill         ; clear PairInUseBy w/ $FF

        ; now re-initialize CARD structure (in WRAM bank #3)

        ld              hl, CARD + 2
        ld              bc, PALTAB_DMY0 & $FF00 + 240  ; do 240 cards
        ld              de, CARD_RECSIZE

        ld              a, $81
.CardLoop
        bit             1, [hl]
        jr              z, .Skip

        set             7, [hl]         ; set Re-analyze (bit #7)
        inc             l
        ld              [hl+], a
        ld              [hl], b         ; <- PALTAB_DMY0
        dec             l
        dec             l
.Skip
        add             hl, de          ; pos @ next CARD's Single/Pair

        dec             c
        jr              nz, .CardLoop

        ldh             a, [STIC_Mode]  ; a <- STIC_Mode
        or              a               ; are we now in FGBG mode ?
        jr              nz, .FGBG       ; yes, so switch to FGBG

        ; in FG/BG mode, so switch to CS mode

        ld              hl, SYSTEM_RAM + 1
.CS_Loop
        push            hl              ; save pos in SYSTEM_RAM

        ld              a, 1
        ldh             [cSVBK], a
        ldh             [rSVBK], a      ; switch to WRAM bank w/ SYSTEM_RAM

        ld              d, [hl]
        dec             l
        ld              e, [hl]         ; de <- BACKTAB attribute word

        ld              a, 3
        ldh             [cSVBK], a
        ldh             [rSVBK], a      ; switch to WRAM bank w/ CARD, etc.

        call            BktWrite_CS     ; re-write it in CS mode

        pop             hl              ; restore pos in SYSTEM_RAM

        inc             hl              ; pos @ LOW of next BACKTAB word
        inc             l               ; pos @ HIGH of next BACKTAB word

        ld              a, l
        cp              ( SYSTEM_RAM + $00F0 * 2 + 1 ) & $FF
        jr              nz, .CS_Loop

        ld              a, h
        cp              ( SYSTEM_RAM + $00F0 * 2 + 1 ) >> 8
        jr              nz, .CS_Loop

        ret

.FGBG

        ; in CS mode, so switch to FG/BG mode

        ld              hl, SYSTEM_RAM + 1
.FGBG_Loop
        push            hl              ; save pos in SYSTEM_RAM

        ld              a, 1
        ldh             [cSVBK], a
        ldh             [rSVBK], a      ; switch to WRAM bank w/ SYSTEM_RAM

        ld              d, [hl]
        dec             l
        ld              e, [hl]         ; de <- BACKTAB attribute word

        ld              a, 3
        ldh             [cSVBK], a
        ldh             [rSVBK], a      ; switch to WRAM bank w/ CARD, etc.

        call            BktWrite_FGBG   ; re-write it in FG/BG mode

        pop             hl              ; restore pos in SYSTEM_RAM

        inc             hl              ; pos @ LOW of next BACKTAB word
        inc             l               ; pos @ HIGH of next BACKTAB word

        ld              a, l
        cp              ( SYSTEM_RAM + $0F0 * 2 + 1 ) & $FF
        jr              nz, .FGBG_Loop

        ld              a, h
        cp              ( SYSTEM_RAM + $0F0 * 2 + 1 ) >> 8
        jr              nz, .FGBG_Loop

        ret

; **
; BktWrite_FGBGGROM0:
;
; Parameters:       hl = SYSTEM_RAM address
;                   d  = HIGH byte of BACKTAB word
;
; Modifies:         All registers.
;

BktWrite_FGBGGROM0::

        ; setup pointer to CARD, so we can begin to update the record

        add             hl, hl          ; hl = SYSTEM_RAM << 1 + CardNum << 2
        add             hl, hl          ;    = SYSTEM_RAM << 2 + CardNum << 3

        ld              a, h
        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              h, a            ; hl -> CARD[ CardNum * 8 ]

        ; extract CardNum from the CARD address

        xor             l               ; a = ??n??NNN (NNN = CardNum's MSBs)
        ld              [S1_CN + 1], a  ; store it out to "CardNum"

        ;               FOREGROUND / BACKGROUND MODE
        ;
        ;       15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
        ;        ?  ?  B  b  t  b  b  n  n  n  n  n  n  f  f  f

        ; set CARD's DataAddr = GROM #0 (already determined)

        xor             a
        ld              [hl+], a        ; store out LOW byte of DataAddr
        ld              [hl+], a        ; store out HIGH byte of DataAddr

        ld              b, [hl]         ; b <- UniqueColors + Flags

        inc             a               ; = ld a, 1
        ld              [hl+], a        ; set UniqueColors = 1, clear Flags

        ld              a, d            ; a = ??Bb.bb. (HIGH of BACKTAB word)
        and             $36             ;   = ..Bb.bb.
        ld              e, a            ; save New SingleColors in e

        ld              c, [hl]         ; <- Old Single/PairColors / Sq12Col
        ld              [hl+], a        ; -> New Single/PairColors / Sq12Col

        ld              a, b            ; a = UniqueColors + Flags
        and             $07             ; isolate UniqueColors in bits 0-2
        ld              d, a            ; and save it in d

        ld              b, [hl]                   ; <- Old Single/PairInUseBy
                                                  ;    / Sq34Col
        ld              [hl], SingleInUseBy >> 8  ; -> New Single/PairInUseBy
                                                  ;    / Sq34Col
        inc             l

; **
; BktGetCols_FGBGGROM0:  Get a new color index for the background color.
;
; Parameters:            b  = Old Single/PairInUseBy / Sq34Col
;                        c  = Old Single/PairColors / Sq12Col
;                        d  = UniqueColors
;                        e  = New SingleColors
;                        hl = CARD @ PalNdx
;
; Modifies:              All registers.
;
; Notes:                 Called by BktGetCols and part of BktWrite_FGBGGROM0
;

BktGetCols_FGBGGROM0::

        ; "de-register" card from Single/PairInUseBy (freed when reaches 0)

        ld              a, [bc]         ; a <- Single/PairInUseBy
        dec             a
        ld              [bc], a         ; de-register from Single/PairInUseBy

        ; "de-register" cards previous ColorIndices from PALTAB_1

        ld              a, [hl+]        ; a <- ColorIndices
        rlca                            ; put 1st ColorIndex in bits 1-2

        ld              c, [hl]         ; c <- $C0 + ( PalNdx + 1 ) << 3

        push            hl              ; save CARD @ PaletteIndex

        ld              h, PALTAB_1 >> 8
.DeRegLoop
        ld              b, a            ; save ColorIndices in b
        and             $06             ; a = .....cc.
        or              c               ;   = 11pppcc1

        ld              l, a            ; hl -> PALTAB_1 (InUseBy)

        dec             [hl]            ; de-register index from PALTAB_1

        ld              a, b
        rrca
        rrca                            ; put next ColorIndex in bits 1-2

        dec             d               ; de-registered all colors ?
        jr              nz, .DeRegLoop  ; no, so keep looping

        ; check if BGColor has already been found & we can use its indices

        ld              l, e            ; hl -> SingleInUseBy[ ..Bb.bb. ]

        inc             [hl]            ; register card w/ SingleInUseBy
        jr              z, .SearchPal   ; 1st card, so need to search palette

        inc             l               ; hl -> SingleIndices[ ..Bb.bb1 ]

        ld              a, [hl]         ; a <- 11pppbb1
        ld              e, a            ; save a copy of SingleIndices in e
        ld              l, a            ; hl -> BG's PALTAB_1 (InUseBy)

        inc             [hl]            ; register card w/ BG's color slot

        ; restore card and update PaletteIndex & ColorIndices

        pop             hl              ; restore CARD @ PaletteIndex

        ld              [hl-], a        ; store out PaletteIndex
        ld              [hl], d         ; store out ColorIndices (0)

        ld              hl, S1_CN + 1

        ld              l, [hl]
        ld              h, CNToBDD >> 8 ; hl -> CNToBDD[ CardNum ]

        ld              c, [hl]
        inc             h
        ld              b, [hl]         ; bc -> BDD_TileNdx

        ; update Background Display Data for tile indices

        and             $C7             ; a = 11...bb1 (TileNdx)

        ld              [bc], a         ; store out TileNdx to BDD_TileNdx

        ; update Background Display Data for tile attributes

        xor             e               ; a = ..ppp...
        rrca                            ;   = ...ppp..
        rrca                            ;   = ....ppp.
        rrca                            ;   = .....ppp (PalNdx)

        set             1, b            ; bc -> BDD_TileAttr

        ld              [bc], a         ; store out PalNdx to BDD_TileAttr

        ret

.SearchPal

        ld              a, e            ; a = ..Bb.bb. (New SingleColors)
        inc             a               ;     ..Bb.bb1 *

        ld              [S1_CS + 1 ], a ; store out to "ColorSingle"

        bit             5, a            ; is B (of bBbb) set ?
        jr              z, .NoB         ; no, so leave bit #3 reset

        xor             $28             ;   = ...b1bb1 (BGCol << 1)
.NoB
        add             ColorToRGB & $FF - 1
        ld              c, a            ; c -> ColorToRGB[ BGCol * 2 ]

        ld              a, [c]          ; a <- LOW byte of BGColRGB
        ld              [S1_BG + 1], a  ; store it out to "BackGround"
        ld              [S1_BL + 1], a  ; store it out to "BackGround LOW"

        inc             c               ; pos @ HIGH byte of BGColRGB

        ld              a, [c]          ; a <- HIGH byte of BGColRGB
        ld              [S1_BH + 1], a  ; store it out to "BackGround HIGH"

        ld              bc, Case1Tab + 1
        ld              de, $C0FF       ; d = BestCase (Level 4 = worst)
        ld              l, d            ; hl -> PALTAB_1 (InUseBy)

        jp              S1_ColLoop      ; search palette for one color

; **
; BktWrite_FGBG:  Write to the background in FG/BG mode.
;
; Parameters:     hl = SYSTEM_RAM address (i.e., $D000-$D1DE)
;                 de = BACKTAB attribute word
;
;                 [rSVBK] must already be set to WRAM bank #3.
;
; Modifies:       All registers.
;

BktWrite_FGBG::

;        ld              a, l
;        cp              ( SYSTEM_RAM + ( 20 * 0 + 19 ) << 1 ) & $00FF
;        jr              nz, .Ahead
;
;        ld              a, h
;        cp              ( SYSTEM_RAM + ( 20 * 0 + 19 ) << 1 ) >> 8
;        jr              nz, .Ahead
;
;        ld              b, b            ; BREAK !!!
;.Ahead
        ld              c, ( GRAM_8 - MakeUp_GROM >> 1 ) >> 8

        bit             3, d            ; is GRAM toggle (bit #3) set ?
        jr              nz, .InGRAM     ; yes, so don't replace "GraphicBase"

        ld              a, d            ; a = ??Bb0bbN (HIGH of BACKTAB word)
        rra                             ;   = ???Bb0bb (CY = N)
        ld              a, e            ; a = nnnnnfff (LOW of BACKTAB word)
        rra                             ;   = Nnnnnnff

        and             $FC             ;   = Nnnnnn.. (CardNdx)

        jp              z, BktWrite_FGBGGROM0  ; use special rt. if GROM #0

        ld              c, ( GROM_8 - MakeUp_GROM >> 1 ) >> 8

.InGRAM

        ; setup pointer to CARD, so we can begin to update the record

        add             hl, hl          ; hl = SYSTEM_RAM << 1 + CardNum << 2
        add             hl, hl          ;    = SYSTEM_RAM << 2 + CardNum << 3

        ld              a, h
        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              h, a            ; hl -> CARD[ CardNum * 8 ]

        ; extract CardNum from the CARD address

        xor             l               ; a = ??n??NNN (NNN = CardNum's MSBs)
        ld              [S2_CN + 1], a  ; store it out to "CardNum"

        ;               FOREGROUND / BACKGROUND MODE
        ;
        ;       15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
        ;        ?  ?  B  b  t  b  b  n  n  n  n  n  n  f  f  f

        ; extract DataAddr, MakeUpNdx, and GraphicBase from BACKTAB word

        ld              a, e            ; a = Nnnnnfff
        and             $F8             ;   = Nnnnn...
        ld              [hl+], a        ; store out LOW byte of DataAddr

        add             a, a            ;   = nnnn.... (CY = N)
        ld              [S2_ML + 1], a  ; store out "MakeUpNdx LOW"

        ld              a, d              ; a = ??Bbtbbn
        rla                               ;   = ?BbtbbnN
        and             $03               ;   = ......nN
        add             MakeUp_GROM >> 8  ;   = 11.1..nN

        ld              [S2_MH + 1], a  ; store out "MakeUpNdx HIGH"

        rra                             ; a = .11.1..n
        add             c               ; add "GraphicBase"
        ld              [hl+], a        ; store out HIGH byte of DataAddr

        ld              a, c            ; a = LOW byte of GRAM or GROM
        ld              [S2_GB + 1], a  ; store out "GraphicBase"

        ; see if we can avoid searching the palette for new color indices

        ld              b, [hl]         ; b <- UniqueColors + Flags

        ld              a, b            ; a = rcannuuu
        and             $C1             ; is   ( ReAnalyze = 1 )
                                        ;   or ( ColoredSquares = 1 )
                                        ;   or ( UniqueColors = 1 ) TRUE ?
        jr              nz, .GetColors  ; yes, so just extract FG & BG colors

        inc             l               ; skip UniqueColors + Flags

        ; extract AND compare FG & BG colors from BACKTAB word w/ old values

        ld              a, e            ; a = nnnnnfff
        and             $07             ;   = .....fff
        rrca                            ;   = f.....ff
        rrca                            ;   = ff.....f

        xor             d               ;   = ??Bbtbb?
        and             $C9             ;   = ??..t..? (preserve fff)
        xor             d               ;   = ffBb.bbf (set Bb.bb)

        cp              [hl]            ; "New" PairColors = Old PairColors ?
        jr              nz, .NewFBC     ; no, so get new color indices

        ; FG & BG colors are the same, so don't get new color indices

        ld              a, $C9          ; a = OpCode for RET
        ld              [S2_B1], a      ; store out "Bypass #1" (YES)

        inc             l               ; pos @ Single/PairInUseBy / Sq34Col
        inc             l               ; pos @ ColorIndices

        ld              a, [hl+]        ; a <- ColorIndices

        inc             l               ; pos @ PaletteIndex

        push            hl              ; save CARD @ TileNdx

        ld              l, [hl]         ; l <- TileNdx (1-255)
        ld              h, TileInUseBy >> 8

        dec             [hl]            ; de-register tile from TileInUseBy

        jp              S2_ML           ; continue @ "MakeUpNdx LOW"

.GetColors

        ld              [hl], 2         ; set UniqueColors = 2, clear Flags
        inc             l

        ; extract FG and BG color values from BACKTAB word (do not compare)

        ld              a, e            ; a = nnnnnfff
        and             $07             ;   = .....fff
        rrca                            ;   = f.....ff
        rrca                            ;   = ff.....f

        xor             d               ;   = ??Bbtbb?
        and             $C9             ;   = ??..t..? (preserve fff)
        xor             d               ;   = ffBb.bbf (set Bb.bb)
.NewFBC
        ld              e, a            ; e = ffbB.bbf (New PairColors)

        ld              c, [hl]         ; <- Old Single/PairColors / Sq12Col
        ld              [hl+], a        ; -> New Single/PairColors / Sq12Col

        ld              a, b            ; a = UniqueColors + Flags
        and             $07             ; isolate UniqueColors in bits 0-2
        ld              d, a            ; and save it in d

        ld              b, [hl]                ; <- Old Single/PairInUseBy
                                               ;    / Sq34Col
        ld              [hl], PairInUseBy >> 8 ; -> New Single/PairInUseBy
                                               ;    / Sq34Col
        inc             l

; **
; BktGetCols_FGBG:  Get new color indices for both the foreground
;                   and background colors.
;
; Parameters:       b  = Old Single/PairInUseBy
;                   c  = Old Single/PairColors
;                   d  = UniqueColors
;                   e  = New Single/PairColors
;                   hl = CARD @ PalNdx
;
; Modifies:         All registers.
;
; Notes:            Called by BktGetCols and part of BktWrite_FGBG.
;

BktGetCols_FGBG::

        ; "de-register" card from PairInUseBy (free up pair if it reaches 0)

        ld              a, [bc]
        dec             a
        ld              [bc], a         ; de-reg. from Single/PairsInUseBy

        ld              a, [hl+]        ; a <- ColorIndices
        rlca                            ; put 1st ColorIndex in bits 1-2

        ld              c, [hl]         ; c <- $C0 + ( PalNdx + 1 ) << 3
        inc             l

        push            hl              ; save CARD @ TileNdx

        ; "de-register" card from TileInUseBy (free up tile if it reaches 0)

        ld              l, [hl]         ; l <- TileNdx (1-255)
        ld              h, TileInUseBy >> 8

        dec             [hl]            ; de-register card from TileInUseBy

        ; "de-register" cards previous ColorIndices from PALTAB_1

        ld              h, PALTAB_1 >> 8
.DeRegLoop
        ld              b, a            ; save ColorIndices in b
        and             $06             ; a = .....cc.
        or              c               ;   = 11pppcc1

        ld              l, a            ; hl -> PALTAB_1 (InUseBy)

        dec             [hl]            ; de-register index from PALTAB_1

        ld              a, b
        rrca
        rrca                            ; put next ColorIndex in bits 1-2

        dec             d               ; de-registered all colors ?
        jr              nz, .DeRegLoop  ; no, so keep looping

        ; check if valid color & palette indices exist for this color pair

        ld              l, e
        inc             h               ; hl -> PairInUseBy[ fffbBbb. ]

        inc             [hl]            ; "register" card with PairInUseBy

        jr              z, .SearchPal   ; if 1st card, need to search palette

        ; found valid color & palette indices for this color pair

        xor             a               ; a = OpCode for NOP
        ld              [S2_B1], a      ; store out "Bypass #1" (NO)
        ld              a, $C9          ; a = OpCode for RET
        ld              [S2_B2], a      ; store out "Bypass #2" (YES)

        ; -------------------------------------------------------------------
        ;                  Update palette information
        ; -------------------------------------------------------------------

        inc             h               ; de -> PairIndices[ fffbBbbF ]

        ld              a, [hl]         ; a <- 1pppffbb

        ld              c, a            ; save a copy of ColorIndices in c

        rrca                            ; a = b1pppffb
        or              $C1             ;   = 11pppff1

        ld              l, a
        ld              h, PALTAB_1 >> 8  ; hl -> FG's PALTAB_1

        and             $F9             ; a = 11ppp..1

        ld              b, a            ; save $C0 + PaletteIndex << 3 in b

        ld              [S2_PI + 1], a  ; store out "PaletteIndex"

        ; register card w/ PALTAB_1 with indices pulled from PairIndices

        ld              a, c            ; a = 1pppffbb
        rlca                            ;   = pppffbb1
        and             $06             ;   = .....bb.
        or              b               ;   = 11pppbb1

        inc             [hl]            ; register card w/ FG's color slot

        ld              l, a            ; l = LOW byte of BG's PALTAB_1

        inc             [hl]            ; register card w/ BG's color slot

        ld              a, c            ; a = 1pppffbb
        and             $0F             ;   = ....ffbb (ColorIndices)

        ld              h, TileInUseBy >> 8 ; *

        jp              S2_ML           ; continue @ "MakeUpNdx LOW"

.SearchPal

        ; did NOT find valid color & palette indices for this color pair

        xor             a               ; a = OpCode for NOP
        ld              [S2_B1], a      ; store out "Bypass #1" (NO)
        ld              [S2_B2], a      ; store out "Bypass #2" (NO)

        ld              a, e            ; a = ffBb.bbf

        ld              [S2_CP + 1], a  ; store out to "ColorPair"

        and             $36             ;   = ..Bb.bb.

        bit             5, a            ; is bit #5 (B of bBbb) set ?
        jr              z, .NoB         ; no, so it leave bit #3 reset

        xor             $28             ;   = ...bBbb. (BGCol << 1)
.NoB
        add             ColorToRGB & $FF
        ld              c, a            ; c -> ColorToRGB[ BGCol * 2 ]

        ld              a, [c]          ; a <- LOW byte of BGColRGB
        ld              [S2_BG + 1], a  ; store it out to "BackGround"
        ld              [S2_BL + 1], a  ; store it out to "BackGround LOW"

        ld              d, a            ; save a copy of LOW of BGColRGB in d

        inc             c               ; pos @ HIGH byte of BGColRGB

        ld              a, [c]          ; a <- HIGH byte of BGColRGB
        ld              [S2_BH + 1], a  ; store it out to "BackGround HIGH"

        ld              a, e            ; a = ffBb.bbf
        rlca                            ;   = fBb.bbff
        rlca                            ;   = Bb.bbfff
        rlca                            ;   = b.bbfffB (CY = B) *

        or              $F1             ;   = 1111fff1 ($F1 + FGCol << 1)

        add             ( ColorToRGB - $F1 ) & $FF  ; CY = 1
        ld              c, a                        ; c -> ColorToRGB
                                                    ;      [ FGCol * 2 ]

        ld              a, [c]          ; a <- LOW byte of FGColRGB
        ld              [S2_FL + 1], a  ; store it out to "ForeGround LOW"

        sbc             d               ; a =   LOW byte of FGColRGB
                                        ;     - LOW byte of BGColRGB
                                        ;     - 1

        ld              [S2_FG + 1], a  ; store it out to "ForeGround"

        inc             c               ; pos @ HIGH byte of BGColRGB

        ld              a, [c]          ; a <- HIGH byte of FGColRGB
        ld              [S2_FH + 1], a  ; store it out to "ForeGround HIGH"

        ld              hl, PALTAB_0         ; hl -> PALTAB_0 (RGB)
        ld              de, 7 << 12 + $80FF  ; d = BestCase (Level 7 = worst)

        jp              S2_PalLoop

; ---------------------------------------------------------------------------
;                   Routines copied to WRAM #0 by Bkg_Init
; ---------------------------------------------------------------------------

BktWrite_ROM:                           ; must reside in WRAM bank #0

        ; begin LONG setup process for BktWrite_XX :(

        ld              [BW_SP + 1], sp  ; save stack pointer
        ld              sp, SP_1         ; use STACK_1 for BktWrite_XX

        ei                              ; interrupts can't be off for so long

        ld              a, c            ; a = #/WRAM bank where CARD, etc. is
        ldh             [cSVBK], a      ; note new WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        ld              a, [c]          ; a <- [cROMB0] (current ROM bank #)
        ld              c, a            ; save it in c (for pushing)

        push            bc              ; save Status / [cROMB0]

        ld              a, BANK(BKG)    ; a = #/bank where BktWrite_XX is
        ldh             [cROMB0], a     ; note new ROM bank # for ISRs
        ld              [bc], a         ; switch to new ROM bank

rBW_BW: call            BktWrite_CS     ; hl = SYSTEM_RAM, de = BACKTAB word

        pop             bc              ; restore Status / [cROMB0]

        ld              a, c            ; a = [cROMB0] (previous ROM bank #)
        ld              c, cROMB0 & $FF ; c -> cROMB0 (current ROM bank #)

        ld              [c], a          ; note previous ROM bank # for ISRs
        ld              [bc], a         ; switch to previous ROM bank

        ld              a, 1            ; a = #/WRAM bank where R0, etc. is
        ldh             [cSVBK], a      ; note new WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        di                              ; disable interrupts to restore stack

rBW_SP: ld              sp, 0           ; restore stack pointer

        reti                            ; return, enabling interrupts

        ; ^ 62 cycles, 40 bytes

BktWrite_ROM_End:

; **
; SrchPal1_ROM:   Search PALTAB_0 for BGColRGB.
;
; Parameters:     b  = Case1Tab >> 8
;                 e  = $FF
;                 d  = $80 + 4 << 4 (WorstCase)
;                 hl = PALTAB_1
;
;                 [rSVBK] must already be set to WRAM bank #3.
;                 [rROMB0] must already be set to ROM bank #8.
;
;                 * S1_CN = CardNum           S1_PI = PaletteIndex
;                 * S1_CS = ColorSingle
;
;                 * Must be set in advance.
;
; Modifies:       All registers.
;

SrchPal1_ROM::

        ; gather info on which slots match BGColRGB and which are free

rS1_PalLoop:

        ld              c, $01          ; bit #0 = last-color marker

rS1_ColLoop:

        dec             h               ; pos @ PALTAB_0 (RGB)

        ld              a, [hl+]        ; a <- LOW byte of color slot's RGB

rS1_BG: sub             0               ; is it = BG color RGB ?
        add             e               ; CY reset if TRUE
        rl              c               ; record "Found" bit in c

        inc             h               ; re-pos @ PALTAB_1 (InUseBy)

        ld              a, [hl+]        ; a <- color slot's InUseBy

        add             e               ; CY reset if InUseBy = 0
        rl              c               ; record "Free" bit in b

        jr              nc, rS1_ColLoop ; if CY = 1, last color of palette

        ; using "Free" and "Found", look-up the best case for this palette

        ld              a, [bc]         ; a <- !lllbb.? (CurCase)

        cp              d               ; is CurCase < BestCase ?
        jr              nc, rS1_PalNext ; no, so keep evaluating palettes

        ; CurCase is better than BestCase, so set BestCase = CurCase

        ld              d, a            ; replace BestCase w/ CurCase

        ld              a, l            ; a = $C0 + ( PaletteIndex + 1 ) << 3
        ld              [S1_PI + 1], a  ; store it out to "PaletteIndex"

rS1_PalNext:

        ; check more palettes if possible, looking for the best case

        add             l               ; reached end-of-palettes ?
        jr              c, rS1_PalLoop  ; no, so keep evaluating palettes

        ; checked every color of PALTAB, so absolute best case is in d

        ld              a, d            ; a = !lllbb.? (lll = Level)

        rrca                            ; is BestCase still = WorstCase ?
        jp              nc, S1_DeadLock ; yes, so found no better case

        ; -------------------------------------------------------------------
        ;                  Update tile & palette information
        ; -------------------------------------------------------------------

rS1_CN: ld              hl, CNToBDD     ; hl -> CNToBDD[ CardNum ]

        ld              e, [hl]
        inc             h
        ld              d, [hl]         ; de -> BDD_TileNdx

        ; update Background Display Data for tile indices

        or              $F9             ; a = 11111bb1 (bb = BGColNdx)

        ld              [de], a         ; store out TileNdx to BDD_TileNdx

rS1_PI: add             0               ;   = 11pppbb1

rS1_CS: ld              hl, SingleIndices  ; hl -> SingleIndices
                                           ;       [ SingleColors ]

        ld              [hl], a         ; store out to SingleIndices

        ; register card w/ color slot, and store out new RGB color values

        ld              l, a            ; hl -> BG's PALTAB_1 (InUseBy)

        inc             [hl]            ; register card w/ BG's color slot

        dec             h               ; hl -> BG's PALTAB_0 (RGB)

rS1_BH: ld              [hl], 0         ; store out HIGH byte of BGColRGB
        dec             l
rS1_BL: ld              [hl], 0         ; store out LOW byte of BGColRGB

        ; restore card and update PaletteIndex & ColorIndices

        pop             hl              ; restore CARD @ PaletteIndex

        ld              [hl-], a        ; store out PaletteIndex
        ld              [hl], b         ; store out ColorIndices (b.0-1 = 0)

        ; update Background Display Data for tile attributes

        and             $38             ; a = ..ppp...
        rrca                            ;   = ...ppp..
        rrca                            ;   = ....ppp.
        rrca                            ;   = .....ppp (PalNdx)

        set             1, d            ; de -> BDD_TileAttr

        ld              [de], a         ; store out PalNdx to BDD_TileAttr

        ret

SrchPal1_ROM_End::

; **
; SrchPal2_ROM:   Search PALTAB_0 for FGColRGB and BGColRGB.
;
; Parameters:     e  = $FF
;                 d  = 7 << 4 (WorstCase)
;                 hl = PALTAB_0
;
;                 [rSVBK] must already be set to WRAM bank #3.
;                 [rROMB0] must already be set to ROM bank #8.
;
;                 * S2_ML = MakeUpNdx LOW   * S2_MH = MakeUpNdx HIGH
;                 * S2_GB = GraphicBase     * S2_CN = CardNum
;                 * S2_B1 = Bypass #1         S2_PI = PaletteIndex
;                 * S2_B2 = Bypass #2         S2_CP = ColorPair
;
;                 * Must be set in advance.
;
; Modifies:       All registers.
;

SrchPal2_ROM::

        ; gather info on which slots = FGColRGB or BGColRGB, & which are free

rS2_PalLoop:

        ld              b, Case2Tab >> 12 + $10  ; bit #4 = last-color marker

rS2_ColLoop:

        ld              a, [hl+]        ; a <- LOW byte of color slot's RGB

rS2_BG: sub             0               ; is it = BG color RGB ?
        add             e               ; CY reset if TRUE
        rl              c               ; record "Found" bit in c

rS2_FG: sub             0               ; is it = FG color RGB ?
        add             e               ; CY reset if TRUE
        rl              c               ; record "Found" bit in c

        inc             h               ; pos @ PALTAB_1 (InUseBy)

        ld              a, [hl+]        ; a <- LOW of color slot's InUseBy

        add             e               ; CY reset if InUseBy = 0
        rl              b               ; record "Free" bit in b

        dec             h               ; re-pos @ PALTAB_0 (RGB)

        jr              nc, rS2_ColLoop ; if CY = 1, last color of palette

        ; ^ 696 cycles (saves 64 cycles)

        ; using "Free" and "Found", look-up the best case for this palette

        ld              a, [bc]         ; a <- !lllffbb

        cp              d               ; is CurCase < BestCase ?
        jr              nc, rS2_PalNext ; no, so try next palette

        ; CurCase is better than BestCase, so set BestCase = CurCase

        ld              d, a            ; replace BestCase w/ CurCase

        ld              a, l            ; a = ??PPP... (PPP = PalNdx + 1 )
        sub             1 << 3 - 1      ;   = 11ppp..1 (ppp = PalNdx)

        ld              [S2_PI + 1], a  ; store it out to "PaletteIndex"

rS2_PalNext:

        add             l               ; reached end-of-palettes ?
        jr              c, rS2_PalLoop  ; no, so keep evaluating palettes

        ; checked every color of PALTAB

        ld              a, d            ; a = BG color index in bits 0-1
                                        ;     FG color index in bits 2-3
                                        ;     Level number in bits 4-6

        and             $0F             ; is BestCase still = WorstCase ?
        jp              z, S2_DeadLock  ; yes, so found no better case

rS2_GetMakeUp:

        ; form new MakeUpNdx from newly acquired color indices

        ld              h, TileInUseBy >> 8

rS2_ML: add             0               ; add LOW byte of NewMakeUpNdx
        ld              c, a            ; and save it in c

rS2_MH: ld              b, 0            ; bc -> MakeUp_GRxM[
                                        ;   + CardIndex     << 4
                                        ;   + FG ColorIndex << 2
                                        ;   + BG ColorIndex << 0 ]

rS2_GB: ld              a, 0            ; a <- "GraphicBase"
        ldh             [cSVBK], a      ; note current WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        ld              a, [bc]         ; a <- MakeUp_GRxM[ NewMakeUpNdx ]
        or              a               ; is NewTileNdx = 0 ?

        ld              l, a            ; hl -> TileInUseBy[ NewTileNdx ]

        jr              nz, rS2_RegTile ; no, so tile already exists in VRAM

        ; -------------------------------------------------------------------
        ;      Tile with new makeup not in VRAM, so allocate new tile
        ; -------------------------------------------------------------------

        ld              l, [hl]         ; l <- TileSrchNdx (0-255)
.TileSrch
        inc             l               ; pos @ next tile, wrap from 255 -> 0
        cp              [hl]            ; is TileInUseBy[ l ] = 0 ?
        jr              nz, .TileSrch   ; not yet, so continue the search

        ld              a, l              ; a = NewTileSrchNdx (= NewTileNdx)
        ld              [TileInUseBy], a  ; update TileSrchNdx

        ; let MakeUp_GRxM know that a tile w/ new makeup now exists in VRAM

        ld              [bc], a         ; MakeUp_GRxM[ NewMakeUpNdx ] <-
                                        ;   NewTileNdx

        ld              a, 6            ; switch to WRAM bank where TILE is
        ldh             [cSVBK], a      ; note current WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        ld              h, TILE >> 11   ; TILE aligned on multiples of 2048

        add             hl, hl
        add             hl, hl
        add             hl, hl          ; hl -> TILE[ NewTileNdx * 8 ]

        ; load in newly allocated tile's old MakeUpNdx & update it w/ the new

        ld              d, [hl]         ; d <- HIGH byte of OldMakeUpNdx
        ld              a, b
        ld              [hl+], a        ; store out HIGH byte of NewMakeUpNdx

        ld              e, [hl]         ; e <- LOW byte of OldMakeUpNdx
        ld              a, c
        ld              [hl+], a        ; store out LOW byte of NEW MakeUpNdx

        inc             l               ; skip OldMakeUpBank + Queue

        push            de              ; save OldMakeUpNdx

        ; update fields that affect copying with interrupts disabled

        and             $0F             ;   = ....ffbb
        di                              ; don't allow CopyToVRAM to interrupt
        add             TC00_C0 >> 8    ; add TC00_C0's page "base"

        ld              [hl+], a        ; store out ColorIndices

        ld              a, c            ; a = nnnnffbb
        and             $F0             ;   = nnnn....
        srl             b               ; put N (bit #4 of CardIndex) in CY
        rra                             ;   = Nnnnn...

        ld              [hl+], a        ; store out LOW byte of DataAddr

        ld              a, [S2_GB + 1]  ; a <- "GraphicBase"
        ld              d, a            ; save a copy in d
        add             b

        ld              [hl-], a        ; store out HIGH byte of DataAddr

        dec             l               ; skip LOW byte of DataAddr
        dec             l               ; skip ColorIndices

        ld              a, [hl]         ; a <- old MakeUpBank + Queue
        ld              [hl], d         ; store out new MakeUpBank + Queue
        ei                              ; allow CopyToVRAM to interrupt again

        ; NOTE:  Even if VBlankFunc interrupts at this exact point and
        ;        CopyToVRAM copies and removes the tile from its queue,
        ;        we can be sure that it was copied with the most up-to-date
        ;        info (namely, ColorIndices and DataAddr), and that we
        ;        don't need to add it again.

        bit             6, a            ; is COPYLIST-Queue (bit #6) set ?
        jr              nz, .QueueSkip  ; yes, so don't queue it again

        ; tile was not already in COPYLIST-Queue, so add it

        ld              e, l
        ld              d, h            ; save TILE @ MakeUpBank + Queue
        di                              ; don't allow CopyToVRAM to interrupt
        ld              hl, CL_BOTTOM

        inc             [hl]            ; inc CL_BOTTOM

        ld              l, [hl]
        dec             l               ; l <- CL_BOTTOM (before increment)

        inc             h               ; hl -> COPYLIST[ CL_BOTTOM + 1 ]

        ld              [hl], e         ; store out LOW byte of TILE address
        inc             h
        ld              [hl], d         ; store out HIGH byte of TILE address
        ei                              ; allow CopyToVRAM to interrupt again

.QueueSkip

        ; using the old MakeUpNdx, remove tile from its previous MakeUp_GRxM

        pop             de              ; restore old MakeUpBank + Queue

        ldh             [cSVBK], a      ; note current WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to WRAM bank w/ MakeUp_GRxM

        xor             a
        ld              [de], a         ; MakeUp_GRxM[ OldMakeUpNdx ] <- 0

        ; -------------------------------------------------------------------
        ;                   Update tile information
        ; -------------------------------------------------------------------

        ; register tile w/ TileInUseBy (prevent it from being re-used)

        ld              hl, TileInUseBy

        ld              l, [hl]         ; l <- TileSrchNdx (= NewTileNdx)
rS2_RegTile:
        inc             [hl]            ; register NewTileNdx w/ TileInUseBy

        ; update Background Display Data for tile indices

        ld              a, 3            ; a = #/WRAM bank that CARD is in
        ldh             [cSVBK], a      ; note current WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        ld              a, l            ; a = NewTileNdx

rS2_CN: ld              hl, CNToBDD     ; hl -> CNToBDD[ CardNum ]

        ld              e, [hl]
        inc             h
        ld              d, [hl]         ; de -> BDD_TileNdx

        ld              [de], a         ; store out NewTileNdx to BDD_TileNdx

        pop             hl              ; restore CARD @ TileNdx

        ld              [hl-], a        ; store out NewTileNdx (1-255)

rS2_B1: ret                             ; ret if BGCol & FGCol didn't change

        ; -------------------------------------------------------------------
        ;                  Update palette information
        ; -------------------------------------------------------------------

rS2_PI: ld              a, 0            ; a <- $C1 + PalNdx << 3

        ld              [hl-], a        ; store out PaletteIndex

        ; update Background Display Data for tile attributes

        and             $38             ; a = ..ppp...
        rrca                            ;   = ...ppp..
        rrca                            ;   = ....ppp.
        rrca                            ;   = .....ppp

        set             1, d            ; de -> BDD_TileAttr

        ld              [de], a         ; store out PalNdx to BDD_TileAttr

        ld              a, c            ; a = nnnnffbb (LOW of MakeUpNdx)
        and             $0F             ;   = ....ffbb (CY = 0)

        ld              [hl+], a        ; store out ColorIndices

rS2_B2: ret                             ; ret if pair found in PairIndices

        ld              c, a            ; save ColorIndices in c
        ld              b, [hl]         ; save PaletteIndex in b *

        ; register card w/ color slot, and store out new RGB color values

        rrca                            ;   = b....ffb
        or              b               ;   = 11pppff1
        dec             a               ;   = 11pppff.

        ld              l, a            ; l = LOW byte of BG's PALTAB_0

        ; store out new color & palette indices to PairIndices

        add             a, a            ;   = 1pppff..
        or              c               ;   = 1pppffbb

rS2_CP: ld              [PairIndices], a  ; store out to PairIndices

        add             a, a            ; a = pppffbb.
        and             $06             ;   = .....bb.
        or              b               ;   = 11pppbb1

        ld              h, PALTAB_0 >> 8  ; hl -> BG's PALTAB_0 (RGB)

rS2_FL: ld              [hl], 0         ; store out LOW byte of FGColRGB
        inc             l
rS2_FH: ld              [hl], 0         ; store out HIGH byte of FGColRGB

        inc             h               ; hl -> BG's PALTAB_1 (InUseBy)

        inc             [hl]            ; register card w/ FG's color slot

        ld              l, a            ; hl -> FG's PALTAB_1 (InUseBy)

        inc             [hl]            ; register card w/ BG's color slot

        dec             h               ; hl -> FG's PALTAB_0 (RGB)

rS2_BH: ld              [hl], 0         ; store out HIGH byte of BGColRGB
        dec             l
rS2_BL: ld              [hl], 0         ; store out LOW byte of BGColRGB

        ret

SrchPal2_ROM_End::

; **
; SrchPal4_ROM:   Search PALTAB_0 for Sq1ColRGB, Sq2ColRGB, Sq3ColRGB,
;                 and Sq4ColRGB.
;
; Parameters:     hl = PALTAB_1 (InUseBy)
;
;                 [rSVBK] must already be set to WRAM bank #3.
;                 [rROMB0] must already be set to ROM bank #8.
;
;                 * S4_S1 = Search #1       * S4_S2 = Search #2
;                 * S4_S3 = Search #3       * S4_S4 = Search #4
;                 * S4_B1 = BackBit #1      * S4_B2 = BackBit #2
;                 * S4_B3 = BackBit #3      * S4_B4 = BackBit #4
;                 * S4_L1 = Low #1          * S4_L2 = Low #2
;                 * S4_L3 = Low #3          * S4_L4 = Low #4
;                 * S4_H1 = High #1         * S4_H2 = High #2
;                 * S4_H3 = High #3         * S4_H4 = High #4
;                 * S4_BC = BestCase        * S4_CN = CardNum
;                   S4_CI = ColorIndices      S4_PI = PaletteIndex
;
;                 * Must be set in advance.
;
; Modifies:       All registers.
;

SrchPal4_ROM::

rS4_PalLoop:

        ; -------------------------------------------------------------------
        ;             Search PALTAB_1 for free color slots
        ; -------------------------------------------------------------------

        ld              bc, $FF << 8 + $11

rS4_FreeLoop:

        inc             l               ; skip LOW byte (not used)
        ld              a, [hl+]        ; a <- InUseBy

        add             b               ; CY reset if InUseBy = 0
        rl              c               ; record "Free" bit in c

        jr              nc, rS4_FreeLoop

        ld              de, -1 * ( PALTAB_1 - PALTAB_0 + 1 << 3 )
        add             hl, de          ; hl -> PALTAB_0 (RGB) @ color #0

        ; -------------------------------------------------------------------
        ;     Search PALTAB_0 for color matches w/ Sq1, Sq2, Sq3, and Sq4
        ; -------------------------------------------------------------------

SQ_NUM  SET             1
        REPT            4               ; check all 4 squares

SQ      EQUS            STRSUB( "{SQ_NUM}", 2, 1 )

rS4_RX  EQUS            "rS4_R{SQ}"     ; "Repeat #X"
rS4_SX  EQUS            "rS4_S{SQ}"     ; "Search #X"
rS4_BX  EQUS            "rS4_B{SQ}"     ; "BackBit #X"
rS4_LX  EQUS            "rS4_L{SQ}"     ; "Low #X"
rS4_HX  EQUS            "rS4_H{SQ}"     ; "High #X"

rS4_RX: ld              a, [hl+]        ; a <- LOW byte of color's RGB value

rS4_SX: sub             0               ; is it = LOW byte of SqXColRGB ?
        add             b               ; CY reset if TRUE
        rl              c               ; record "Found" bit in c

        inc             l               ; skip HIGH byte of color's RGB value

        jr              nc, rS4_RX      ; repeat until we've checked 4 colors

        ; based on Found & Free, get SqXLevel, SqXColNdx, and SqXNewFree

rS4_BX: ld              b, Case4Tab >> 8  ; bc -> Case4Tab[
                                          ;     BackBit << 8
                                          ;   + Free    << 4
                                          ;   + Found   << 0 ]

        ld              a, [bc]         ; a <- SqXLevel   in bits 6-7
                                        ;      SqXColNdx  in bits 4-5
                                        ;      SqXNewFree in bits 0-3

        add             a, a            ; is byte from Case4Tab = 0 ?

        IF              SQ_NUM == 1
        jp              z, S4_PalNext1  ; yes, so failed to position color
        ELSE
        jr              z, rS4_PalNext1 ; yes, so failed to position color
        ENDC

        ; position found for SqXColRGB, so record SqXLevel and SqXColNdx

        rl              e
        add             a, a
        rl              e               ; put SqXLevel in bits 0-1 of e

        add             a, a

        rl              d               ; CY = 1
        rla                             ; set bit #0 (end-of-colors marker)
        rl              d               ; put SqXColNdx in bits 0-1 of d

        IF              SQ_NUM < 4

        swap            a               ; put SqXNewFree in bits 0-3
        ld              c, a            ; and save them in c

        ; store out SqXColRGB to PALTAB_0, so other squares can re-use it

        ld              a, d            ; a = ??????cc (cc = SqXColNdx)

        add             a, a            ;   = ?????cc.
        or              $F9             ;   = 11111cc1
        add             l               ;   = 11pppcc1 (ppp = PalNdx)

        ld              l, a            ; hl -> PALTAB_0 @ color # SqXColNdx

rS4_HX: ld              [hl], 0         ; store out HIGH byte of SqXColRGB
        dec             l
rS4_LX: ld              [hl], 0         ; store out LOW byte of SqXColRGB

        and             $F8             ; a = 11ppp...

        ld              l, a            ; hl -> PALTAB_0 @ color #0

        ld              b, $FF

        PURGE           rS4_LX, rS4_HX
        ENDC

SQ_NUM  SET             SQ_NUM + 1
        PURGE           SQ, rS4_RX, rS4_SX, rS4_BX
        ENDR

        ld              c, e
        ld              b, MasterLevel >> 8  ; bc -> MasterLevel[
                                             ;   + Sq1Lvl << 6
                                             ;   + Sq2Lvl << 4
                                             ;   + Sq3Lvl << 2
                                             ;   + Sq4Lvl << 0 ]

        ld              a, [bc]         ; a <- CurCase (for this palette)
rS4_BC: cp              0               ; is it < BestCase ?
        jr              c, rS4_NewBest  ; no, so don't replace it
rS4_PalNext1:
        ld              a, l            ; a = 11PPP... (PPP = PalNdx + 1)
rS4_PalNext2:
        inc             h               ; hl -> PALTAB_1 (InUseBy)

        or              a               ; have we checked all 8 palettes ?
        jp              nz, SrchPal4    ; no, so keep checking

        ld              a, [S4_BC + 1]  ; a <- BestCase
        inc             a               ; was it = $FF (WorstCase) ?
        jr              nz, rS4_CN      ; no, so found at least 1 usable case

        ; code to handle a "deadlocked" palette should go here

        jp              S4_DeadLock     ; keep in ROM to conserve RAM

rS4_NewBest:

        ; found a better case, so record level, palette, and color indices

        or              a               ; is BestCase now = 0 ?
        ld              [S4_BC + 1], a  ; store out new BestCase

        ld              a, d            ; a = 11223344 (ColorIndices)
        ld              [S4_CI + 1], a  ; store out ColorIndices

        ld              a, l            ; a = 11PPP... (PPP = PalNdx + 1)
        ld              [S4_PI + 1], a  ; store out PaletteIndex

        jr              nz, rS4_PalNext2 ; no, so didn't find a "perfect" case

rS4_CN: ld              hl, CNToBDD     ; hl -> CNToBDD[ CardNum ]

        ld              e, [hl]
        inc             h
        ld              d, [hl]         ; de -> BDD_TileNdx

rS4_CI: ld              a, 0            ; a = 11223344 (ColorIndices)

        ld              [de], a         ; store out NewTileNdx

        ; "compact" ColorIndices and get UniqueColors (1-4)

        ld              l, a
        ld              h, UniqueTab >> 8

        ld              c, [hl]         ; c <- UniqueColorIndices
        inc             h
        ld              b, [hl]         ; b <- UniqueColors (1-4)

        ; store out Sq4ColRGB to PALTAB_0 (was not stored out above)

        add             a, a            ; a = ?????cc. (cc = Sq4ColNdx)
        or              $F9             ;   = 11111cc1
rS4_PI: add             0               ;   = 11pppcc1 (ppp = PalNdx)

        ld              l, a
        ld              h, PALTAB_0 >> 8  ; hl -> PALTAB_0 (RGB)

rS4_H4: ld              [hl], 0         ; store out HIGH byte of Sq4ColRGB
        dec             l
rS4_L4: ld              [hl], 0         ; store out LOW byte of Sq4ColRGB

        ; restore CARD and update PaletteIndex and UniqueColorIndices

        and             $F9             ; a = 11ppp..1 (ppp = PalNdx)

        pop             hl              ; restore CARD @ PaletteIndex

        ld              [hl-], a        ; store out PaletteIndex
        ld              [hl], c         ; store out ColorIndices (compacted)
        ld              c, a            ; save PaletteIndex in c

        ; store out PaletteIndex to BDD_TileAttr & set Upper Tile Bank bit

        and             $78             ; a = .1ppp... (ppp = PalNdx)
        rrca                            ;   = ..1ppp..
        rrca                            ;   = ...1ppp.
        rrca                            ;   = ....1ppp

        set             1, d            ; de -> BDD_TileAttr

        ld              [de], a         ; store out PaletteIndex

        ; update UniqueColors + Flags

        ld              d, [hl]         ; d = ColorIndices (compacted)
        dec             l

        dec             l               ; skip Single/PairInUseBy / Sq34Col
        dec             l               ; skip Single/PairColors / Sq12Col

        ld              a, [hl]         ; a <- UniqueColors + Flags

        and             $18             ; preserve CSNdx in bits 3-4
        or              $40             ; set ColoredSquares (bit #6)
        or              b               ; set UniqueColors (bits #0-#2)

        ld              [hl], a         ; update UniqueColors + Flags

        ; register Sq1ColNdx, Sq2ColNdx, Sq3ColNdx, & Sq4ColNdx w/ PALTAB_1

        ld              a, d
        rlca                            ; put 1st ColorIndex in bits 1-2

        ld              h, PALTAB_1 >> 8  ; hl -> PALTAB_1 (InUseBy)
.RegLoop
        ld              d, a            ; save UniqueColorIndices in d
        and             $06             ; a = .....cc.
        or              c               ;   = 11pppcc1

        ld              l, a            ; hl -> PALTAB_1 (InUseBy)

        inc             [hl]            ; register UniqueColor w/ PALTAB_1

        ld              a, d
        rrca
        rrca                            ; advance to next ColorIndex

        dec             b               ; dec UniqueColors
        jr              nz, .RegLoop

        ret

SrchPal4_ROM_End::

; **
; S1_DeadLock:
;
; Parameters:   h = PALTAB_1 >> 8 + 1
;

S1_DeadLock:

        ld              b, b            ; BREAK !!!

        ; re-initialize SingleInUseBy structure (in WRAM bank #3)

        dec             h               ; pos @ SingleInUseBy / PALTAB_1

        ld              d, $C0 >> 1     ; clear $00-$BF (even locations)
        ld              a, $FF
.SngLoop
        ld              [hl+], a        ; set InUseBy = $FF
        inc             l               ; skip HIGH byte (not used)

        dec             d
        jr              nz, .SngLoop

        ; re-initialize PALTAB_1 structure (in WRAM bank #3)

        xor             a
.PalLoop
        inc             l               ; skip LOW byte (not used)
        ld              [hl+], a        ; set InUseBy = 0

        cp              l
        jr              nz, .PalLoop

        ; re-initialize PairInUseBy structure (in WRAM bank #3)

        dec             a               ; a = $FF
.PairLoop
        ld              [hl+], a        ; set InUseBy = $FF

        cp              l
        jr              nz, .PairLoop

        ld              [hl+], a        ; set PairInUseBy[ 255 ] = $FF

        ; re-initialize CARD structure (in WRAM bank #3)

        ld              hl, CARD + 6
        ld              c, 240          ; do 240 cards
        ld              e, CARD_RECSIZE + 4

        ld              a, $80
.CardLoop
        ld              [hl-], a        ; set PaletteIndex = $80

        dec             l               ; pos @ Single/PairInUseBy / Sq34Col

        res             7, [hl]         ; point to ROM so we won't de-reg.

        dec             l               ; pos @ Single/PairColors / Sq12Col
        dec             l               ; pos @ UniqueColors + Flags

        set             7, [hl]         ; set Re-analyze (bit #7)

        add             hl, de          ; pos @ next CARD

        dec             c
        jr              nz, .CardLoop

        ld              [BktGetCols_Flag], a  ; set BktGetCols_Flag <> 0

        ; restore UniqueColors + Flag of current card (overwritten above)

        pop             hl              ; restore CARD @ PaletteIndex

        push            hl              ; save CARD @ PaletteIndex

        dec             l               ; skip PaletteIndex
        dec             l               ; skip ColorIndices
        dec             l               ; skip BGCol / Sq34Col
        dec             l               ; skip FGCol / Sq12Col

        res             7, [hl]         ; reset Re-analyze (bit #7)

        ld              hl, PALTAB_0    ; hl -> PALTAB_0 (RGB)
        ld              bc, Case1Tab + 1
        ld              de, 4 << 12 + $80FF

        jp              S1_ColLoop

; **
; S2_DeadLock:
;
; Parameters:   a = 0
;               l = 0
;               h = PALTAB_1 >> 8
;

S2_DeadLock:

        ld              b, b            ; BREAK !!!

        ; re-initialize SingleInUseBy structure (in WRAM bank #3)

        ld              d, $C0 >> 1     ; clear $00-$BF (even locations)

        dec             a               ; a = $FF
.SngLoop
        ld              [hl+], a        ; set InUseBy = $FF
        inc             l               ; skip HIGH byte (not used)

        dec             d
        jr              nz, .SngLoop

        ; re-initialize PALTAB_1 structure (in WRAM bank #3)

        xor             a               ; a = 0
.PalLoop
        inc             l               ; skip LOW byte (not used)
        ld              [hl+], a        ; set InUseBy = 0

        cp              l
        jr              nz, .PalLoop

        ; re-initialize PairInUseBy structure (in WRAM bank #3)

        dec             a               ; a = $FF
.PairLoop
        ld              [hl+], a        ; set InUseBy = $FF

        cp              l
        jr              nz, .PairLoop

        ld              [hl], a         ; set PairInUseBy[ 255 ] = $FF

        ; re-initialize CARD structure (in WRAM bank #3)

        ld              hl, CARD + 6
        ld              c, 240          ; do 240 cards
        ld              e, CARD_RECSIZE + 4

        ld              a, $80
.CardLoop
        ld              [hl-], a        ; set PaletteIndex = $80

        dec             l               ; pos @ Single/PairInUseBy / Sq34Col

        res             7, [hl]         ; point to ROM so we won't de-reg.

        dec             l               ; pos @ Single/PairColors / Sq12Col
        dec             l               ; pos @ UniqueColors + Flags

        set             7, [hl]         ; set Re-analyze (bit #7)

        add             hl, de          ; pos @ next CARD

        dec             c
        jr              nz, .CardLoop

        ld              [BktGetCols_Flag], a  ; set BktGetCols_Flag <> 0

        ; restore UniqueColors + Flag of current card (overwritten above)

        pop             hl              ; restore CARD @ TileNdx

        push            hl              ; save CARD @ TileNdx

        ld              a, l
        sub             7 - 2
        ld              l, a            ; hl -> CARD @ UniqueColors + Flags

        res             7, [hl]         ; reset Re-analyze (bit #7)

        ld              hl, PALTAB_0    ; hl -> PALTAB_0 (RGB)
        ld              de, 7 << 12 + $80FF

        jp              SrchPal2

; **
; S4_DeadLock:
;
; Parameters:   a = 0
;               h = PALTAB_1 >> 8
;
; Modifies:     af, bc, de, hl
;

S4_DeadLock:

        ld              b, b            ; BREAK !!!

        ; re-initialize SingleInUseBy structure (in WRAM bank #3)

        ld              d, $C0 >> 1     ; clear $00-$BF (even locations)
        dec             a               ; a = $FF
.SngLoop
        ld              [hl+], a        ; set InUseBy = $FF
        inc             l               ; skip HIGH byte (not used)

        dec             d
        jr              nz, .SngLoop

        ; re-initialize PALTAB_1 structure (in WRAM bank #3)

        xor             a
.PalLoop
        inc             l               ; skip LOW byte (not used)
        ld              [hl+], a        ; set InUseBy = 0

        cp              l
        jr              nz, .PalLoop

        ; re-initialize PairInUseBy structure (in WRAM bank #3)

        dec             a               ; a = $FF
.PairLoop
        ld              [hl+], a        ; set InUseBy = $FF

        cp              l
        jr              nz, .PairLoop

        ld              [hl+], a        ; set PairInUseBy[ 255 ] = $FF

        ; re-initialize CARD structure (in WRAM bank #3)

        ld              hl, CARD + 6
        ld              c, 240          ; do 240 cards
        ld              e, CARD_RECSIZE + 4

        ld              a, $80
.CardLoop
        ld              [hl-], a        ; set PaletteIndex = $80

        dec             l               ; pos @ Single/PairInUseBy / Sq34Col

        res             7, [hl]         ; point to ROM so we won't de-reg.

        dec             l               ; pos @ Single/PairColors / Sq12Col
        dec             l               ; pos @ UniqueColors + Flags

        set             7, [hl]         ; set Re-analyze (bit #7)

        add             hl, de          ; pos @ next CARD

        dec             c
        jr              nz, .CardLoop

        ld              [BktGetCols_Flag], a  ; set BktGetCols_Flag <> 0

        ld              hl, PALTAB_1    ; hl -> PALTAB_1 (InUseBy)

        jp              SrchPal4

STIC_OverFlow::

        ; _simple_ emulation of STIC "overflow" at left edge of screen

        ld              hl, CARD + 19 * CARD_RECSIZE + 6
        ld              de, 20 * CARD_RECSIZE
        ld              bc, $F907

ROW_NUM SET             0
        REPT            12

        ld              a, [hl-]        ; a = 11ppp..1 (ppp = PalNdx)

        rrca                            ;   = 111ppp..
        rrca                            ;   = .111ppp.
        rrca                            ;   = ..111ppp
        and             c               ;   = .....ppp

        ld              [BDD_TileAttr + 32 * ROW_NUM + 31], a

        ld              a, [hl+]        ; a = ....ffbb (ColorIndices)

        add             a, a            ;   = ...ffbb.
        or              b               ;   = 11111bb1

        ld              [BDD_TileNdx + 32 * ROW_NUM + 31], a

        IF              ROW_NUM != 11
        add             hl, de          ; pos CARD @ card #19 of next row
        ENDC

ROW_NUM SET             ROW_NUM + 1
        ENDR

        ret

; **
; ScrollPrep:
;
; Parameters:   de = SYSTEM_RAM loc. for double-registry (i.e., Row #0)
;               hl = SYSTEM_RAM loc. for de-registry (i.e., Row #11)
;
; Modifies:     af, bc, de, hl
;

ScrollPrep::

        ; -------------------------------------------------------------------
        ;       Handle registration/de-registration of Row's #0 and #11
        ; -------------------------------------------------------------------

        ld              a, [de]         ; a <- HIGH of Row #0's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .RegHigh    ; yes, so need to de-register, etc.

        dec             e               ; pos @ LOW of Row #0's BACKTAB word
        dec             l               ; pos @ LOW of Row #11's BACKTAB word

        ld              a, [de]         ; a <- LOW of Row #0's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .RegLow     ; yes, so need to de-register, etc.
.RegLoop
        dec             e               ; advance Row #0
        dec             l               ; advance Row #11

        ld              a, [de]         ; a <- HIGH of Row #0's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .RegHigh    ; yes, so need to de-register, etc.

        dec             e               ; pos @ LOW of Row #0's BACKTAB word
        dec             l               ; pos @ LOW of Row #11's BACKTAB word

        ld              a, [de]         ; a <- LOW of Row #0's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              z, .RegLoop     ; no, so check next BACKTAB word

        jr              .RegLow

.RegHigh
        inc             a               ; is it StopByte $FF ?
        ret             z               ; yes, so stop checking cards

        dec             e               ; pos @ LOW of Row #0's BACKTAB word
        dec             l               ; pos @ LOW of Row #11's BACKTAB word
.RegLow
        ; save

        push            de              ; save Row #0's pos in SYSTEM_RAM
        push            hl              ; save Row #11's pos in SYSTEM_RAM

        ld              a, 3            ; a = #/WRAM bank where CARD, etc. is
        ldh             [cSVBK], a      ; note new WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        push            de              ; save Row #0's pos in SYSTEM_RAM

        ; -------------------------------------------------------------------
        ;              Perform de-registration of Row #11's card
        ; -------------------------------------------------------------------

        ; set up pointer to Row #11's CARD record for "de-registration"

        ld              de, CARD - SYSTEM_RAM << 2 & $FFFF + 2

        add             hl, hl          ; hl = SYSTEM_RAM << 1 + CardNum << 2
        add             hl, hl          ;    = SYSTEM_RAM << 2 + CardNum << 3
        add             hl, de          ; hl -> CARD @ UniqueColors + Flags

        ld              a, [hl+]        ; a <- UniqueColors + Flags
        and             $07             ; isolate UniqueColors in bits 0-2
        ld              b, a            ; and save in b

        ld              e, [hl]         ; e <- Single/PairColors  / Sq12Col
        inc             l
        ld              d, [hl]         ; d <- Single/PairInUseBy / Sq34Col
        inc             l

        ; "de-register" card from Single/PairInUseBy (freed when reaches 0)

        ld              a, [de]         ; a <- Single/PairInUseBy
        dec             a
        ld              [de], a         ; de-register from Single/PairInUseBy

        ld              a, [hl+]        ; a <- ColorIndices
        rlca                            ; put 1st ColorIndex in bits 1-2

        ld              d, [hl]         ; d <- $C1 + PaletteIndex << 3
        inc             l

        ; "de-register" card from TileInUseBy (free up tile if it reaches 0)

        ld              l, [hl]         ; l <- TileNdx (1-255)
        ld              h, TileInUseBy >> 8

        dec             [hl]            ; de-register tile from TileInUseBy

        ; "de-register" card's ColorIndices from PALTAB_1

        ld              h, PALTAB_1 >> 8
.DeRegLoop
        ld              e, a            ; save ColorIndices in e
        and             $06             ; a = .....cc.
        or              d               ;   = 11pppcc1

        ld              l, a            ; hl -> PALTAB_1 (InUseBy)

        dec             [hl]            ; de-register index from PALTAB_1

        ld              a, e
        rrca
        rrca                            ; put next ColorIndex in bits 1-2

        dec             b               ; de-registered all colors ?
        jr              nz, .DeRegLoop  ; no, so keep looping

        ; -------------------------------------------------------------------
        ;            Perform double-registration of Row #0's card
        ; -------------------------------------------------------------------

        ; set up pointer to Row #0's CARD record for "double-registration"

        pop             hl              ; rest. Row #0's pos in SYSTEM_RAM

        ld              de, CARD - SYSTEM_RAM << 2 & $FFFF + 2

        add             hl, hl          ; hl = SYSTEM_RAM << 1 + CardNum << 2
        add             hl, hl          ;    = SYSTEM_RAM << 2 + CardNum << 3
        add             hl, de          ; hl -> CARD @ UniqueColors + Flags

        ld              a, [hl+]        ; a <- UniqueColors + Flags
        and             $07             ; isolate UniqueColors in bits 0-2
        ld              b, a            ; and save in b

        ld              e, [hl]         ; e <- Single/PairColors  / Sq12Col
        inc             l
        ld              d, [hl]         ; d <- Single/PairInUseBy / Sq34Col
        inc             l

        ; "double-register" card with Single/PairInUseBy

        ld              a, [de]         ; a <- Single/PairInUseBy
        inc             a
        ld              [de], a         ; register 2x with Single/PairInUseBy

        ld              a, [hl+]        ; a <- ColorIndices
        rlca                            ; put 1st ColorIndex in bits 1-2

        ld              d, [hl]         ; d <- $C1 + PaletteIndex << 3
        inc             l

        ; "double-register" card with TileInUseBy

        ld              l, [hl]         ; l <- TileNdx (1-255)
        ld              h, TileInUseBy >> 8

        inc             [hl]            ; register tile 2x with TileInUseBy

        ; "double-register" card's ColorIndices with PALTAB_1

        ld              h, PALTAB_1 >> 8
.DblRegLoop
        ld              e, a            ; save ColorIndices in e
        and             $06             ; a = .....cc.
        or              d               ;   = 11pppcc1

        ld              l, a            ; hl -> PALTAB_1 (InUseBy)

        inc             [hl]            ; double-register index with PALTAB_1

        ld              a, e
        rrca
        rrca                            ; put next ColorIndex in bits 1-2

        dec             b               ; de-registered all colors ?
        jr              nz, .DblRegLoop ; no, so keep looping

        ; restore

        pop             hl              ; rest. Row #11's pos in SYSTEM_RAM
        pop             de              ; rest. Row #0's pos in SYSTEM_RAM

        ld              a, 1            ; a = #/WRAM bank where SYSTEM_RAM is
        ldh             [cSVBK], a      ; note new WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        jr              .RegLoop

.RegDone

        ret

; **
; ScrollDown::
;
; Parameters:   None.
;
; Modifies:     af, bc, de, hl
;

ScrollDown_BJ::

        ;

        ld              hl, SYSTEM_RAM + ( 20 * 2 + 16 ) * 2 + 1
        ld              c, $FE

        ld              a, [hl]         ; a <- contents of StopByte location
        ld              [hl], $FF       ; overwrite with $FF
        ld              [c], a          ; save contents to HiRAM

        ld              hl, SYSTEM_RAM + ( 20 * 11 + 19 ) * 2 + 1  ; De-reg.
        ld              de, SYSTEM_RAM + ( 20 * 2 + 19 ) * 2 + 1   ; Dbl-reg.

        call            ScrollPrep      ; perform de-registry / dbl-registry

        ld              a, [c]          ; a <- contents of StopByte location
        ld              [de], a         ; restore contents

        ;

        ld              hl, SYSTEM_RAM + ( 20 * 6 + 7 ) * 2 + 1
        ld              c, $FE

        ld              a, [hl]         ; a <- contents of StopByte location
        ld              [hl], l         ; overwrite with $FF
        ld              [c], a          ; save contents to HiRAM

        ld              hl, SYSTEM_RAM + ( 20 * 11 + 16 ) * 2 + 1  ; De-reg.
        ld              de, SYSTEM_RAM + ( 20 * 0 + 16 ) * 2 + 1   ; Dbl-reg.

        call            ScrollPrep      ; perform de-registry / dbl-registry

        ld              a, [c]          ; a <- contents of StopByte location
        ld              [de], a         ; restore contents

        ;

        ld              hl, SYSTEM_RAM + ( 20 * 2 + 0 ) * 2 + 1

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              l, ( SYSTEM_RAM + ( 20 * 5 + 8 ) * 2 + 1 ) & $FF
        dec             c

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              hl, SYSTEM_RAM + ( 20 * 6 + 8 ) * 2 + 1
        dec             c

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              de, SYSTEM_RAM + ( 20 * 10 + 19 ) * 2 + 1  ; Src
        ld              hl, SYSTEM_RAM + ( 20 * 11 + 19 ) * 2 + 1  ; Dst

        call            ScrollBackwards

        ;

        ld              hl, SYSTEM_RAM + ( 20 * 1 + 0 ) * 2 + 1
        dec             c

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              de, SYSTEM_RAM + ( 20 * 1 + 16 ) * 2 + 1  ; Src
        ld              hl, SYSTEM_RAM + ( 20 * 2 + 16 ) * 2 + 1  ; Dst

        call            ScrollBackwards

        ;

        ld              hl, SYSTEM_RAM + ( 20 * 0 + 0 ) * 2 + 1
        dec             c

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              de, SYSTEM_RAM + ( 20 * 0 + 16 ) * 2 + 1  ; Src
        ld              hl, SYSTEM_RAM + ( 20 * 1 + 16 ) * 2 + 1  ; Dst

        call            ScrollBackwards

        jp              ScrlDispatch_Return

; **
; ScrollRight::
;
; Parameters:   None.
;
; Modifies:     af, bc, de, hl
;

ScrollRight::

        ld              hl, SYSTEM_RAM + ( 20 * 6 + 7 ) * 2 + 1
        ld              c, $FE

        ld              a, [hl]         ; a <- contents of StopByte location
        ld              [hl], l         ; overwrite with $FF
        ld              [c], a          ; save contents to HiRAM

        ld              hl, SYSTEM_RAM + ( 20 * 11 + 19 ) * 2 + 1  ; De-reg.
        ld              de, SYSTEM_RAM + ( 20 * 0 + 0 ) * 2 + 1    ; Dbl-reg.

        call            ScrollPrep      ; perform de-registry / dbl-registry

        ld              a, [c]          ; a <- contents of StopByte location
        ld              [de], a         ; restore contents

        ld              hl, SYSTEM_RAM + ( 20 * 0 + 0 ) * 2 + 1

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              l, ( SYSTEM_RAM + ( 20 * 6 + 7 ) * 2 + 1 ) & $FF
        dec             c

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              hl, SYSTEM_RAM + ( 20 * 6 + 8 ) * 2 + 1
        dec             c

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              de, SYSTEM_RAM + ( 20 * 11 + 18 ) * 2 + 1  ; Src
        ld              hl, SYSTEM_RAM + ( 20 * 11 + 19 ) * 2 + 1  ; Dst

        call            ScrollBackwards

        jp              ScrlDispatch_Return

; **
; ScrollDown::
;
; Parameters:   None.
;
; Modifies:     af, bc, de, hl
;

ScrollDown::

        ld              hl, SYSTEM_RAM + ( 20 * 6 + 7 ) * 2 + 1
        ld              c, $FE

        ld              a, [hl]         ; a <- contents of StopByte location
        ld              [hl], l         ; overwrite with $FF
        ld              [c], a          ; save contents to HiRAM

        ; ^ 11 cycles

        ld              hl, SYSTEM_RAM + ( 20 * 11 + 19 ) * 2 + 1  ; De-reg.
        ld              de, SYSTEM_RAM + ( 20 * 0 + 19 ) * 2 + 1   ; Dbl-reg.

        call            ScrollPrep      ; perform de-registry / dbl-registry

        ld              a, [c]          ; a <- contents of StopByte location
        ld              [de], a         ; restore contents

        ; ^ 4 cycles

        ld              hl, SYSTEM_RAM + ( 20 * 0 + 0 ) * 2 + 1

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              l, ( SYSTEM_RAM + ( 20 * 5 + 8 ) * 2 + 1 ) & $FF
        dec             c

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              hl, SYSTEM_RAM + ( 20 * 6 + 8 ) * 2 + 1
        dec             c

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ; ^ 29 cycles

        ld              de, SYSTEM_RAM + ( 20 * 10 + 19 ) * 2 + 1  ; Src
        ld              hl, SYSTEM_RAM + ( 20 * 11 + 19 ) * 2 + 1  ; Dst

        call            ScrollBackwards

        jp              ScrlDispatch_Return

; **
; ScrollBackwards::
;
; Parameters:   hl = SYSTEM_RAM address for the destination (i.e., Row #11)
;               de = SYSTEM_RAM address for the source (i.e., Row #10)
;               bc = StopByte address
;
; Modifies:     af, bc, de, hl
;

ScrollBackwards::

        ; -------------------------------------------------------------------
        ;            Perform copying of Rows #0-#10 to Rows #1-#11
        ; -------------------------------------------------------------------

        ; BACKTAB attribute word comparison -- entry point

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh   ; yes, so need to copy CARD, etc.

        dec             e               ; pos @ LOW of Src's BACKTAB word
        dec             l               ; pos @ LOW of Dst's BACKTAB word

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .CopyLow    ; yes, so need to copy CARD, etc.

.CopyLoop

        ; BACKTAB attribute word comparison -- unroll pass #1

        dec             e               ; advance Src
        dec             l               ; advance Dst

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh   ; yes, so need to copy CARD, etc.

        dec             e               ; pos @ LOW of Src's BACKTAB word
        dec             l               ; pos @ LOW of Dst's BACKTAB word

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .CopyLow    ; yes, so need to copy CARD, etc.

        ; BACKTAB attribute word comparison -- unroll pass #2

        dec             e               ; advance Src
        dec             l               ; advance Dst

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh   ; yes, so need to copy CARD, etc.

        dec             e               ; pos @ LOW of Src's BACKTAB word
        dec             l               ; pos @ LOW of Dst's BACKTAB word

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .CopyLow    ; yes, so need to copy CARD, etc.

        ; BACKTAB attribute word comparison -- unroll pass #3

        dec             e               ; advance Src
        dec             l               ; advance Dst

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh   ; yes, so need to copy CARD, etc.

        dec             e               ; pos @ LOW of Src's BACKTAB word
        dec             l               ; pos @ LOW of Dst's BACKTAB word

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .CopyLow    ; yes, so need to copy CARD, etc.

        ; BACKTAB attribute word comparison -- unroll pass #4

        dec             e               ; advance Src
        dec             l               ; advance Dst

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh   ; yes, so need to copy CARD, etc.

        dec             e               ; pos @ LOW of Src's BACKTAB word
        dec             l               ; pos @ LOW of Dst's BACKTAB word

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              z, .CopyLoop    ; no, so check next BACKTAB word

        jr              .CopyLow        ; skip StopByte check & store to HIGH

.CopyHigh

        cp              c               ; reached a StopByte ?
        jr              z, .StopByte    ; yes, so check which one & handle

        ld              [hl-], a        ; store out to HIGH of Dst

        dec             e               ; pos @ LOW of Src's BACKTAB word

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word
.CopyLow
        ld              [hl], a         ; store out to LOW of Dst

        ; save

        push            de              ; save Src
        push            hl              ; save Dst

        ld              a, 3            ; a = #/WRAM bank where CARD, etc. is
        ldh             [cSVBK], a      ; note new WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        ; ^ 24 cycles

        ; set up pointers to CARD (Src and Dst) and Background Display Data

        add             hl, hl          ; hl = SYSTEM_RAM << 1 + CardNum << 2
        add             hl, hl          ;    = SYSTEM_RAM << 2 + CardNum << 3

        ld              a, h
        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              h, a            ; hl -> Dst CARD @ DataAddr

        xor             l               ; a = ??n??NNN (Dst's CardNum)
        ld              b, a            ; save Dst's CardNum in b

        ld              a, d            ; a = HIGH byte of Src

        sla             e
        rla
        sla             e
        rla

        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              d, a            ; de -> Src CARD @ DataAddr

        ; ^ 20 cycles

        ; copy Src CARD to Dst CARD, field by field

        ld              a, [de]         ; <- LOW byte of DataAddr
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- HIGH byte of DataAddr
        inc             e
        ld              [hl+], a        ; ->

        add             c               ; CY = 0 if using GROM #0

        ld              a, [de]         ; <- UniqueColors + Flags
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- Single/PairColors  / Sq12Col
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- Single/PairInUseBy / Sq34Col
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- ColorIndices
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- PaletteIndex
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- TileIndex
        ld              [hl-], a        ; ->

        ; ^ 40 cycles

        ; output TileIndex and PaletteIndex to Background Display Data

        ld              a, b            ; a = Dst's CardNum

        ld              b, [hl]         ; b <- PaletteIndex

        ld              l, a
        ld              h, CNToBDD >> 8 ; hl -> CNToBDD[ Dst's CardNum ]

        ld              a, [de]         ; a <- TileIndex

        jr              c, .NotGROM0    ; if CY = 1, card not using GROM #0

        ld              a, b            ; a = 11pppbb1 (PaletteIndex)
        or              $F9             ;   = 11111bb1
.NotGROM0
        ld              e, [hl]
        inc             h
        ld              d, [hl]         ; de -> BDD_TileNdx

        ld              [de], a         ; store out Src's TileNdx

        ld              a, b            ; a = $C1 + PaletteIndex << 3

        and             $38             ;   = ..ppp...
        rrca                            ;   = ...ppp..
        rrca                            ;   = ....ppp.
        rrca                            ;   = .....ppp

        set             1, d            ; bc -> BDD_PalNdx

        ld              [de], a         ; store out Src's PaletteIndex

        ; ^ 28/30 cycles

        ; restore

        pop             hl              ; rest. Dst
        pop             de              ; rest. Src

        ld              a, 1            ; a = #/WRAM bank where SYSTEM_RAM is
        ldh             [cSVBK], a      ; note new WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        ; ^ 14 cycles

        ; BACKTAB attribute word comparison -- end of normal copy

        dec             e               ; advance Src
        dec             l               ; advance Dst

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh   ; yes, so need to copy CARD, etc.

        dec             e               ; pos @ LOW of Src's BACKTAB word
        dec             l               ; pos @ LOW of Dst's BACKTAB word

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jp              z, .CopyLoop    ; no, so check next BACKTAB word

        jr              .CopyLow        ; skip StopByte check & store to HIGH

        ; -------------------------------------------------------------------
        ;              StopByte -- restore, compare, then handle
        ; -------------------------------------------------------------------

.StopByte

        ld              a, [c]          ; a <- original contents of StopByte
        ld              [de], a         ; restore StopByte location

        inc             c               ; increment StopByte pointer

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh1  ; yes, so need to copy CARD, etc.

        dec             e               ; pos @ LOW of Src's BACKTAB word
        dec             l               ; pos @ LOW of Dst's BACKTAB word

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .CopyLow1   ; yes, so need to copy CARD, etc.
.Adv16
        ld              a, c            ; a = current StopByte pointer
        inc             a               ; reached end-of-copy StopByte ?
        ret             z               ; yes, so no more comparisons to make

        ; BACKTAB attribute word comparison -- 16-bit decrement

        dec             de              ; advance Src
        dec             hl              ; advance Dst

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh   ; yes, so need to copy CARD, etc.

        dec             e               ; pos @ LOW of Src's BACKTAB word
        dec             l               ; pos @ LOW of Dst's BACKTAB word

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jp              z, .CopyLoop    ; no, so check next BACKTAB word

        jp              .CopyLow        ; skip StopByte check & store to HIGH

.CopyHigh1

        ld              [hl-], a        ; store out to HIGH of Dst

        dec             e               ; pos @ LOW of Src's BACKTAB word

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word
.CopyLow1
        ld              [hl], a         ; store out to LOW of Dst

        ; save

        push            de              ; save Src
        push            hl              ; save Dst

        ld              a, 3            ; a = #/WRAM bank where CARD, etc. is
        ldh             [cSVBK], a      ; note new WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        ; set up pointers to CARD (Src and Dst) and Background Display Data

        add             hl, hl          ; hl = SYSTEM_RAM << 1 + CardNum << 2
        add             hl, hl          ;    = SYSTEM_RAM << 2 + CardNum << 3

        ld              a, h
        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              h, a            ; hl -> Dst CARD @ DataAddr

        xor             l               ; a = ??n??NNN (Dst's CardNum)
        ld              b, a            ; save Dst's CardNum in b

        ld              a, d            ; a = HIGH byte of Src

        sla             e
        rla
        sla             e
        rla

        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              d, a            ; de -> Src CARD @ DataAddr

        ; copy Src CARD to Dst CARD, field by field

        ld              a, [de]         ; <- LOW byte of DataAddr
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- HIGH byte of DataAddr
        inc             e
        ld              [hl+], a        ; ->

        add             c               ; CY = 0 if using GROM #0

        ld              a, [de]         ; <- UniqueColors + Flags
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- Single/PairColors  / Sq12Col
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- Single/PairInUseBy / Sq34Col
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- ColorIndices
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- PaletteIndex
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- TileIndex
        ld              [hl-], a        ; ->

        ; output TileIndex and PaletteIndex to Background Display Data

        ld              a, b            ; a = Dst's CardNum

        ld              b, [hl]         ; b <- PaletteIndex

        ld              l, a
        ld              h, CNToBDD >> 8 ; hl -> CNToBDD[ Dst's CardNum ]

        ld              a, [de]         ; <- TileIndex

        jr              c, .NotGROM0_1  ; if CY = 1, card not using GROM #0

        ld              a, b            ; a = 11pppbb1 (PaletteIndex)
        or              $F9             ;   = 11111bb1
.NotGROM0_1
        ld              e, [hl]
        inc             h
        ld              d, [hl]         ; de -> BDD_TileNdx

        ld              [de], a         ; store out Src's TileNdx

        ld              a, b            ; a = $C1 + PaletteIndex << 3

        and             $38             ;   = ..ppp...
        rrca                            ;   = ...ppp..
        rrca                            ;   = ....ppp.
        rrca                            ;   = .....ppp

        set             1, d            ; bc -> BDD_PalNdx

        ld              [de], a         ; store out Src's PaletteIndex

        ; restore

        pop             hl              ; rest. Dst
        pop             de              ; rest. Src

        ld              a, 1            ; a = #/WRAM bank where SYSTEM_RAM is
        ldh             [cSVBK], a      ; note new WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        jr              .Adv16

; **
; ScrollUp::
;
; Parameters:   None.
;
; Modifies:     af, bc, de, hl
;

ScrollUp::

        ld              hl, SYSTEM_RAM + ( 20 * 10 + 19 ) * 2 + 1
        ld              c, $FE

        ld              a, [hl]         ; a <- contents of StopByte location
        ld              [hl], $FF       ; overwrite with $FF
        ld              [c], a          ; save contents to HiRAM

        ; ^ 11 cycles

        ld              hl, SYSTEM_RAM + ( 20 * 0 + 19 ) * 2 + 1   ; De-reg.
        ld              de, SYSTEM_RAM + ( 20 * 11 + 19 ) * 2 + 1  ; Dbl-reg.

        call            ScrollPrep      ; perform de-registry / dbl-registry

        ld              a, [c]          ; a <- contents of StopByte location
        ld              [de], a         ; restore contents

        ; ^ 4 cycles

        ld              hl, SYSTEM_RAM + ( 20 * 11 + 19 ) * 2

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              l, ( SYSTEM_RAM + ( 20 * 7 + 7 ) * 2 ) & $FF
        dec             c

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              hl, SYSTEM_RAM + ( 20 * 6 + 7 ) * 2
        dec             c

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ; ^ 29 cycles

        ld              de, SYSTEM_RAM + ( 20 * 1 + 0 ) * 2  ; Src
        ld              hl, SYSTEM_RAM + ( 20 * 0 + 0 ) * 2  ; Dst

        call            ScrollForwards

        jp              ScrlDispatch_Return

; **
; ScrollLeft::
;
; Parameters:   None.
;
; Modifies:     af, bc, de, hl
;

ScrollLeft::

        ld              hl, SYSTEM_RAM + ( 20 * 11 + 18 ) * 2 + 1
        ld              c, $FE

        ld              a, [hl]         ; a <- contents of StopByte location
        ld              [hl], $FF       ; overwrite with $FF
        ld              [c], a          ; save contents to HiRAM

        ; ^ 11 cycles

        ld              hl, SYSTEM_RAM + ( 20 * 0 + 0 ) * 2 + 1    ; De-reg.
        ld              de, SYSTEM_RAM + ( 20 * 11 + 19 ) * 2 + 1  ; Dbl-reg.

        call            ScrollPrep      ; perform de-registry / dbl-registry

        ld              a, [c]          ; a <- contents of StopByte location
        ld              [de], a         ; restore contents

        ; ^ 4 cycles

        ld              hl, SYSTEM_RAM + ( 20 * 11 + 19 ) * 2

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              l, ( SYSTEM_RAM + ( 20 * 6 + 8 ) * 2 ) & $FF
        dec             c

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ld              hl, SYSTEM_RAM + ( 20 * 6 + 7 ) * 2
        dec             c

        ld              a, [hl]
        ld              [hl], c
        ld              [c], a

        ; ^ 29 cycles

        ld              de, SYSTEM_RAM + ( 20 * 0 + 1 ) * 2  ; Src
        ld              hl, SYSTEM_RAM + ( 20 * 0 + 0 ) * 2  ; Dst

        call            ScrollForwards

        jp              ScrlDispatch_Return

; **
; ScrollForwards::
;
; Parameters:   hl = SYSTEM_RAM address for the destination (i.e., Row #0)
;               de = SYSTEM_RAM address for the source (i.e., Row #1)
;
; Modifies:     af, bc, de, hl
;

ScrollForwards::

        ; -------------------------------------------------------------------
        ;            Perform copying of Rows #1-#11 to Rows #0-#10
        ; -------------------------------------------------------------------

        ; BACKTAB attribute word comparison -- entry point

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .CopyLow    ; yes, so need to copy CARD, etc.

        inc             e               ; pos @ HIGH of Src's BACKTAB word
        inc             l               ; pos @ HIGH of Dst's BACKTAB word

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh   ; yes, so need to copy CARD, etc.

.CopyLoop

        ; BACKTAB attribute word comparison -- unroll pass #1

        inc             e               ; advance Src
        inc             l               ; advance Dst

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .CopyLow    ; yes, so need to copy CARD, etc.

        inc             e               ; pos @ HIGH of Src's BACKTAB word
        inc             l               ; pos @ HIGH of Dst's BACKTAB word

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh   ; yes, so need to copy CARD, etc.

        ; BACKTAB attribute word comparison -- unroll pass #2

        inc             e               ; advance Src
        inc             l               ; advance Dst

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .CopyLow    ; yes, so need to copy CARD, etc.

        inc             e               ; pos @ HIGH of Src's BACKTAB word
        inc             l               ; pos @ HIGH of Dst's BACKTAB word

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh   ; yes, so need to copy CARD, etc.

        ; BACKTAB attribute word comparison -- unroll pass #3

        inc             e               ; advance Src
        inc             l               ; advance Dst

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .CopyLow    ; yes, so need to copy CARD, etc.

        inc             e               ; pos @ HIGH of Src's BACKTAB word
        inc             l               ; pos @ HIGH of Dst's BACKTAB word

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh   ; yes, so need to copy CARD, etc.

        ; BACKTAB attribute word comparison -- unroll pass #4

        inc             e               ; advance Src
        inc             l               ; advance Dst

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .CopyLow    ; yes, so need to copy CARD, etc.

        inc             e               ; pos @ HIGH of Src's BACKTAB word
        inc             l               ; pos @ HIGH of Dst's BACKTAB word

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              z, .CopyLoop    ; no, so check next BACKTAB word

        jr              .CopyHigh

.CopyLow

        cp              c               ; reached a StopByte ?
        jr              z, .StopByte    ; yes, so check which one & handle it

        ld              [hl+], a        ; store out to LOW of Dst

        inc             e               ; pos @ HIGH of Src's BACKTAB word

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word
.CopyHigh
        ; save

        push            de              ; save Src
        push            hl              ; save Dst

        ld              [hl-], a        ; store out to HIGH of Dst

        ld              a, 3            ; a = #/WRAM bank where CARD, etc. is
        ldh             [cSVBK], a      ; note new WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        ; set up pointers to CARD (Src and Dst) and Background Display Data

        dec             e               ; re-pos @ LOW of Src's BACKTAB word

        add             hl, hl          ; hl = SYSTEM_RAM << 1 + CardNum << 2
        add             hl, hl          ;    = SYSTEM_RAM << 2 + CardNum << 3

        ld              a, h
        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              h, a            ; hl -> Dst CARD @ DataAddr

        xor             l               ; a = ??n??NNN (Dst's CardNum)
        ld              b, a            ; save Dst's CardNum in b

        ld              a, d            ; a = HIGH byte of Src

        sla             e
        rla
        sla             e
        rla

        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              d, a            ; de -> Src CARD @ DataAddr

        ; copy Src CARD to Dst CARD, field by field

        ld              a, [de]         ; <- LOW byte of DataAddr
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- HIGH byte of DataAddr
        inc             e
        ld              [hl+], a        ; ->

        add             c               ; CY = 0 if using GROM #0

        ld              a, [de]         ; <- UniqueColors + Flags
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- Single/PairColors  / Sq12Col
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- Single/PairInUseBy / Sq34Col
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- ColorIndices
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- PaletteIndex
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- TileIndex
        ld              [hl-], a        ; ->

        ; output TileIndex and PaletteIndex to Background Display Data

        ld              a, b            ; a = Dst's CardNum

        ld              b, [hl]         ; b <- PaletteIndex

        ld              l, a
        ld              h, CNToBDD >> 8 ; hl -> CNToBDD[ Dst's CardNum ]

        ld              a, [de]         ; <- TileIndex

        jr              c, .NotGROM0_1  ; if CY = 1, card not using GROM #0

        ld              a, b            ; a = 11pppbb1 (PaletteIndex)
        or              $F9             ;   = 11111bb1
.NotGROM0_1
        ld              e, [hl]
        inc             h
        ld              d, [hl]         ; de -> BDD_TileNdx

        ld              [de], a         ; store out Src's TileNdx

        ld              a, b            ; a = $C1 + PaletteIndex << 3

        and             $38             ;   = ..ppp...
        rrca                            ;   = ...ppp..
        rrca                            ;   = ....ppp.
        rrca                            ;   = .....ppp

        set             1, d            ; bc -> BDD_PalNdx

        ld              [de], a         ; store out Src's PaletteIndex

        ; restore

        pop             hl              ; rest. Dst
        pop             de              ; rest. Src

        ld              a, 1            ; a = #/WRAM bank where SYSTEM_RAM is
        ldh             [cSVBK], a      ; note new WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        ; BACKTAB attribute word comparison -- end of normal copy

        inc             e               ; advance Src
        inc             l               ; advance Dst

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .CopyLow    ; yes, so need to copy CARD, etc.

        inc             e               ; pos @ HIGH of Src's BACKTAB word
        inc             l               ; pos @ HIGH of Dst's BACKTAB word

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jp              z, .CopyLoop    ; no, so check next BACKTAB word

        jr              .CopyHigh       ; skip StopByte check & store to HIGH

        ; -------------------------------------------------------------------
        ;              StopByte -- restore, compare, then handle
        ; -------------------------------------------------------------------

.StopByte

        ld              a, [c]          ; a <- original contents of StopByte
        ld              [de], a         ; restore StopByte location

        inc             c               ; increment StopByte pointer

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jr              nz, .CopyLow1   ; yes, so need to copy CARD, etc.

        inc             e               ; pos @ HIGH of Src's BACKTAB word
        inc             l               ; pos @ HIGH of Dst's BACKTAB word

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jr              nz, .CopyHigh1  ; yes, so need to copy CARD, etc.
.Adv16
        ld              a, c            ; a = current StopByte pointer
        inc             a               ; reached end-of-copy StopByte ?
        ret             z               ; yes, so no more comparisons to make

        ; BACKTAB attribute word comparison -- 16-bit increment

        inc             de              ; advance Src
        inc             hl              ; advance Dst

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word

        cp              [hl]            ; do LOW bytes of Rows differ ?
        jp              nz, .CopyLow    ; yes, so need to copy CARD, etc.

        inc             e               ; pos @ HIGH of Src's BACKTAB word
        inc             l               ; pos @ HIGH of Dst's BACKTAB word

        ld              a, [de]         ; a <- HIGH of Src's BACKTAB word

        cp              [hl]            ; do HIGH bytes of Rows differ ?
        jp              z, .CopyLoop    ; no, so check next BACKTAB word

        jp              .CopyHigh       ; skip StopByte check & store to HIGH

.CopyLow1
        ld              [hl+], a        ; store out to LOW of Dst

        inc             e               ; pos @ HIGH of Src's BACKTAB word

        ld              a, [de]         ; a <- LOW of Src's BACKTAB word
.CopyHigh1
        ; save

        push            de              ; save Src
        push            hl              ; save Dst

        ld              [hl-], a        ; store out to HIGH of Dst

        ld              a, 3            ; a = #/WRAM bank where CARD, etc. is
        ldh             [cSVBK], a      ; note new WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        ; set up pointers to CARD (Src and Dst) and Background Display Data

        dec             e               ; re-pos @ LOW of Src's BACKTAB word

        add             hl, hl          ; hl = SYSTEM_RAM << 1 + CardNum << 2
        add             hl, hl          ;    = SYSTEM_RAM << 2 + CardNum << 3

        ld              a, h
        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              h, a            ; hl -> Dst CARD @ DataAddr

        xor             l               ; a = ??n??NNN (Dst's CardNum)
        ld              b, a            ; save Dst's CardNum in b

        ld              a, d            ; a = HIGH byte of Src

        sla             e
        rla
        sla             e
        rla

        add             ( CARD - SYSTEM_RAM << 2 ) >> 8 & $FF
        ld              d, a            ; de -> Src CARD @ DataAddr

        ; copy Src CARD to Dst CARD, field by field

        ld              a, [de]         ; <- LOW byte of DataAddr
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- HIGH byte of DataAddr
        inc             e
        ld              [hl+], a        ; ->

        add             c               ; CY = 0 if using GROM #0

        ld              a, [de]         ; <- UniqueColors + Flags
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- Single/PairColors  / Sq12Col
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- Single/PairInUseBy / Sq34Col
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- ColorIndices
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- PaletteIndex
        inc             e
        ld              [hl+], a        ; ->

        ld              a, [de]         ; <- TileIndex
        ld              [hl-], a        ; ->

        ; output TileIndex and PaletteIndex to Background Display Data

        ld              a, b            ; a = Dst's CardNum

        ld              b, [hl]         ; b <- PaletteIndex

        ld              l, a
        ld              h, CNToBDD >> 8 ; hl -> CNToBDD[ Dst's CardNum ]

        ld              a, [de]         ; <- TileIndex

        jr              c, .NotGROM0    ; if CY = 1, card not using GROM #0

        ld              a, b            ; a = 11pppbb1 (PaletteIndex)
        or              $F9             ;   = 11111bb1
.NotGROM0
        ld              e, [hl]
        inc             h
        ld              d, [hl]         ; de -> BDD_TileNdx

        ld              [de], a         ; store out Src's TileIndex

        ld              a, b            ; a = $C1 + PaletteIndex << 3

        and             $38             ;   = ..ppp...
        rrca                            ;   = ...ppp..
        rrca                            ;   = ....ppp.
        rrca                            ;   = .....ppp

        set             1, d            ; bc -> BDD_PalNdx

        ld              [de], a         ; store out Src's PaletteIndex

        ; restore

        pop             hl              ; rest. Dst
        pop             de              ; rest. Src

        ld              a, 1            ; a = #/WRAM bank where SYSTEM_RAM is
        ldh             [cSVBK], a      ; note new WRAM bank # for ISRs
        ldh             [rSVBK], a      ; switch to new WRAM bank

        jr              .Adv16
